/* get_next_record 

   Reads next record from stdin into rec_buf and leaves file pointer at start
   of next record. This will be an segment control word if there are any,
   otherwise the start of the record label.
   Returns record_length on successful execution else returns -1 if EOF
   else -2 and error message if record starts badly. In this case it skips
   to the next record mark so that the next record can be read.
   If record "ends" with anything other than a record mark it prints an
   error message, skips to record mark and returns -3. In the latter case
   the record may be usable (but length is unknown).
*/

  
#define SCW_LEN 5
#define FIELD_MARK   0x1E	/* end of field in marc format */
#define SUBFIELD_MARK   0x1F	/* subfield marker ... */
#define RECORD_MARK     0x1D

#define TRUE      1
#define FALSE     0



get_next_record(rec_buf, has_scw) /* get next logical record from file */
     char * rec_buf ;
     int has_scw ;		/* True if there are segment control words */
{
  char scw[SCW_LEN] ;
  char span ;
  register char *p, *op ;
  int slen ;
  int rlen = 0 ;
  register int ch, i ;
  int badstart = FALSE ;

  op = rec_buf ;
  if ( has_scw ) {
    p = scw ;
    for ( i = 0 ; i < SCW_LEN ; i++ ) {
      if ((ch = getchar()) == EOF ) break ;
      if ( ! isdigit(ch) ) badstart = TRUE ;
      *p++ = ch ;
    }
    if ( ch == EOF ) return -1 ;
    if ( badstart ) {
      badstarts++ ;
      if ( ! searching && ! silent )
	fprintf(stderr, "Record starts badly with %.5s, skipping..\n", scw) ;
      if (skip_to_end() == EOF ) return -1 ;
      else return -2 ;
    }
    span = scw[0] ;
    (void) sscanf(scw + 1, "%4d", &slen) ;
    
    /* read next segment into buffer */
    for (i = 0 ; i < slen - SCW_LEN ; i++ ) {
      if ( ( ch = getchar() ) == EOF ) break ;
      *op++ = ch ;
    }
    if ( ch == EOF ) return -1 ;     
    rlen += slen - SCW_LEN ;
    /* spanning indicator '1': record begins but doesn't end in this segment 
       spanning indicator '2': record neither begins or ends in this segment */
    if ( span == '1' ) {
      while (span != '3') {
	p = scw ;
	for ( i = 0 ; i < SCW_LEN ; i++ ) {
	  if ((ch = getchar()) == EOF ) break ;
	  *p++ = ch ;
	}
	if ( ch == EOF ) return -1 ;
	span = scw[0] ;
	(void) sscanf(scw + 1, "%4d", &slen) ;
	/* read segment and append into rec_buf */
	for (i = 0 ; i < slen - SCW_LEN ; i++ ) {
	  if ( ( ch = getchar() ) == EOF ) break ;
	  *op++ = ch ;
	}
	if ( ch == EOF ) return -1 ;     
	rlen += slen - SCW_LEN ;
      }
    }
  }
  else {			/* no segment control words */
    for ( i = 0 ; i < 5 ; i++ )  {
      if ((ch = getchar()) == EOF ) break ;
      if ( ! isdigit(ch) ) badstart = TRUE ;
      *op++ = ch ;
    }
    if ( ch == EOF ) return -1 ;
    if ( badstart ) {
      badstarts++ ;
      if ( ! searching && ! silent )
	fprintf(stderr,
		"Record starts badly with %.5s, skipping..\n", rec_buf) ;
      if (skip_to_end() == EOF) return -1 ;
      else return -2 ;
    }
    (void) sscanf(rec_buf, "%5d", &slen) ;
    for (i = 0 ; i < slen - 5 ; i++ ) {
      if ( ( ch = getchar() ) == EOF ) break ;
      *op++ = ch ;
    }
    if ( ch == EOF ) return -1 ;     
    rlen = slen ;
  } /* No segment control words */
    
  *op = '\0' ;
    
  if ( rec_buf[rlen - 1] != RECORD_MARK ) {
    badends++ ;
    if ( ! searching && ! silent ) {
      fprintf(stderr, "Record ends badly, record follows:\n%s\n", rec_buf) ;
      fprintf(stderr, "Skipped to record mark (or EOF)\n\n") ;
    }
    if ( skip_to_end() == 0 ) return -3 ;
    else return -1 ;
  }

  return rlen ;
}
  
skip_to_end()			/* Skip to next record mark */
{
  register int ch ;
  register int i = 0 ;

  while ( ( ch = getchar() ) != RECORD_MARK && ch != EOF ) i++ ;
  if ( ! silent )
    fprintf(stderr, "Skipped %d chars through record mark\n", i) ;
  if ( ch == RECORD_MARK ) return 0 ;
  else return EOF ;
}

