IDENTIFICATION DIVISION.

program-name projc.
 
 
* PROGRAM PROJC written by Gill Venner,
 
* Modified Aug. 86 by Stephen - up to 30 Subject descriptors,
* 505 removed (again) except for analytics, ..
 
 
* 16 Nov 87 copied to project Sun network SW
 
*   reads 8-bit MARC file, translates each block into ascii, packs segments into whole records
*   selects required tags and subfields, using table (tag-values).
*   adds appropriate punctuation after performing any exception processing
*   (given by numeric "punctuation" in table).
*   prepares file(s) for transfer to apple network
*     separates field groups in each record by # and inserts ^ _ | $ @ thus:
*
*   ^ indicates title-like data
*   | indicates personal name
*   $ indicates corporate name
*   _ indicates no indexing required
*   @ indicates ed/pub data
*
*   nb accents are preceded by \.
 
environment division.
i-o section.
file-control.
                select eight-bit assign DSKC recording mode f.
              select proj assign DSKC recording mode ascii.
              select tmp assign DSKC recording mode ascii.
data division.
file section.
*                                         tmp file is for diagnostics of peculiar records
fd tmp value id diag-id.
01 DREC DISPLAY-7.
 02 DCNO PIC X(10)BB.
 02 DMESS OCCURS 12.
  03 DTAG PIC XXXB.
  03 DI1 PIC X.
  03 DI2 PIC XBB.
*                                         input file (written by ncdrea.cbl)
fd eight-bit value id eight-id.
01 erec display-9.
 02 e pic x occurs 2048.
*                                          output file(s) to go to floppy disc
fd proj value id proj-id.
01 prec display-7.
 02 p  pic x occurs 900.
01 p300rec pic x(300) display-7.
01 p600rec pic x(600) display-7.
01 pr display-7.
 02 fil pic x(300).
 02 p600 pic x(300).
 02 p900 pic x(300).
 
working-storage section.
*                                      working storage for pack and convert sections
*                                      *********************************************
01 block-accents display-7.
 02 b-acc pic x occurs 2048.
*                       the following redefinitions help us to translate 8-bit code into 7-bit ascii
01 ebcdc display-9.
 02 fil pic xxx.
 02 e-ch pic x .
01 bin redefines ebcdc comp pic 99999.
01 asci redefines bin display-7.
 02 fil pic xxxx.
 02 a-ch pic x.
01 chars.
 02 char pic x occurs 6.
77 eight-id pic x(9).
*                            hex is used to convert length of o/p file into hex
77 hex pic 99 comp occurs 6.
*                            the following constants are used during 8-bit translation
77 hex-5E pic 9(5) comp value 94.
77 hex-61 pic 9(5) comp value 97.
77 hex-7B pic 9(5) comp value 123.
77 hex-80 pic 9(5) comp value 128.
77 hex-E0 pic 9(5) comp value 224.
77 hex-BB pic 9(5) comp value 187.
77 hex-FE pic 9(5) comp value 254.
77 hex-BC pic 9(5) comp value 188.
*                                current MARC block translated into ascii (accents are in corres. position in BLOCK-ACCENTS)
01 current-block display-7.
 02 first-scw pic x(5) .
 02 fil pic x(2043).
01 cb redefines current-block display-7.
 02 cbc pic x occurs 2048.
01 cg redefines cb display-7 pic x(300).
01 diag-id.
 02 diag-stem pic xxxx.
 02 fil pic x(5) value "  DIA".
*                           current MARC record, assembled from segments (accents are in corres. position in RECORD-ACCENTS)
01 frec display-7.
 02 fstart.
  03 frl pic x(5).
  03 fstatus pic x.
  03 fil pic x(4).
 02 fcno redefines fstart.
  03 fcno3 pic xxx.
  03 fil pic x(5).
  03 f9 pic x.
  03 f10 pic x.
 02 fil pic xx.
 02 fbase pic 9(5).
 02 fil pic x(7).
 02 fdirs occurs 100.
  03 ftg pic xxx.
  03 fil pic x(9).
01 fr  redefines frec display-7.
 02 f pic x occurs 4095.
01 fg redefines fr display-7.
 02 fil pic x(300).
01 previous-fcno  display-7.
 02 previous-char1 pic x.
 02 fil pic x(9).
77 ffirst pic 9 comp value 0.
 88 searching value 0.
77 skipno pic 9999.
77 save-fcno pic x(10) display-7.
01 proj-id.
 02 proj-stem pic x(4).
 02 proj-no pic xx value "01".
 02 fil pic xxx value "TMP".
77 pfileno pic 99 value 1.
01 record-accents display-7.
 02 r-acc pic x occurs 4000.
*                          the following redefinitions are merely space-saving
77 record-save-area redefines record-accents pic x(900) display-7.
01 rsubjs redefines record-accents display-7.
 02 rsubj occurs 30.
  03 rmark pic x.
  03 rsubjct pic x(120).
 02 RPUNC PIC X OCCURS 30.
01 scw.
 02 spanning-indicator pic 9.
 88 whole-record value 0.
 88 last-segment-of-record value 3.
 02 segment-length pic 9(4).
01 next-scw.
 02 next-scw-ch pic x occurs 5.
*                                          l is no. of chars. left in segment
77 l pic 9(4) comp.
*                                          v is char. position in rec. being assembled
77 v pic 9(4) comp value 1.
*                                          u is char. position in current i/p block
77 u pic 9(4) comp.
77 number-of-blocks pic 9(5) comp value 0.
77 number-of-records pic 9(5) comp value 0.
77 tot-no-of-recs pic 9(5) comp value 0.
*                                                working storage for process record section
*                                                ******************************************
77 a pic 999 comp.
01 a082s display-7.
*                            nb assumes no dewey > 30 chs
 02 a082 pic x(40) occurs 10.
77 a082-count pic 99 comp.
77 a240 pic 999 comp.
77 a245 pic 999 comp.
77 A505 PIC 9.
77 ADISP PIC 99999.
77 ANALYT PIC 999.
77 bell pic 99 comp value 14.
01 bell-field redefines bell display-7.
 02 fil pic xxxx.
 02 bell-char pic x.
77 b pic 999 comp.
77 c  pic 9999 comp.
77 cc pic 9999 comp.
*                                          save area for appropriate piece of tag-values
01 current-tag display-7.
 02 ctag pic xxx.
 02 cind pic x.
 02 csflds occurs 7.
  03 csf pic x.
  03 cpunc pic xxx.
77 D PIC 99 COMP.
77 date1 pic 9999.
77 datecode pic 999.
77 DCed pic xx.
77 exch-id pic x(9).
*                   FDIR is save area for 12-byte directory entry (in FREC) for current tag
01 fdir display-7.
 02 ftag pic 999.
 02 fgp redefines ftag.
  03 fhash pic x.
  03 fil pic xx.
 02 flev pic 9.
 02 flen pic 999.
 02 frep pic 9.
 02 fdis pic 9999.
*                               required subfields for one tag are "stringed" into GFIELD with punc
01 gfield display-7 value space.
 02 g pic x occurs 500.
77 i1 pic 9.
77 i2 pic 9.
77 ind1 pic x.
01 initi display-7.
 02 init1 pic x.
 02 fil pic x.
 02 init3 pic x.
 02 FIL PIC X.
 02 INIT5 PIC X.
77 j pic 999 comp.
77 LANG PIC XXX DISPLAY-7.
77 mark pic x display-7.
77 max-chars pic 9(7) comp.
77 MESS PIC X(7).
77 next-year pic 9999 value 1987.
77 pc pic 9999 comp .
77 pcj pic 9(6) comp value 0.
77 pcsave pic 9999 comp value 0.
*                               used to interpret pseudo punctuation character in TAG-VALUES
01 punc display-7.
 02 punc1 pic x.
 88 subfield-is-248g  value "1".
 88 subfield-is-100h  value "2".
 88 may-have-article  value "3".
 88 sf-is-082c        value "6".
 88 tag-is-505        value "7".
 02 punc2 pic x.
 02 fil pic x.
*               SS is temporary storage area for wanted subfield
01 ss display-7.
 02 sf pic x.
 02 scgroup.
  03 sc pic x occurs 500.
 02 scgp1 redefines scgroup.
  03 sa1 pic x.
  03 sf1 pic x(499).
 02 scgp2 redefines scgp1.
  03 sa2 pic xx.
  03 sf2 pic x(498).
 02 scgp3 redefines scgp2.
  03 sa3 pic xxx.
  03 sf3 pic x(497).
 02 scgp4 redefines scgp3.
  03 sa4 pic xxxx.
  03 sf4 pic x(496).
 02 scgp5 redefines scgp4.
  03 sa5 pic x(5).
  03 sf5 pic x(495).
01 stemp redefines ss display-7.
 02 st pic x occurs 501.
*                                    the next redefinition is merely space-saving
01 ssubjct redefines stemp display-7.
 02 ssu pic x occurs 240.
* WAS 120
*                    SPECIAL-VALS are used to convert & save accents & other special characters
01 special-vals.
 02 fil comp pic 9(5) value 66.
 02 fil comp pic 9(5) value 68.
 02 fil comp pic 9(5) value 70.
 02 fil comp pic 9(5) value 72.
 02 fil comp pic 9(5) value 74.
 02 fil comp pic 9(5) value 76.
 02 fil comp pic 9(5) value 98.
 02 fil comp pic 9(5) value 100.
 02 fil comp pic 9(5) value 102.
 02 fil comp pic 9(5) value 104.
 02 fil comp pic 9(5) value 106.
 02 fil comp pic 9(5) value 108.
 02 fil comp pic 9(5) value 112.
 02 fil comp pic 9(5) value 116.
01 specials redefines special-vals.
 02 s-c pic 9(5) comp occurs 14.
01 special-ch-vals.
 02 fil pic x(28) value "L O D THAEOEl o d thaeoei th" display-7.
01 special-chs redefines special-ch-vals.
 02 fil occurs 14 display-7.
  03 s-l1 pic x display-7.
  03 s-l2 pic x display-7.
77 skipped pic 99 value 0.
77 site pic 99 value 0.
77 site1 pic 99 value 0.
77 sitechar pic x.
*                       table of tags to be selected from MARC record (with indicators, subfields & required punc)
01 tag-values display-7.
 02 fil pic x(32) value "082 a # b$##c6".
 02 fil pic x(32) value "083 a. #".
 02 fil pic x(32) value "1000a # e # f, #".
 02 fil pic x(32) value "1001a # h2".
 02 fil pic x(32) value "1002a # h2".
 02 fil pic x(32) value "1003a # h2".
 02 fil pic x(32) value "110 a # c. #e (#".
 02 fil pic x(32) value "111 a # c. #e (#".
 02 fil pic x(32) value "240 a3#".
 02 fil pic x(32) value "245 a3  b : i; #j. #k = l. #".
 02 fil pic x(32) value "248 g1  h3  i; #j. #".
 02 fil pic x(32) value "250 a # e, #".
 02 fil pic x(32) value "260 b / c, #".
 02 fil pic x(32) value "440 a3".
 02 fil pic x(32) value "490 a #".
 02 fil pic x(32) value "505 a7".
 02 fil pic x(32) value "6000a # e # f, #".
 02 fil pic x(32) value "6001a # h2".
 02 fil pic x(32) value "6002a # h2".
 02 fil pic x(32) value "6003a # h2".
 02 fil pic x(32) value "610 a # c. #e (#".
 02 fil pic x(32) value "611 a # c. #e (#".
 02 fil pic x(32) value "650 a. #x--#y--#z--#".
 02 fil pic x(32) value "651 a. #x--#y--#z--#".
 02 fil pic x(32) value "7000a # e # f, #".
 02 fil pic x(32) value "7001a # h2".
 02 fil pic x(32) value "7002a # h2".
 02 fil pic x(32) value "7003a # h2".
 02 fil pic x(32) value "710 a # c. #e (#".
 02 fil pic x(32) value "711 a # c. #e (#".
 02 fil pic x(32) value "960 a # b # p#".
 02 fil pic x(32) value "961 a # b # p#".
 02 fil pic x(32) value "962 a # b # p#".
 02 fil pic x(32) value "964 a # b # p#".
 02 fil pic x(32) value "965 a # b # p#".
 02 fil pic x(32) value "966 a # b # p#".
 02 fil pic x(32) value "970 c # ".
01 tag-table redefines tag-values display-7.
 02 ttentry occurs 37.
  03 ttag pic xxx.
  03 tind pic x.
  03 tsf  pic x(28).
77 tag-occs pic 99 comp value 37.
77 total-chars pic 9(7) comp value 0.
*                                     used to assemble final o/p record
01 wanted-fields display-7.
 02 wfield occurs 50.
  03 wtag .
   04 wtagstem pic xx.
   04 wtagtail pic x.
   04 filler pic b.
  03 wi1  pic x.
  03 wi2  pic xb.
  03 wfld pic x(500).
77 wdelim pic XXX.
77 x pic 9(6) comp.
77 y pic 9(6) comp.
77 z pic 9(6) comp.
 
procedure division.
 
************
pack section.
************
 
* get run-time parameters
* open files
* skip required no. of blocks in input file
* then for each block:
*    read into EREC
*    translate each eight-bit character into ascii -> CURRENT-BLOCK
*    except replace accent with \
*    and save translated accent in corresponding char. position in BLOCK-ACCENTS
*    check segment-control-word (SCW) is numeric
*    for each segment:
*        U is char. pos. in i/p block
*        V is char. pos. in record
*        compute segment length
*        and move segment into record area (FREC)
*        if \ move accent to corresponding char. in RECORD-ACCENTS
*    if segment = whole record
*    or segment is last segment of record perform convert
*    then process next segment            ---------------
*    or if no more next block
 
     display "eight-bit filename: " with no advancing.
     accept eight-id.
     open input eight-bit.
     display "no. of blocks to skip: " with no advancing.
     accept skipno.
     display "proj filename stem (4 chs): " with no advancing.
     accept proj-stem.
     move proj-stem to diag-stem.
     display "max. no. of chars (per file): " with no advancing.
     accept max-chars.
     display "previous control no.? (return=start of file) " with no advancing.
     accept previous-fcno.
     if previous-char1 = "B" move "b" to previous-char1.
     if previous-fcno = space move 1 to ffirst display " ",
                              display proj-id, display "*********"
     else                     display "skipping " with no advancing.
     open output proj,tmp.   move 0 to bin.
     move bell-char to p300rec. write p300rec.
     perform read-eight-bit Skipno times.
next-block.
     perform read-eight-bit .
     ADD 1 to number-of-blocks.
     move space to block-accents.
     perform translate-char varying x from 1 by 1 until x > 2048.
    if first-scw not numeric perform error1 go to next-block.
    move first-scw to scw.
    move 6 to u.
next-segment.
    compute l = segment-length - 5.
    if v + l > 4095 perform error2 go to next-block.
    if u + l > 2049 perform error3 go to next-block.
    perform move-char l times.
    if whole-record or last-segment-of-record perform convert.
    if u > 2039 go to next-block.
    perform move-scw varying l from 1 by 1 until l > 5.
    if next-scw not numeric perform error4 go to next-block.
    move next-scw to scw.
    go to next-segment.
 
 
***************
convert section.
***************
 
* if record starts badly report this
* otherwise save "fbase", extract "fcno", perform process-record
*                                         ----------------------
    if frl not numeric perform error5 else
    if fbase not numeric perform error5 else
    if fstatus not = (" " and "c" and "d" and "n" ) perform error5 else
    if ftg(1) not = "001" perform error6
    else move fbase to x
         add 1 to x
         move space to fcno
         unstring fr  delimited by "#" into fcno pointer x
         perform process-record.
    move space to frec, record-accents.
    move 1 to v.
 
pack-and-convert-routines section.
*********************************
 
close-proj.
     add number-of-records to tot-no-of-recs.
     add 4 to total-chars.
     display total-chars," characters = " with no advancing.
     compute hex(1) = total-chars / 1048576.
     compute total-chars = total-chars - hex(1) * 1048576.
     compute hex(2) = total-chars / 65536.
     compute total-chars = total-chars - hex(2) * 65536.
     compute hex(3) = total-chars / 4096.
     compute total-chars = total-chars - hex(3) * 4096.
     compute hex(4) = total-chars / 256.
     compute total-chars = total-chars - 256 * hex(4).
     compute hex(5) = total-chars / 16.
     compute hex(6) = total-chars - hex(5) * 16.
     perform hex-conv varying x from 1 by 1 until x > 6.
     display chars with no advancing.
     display "     blocks read = ",number-of-blocks WITH NO ADVANCING.
     display "     last rec ",save-fcno WITH NO ADVANCING.
     display "     no. of recs ",number-of-records.
     move prec to record-save-area.
     move bell-char to p300rec. write p300rec.
     close proj .
 
eof.
     close tmp.
     display " ". display "*********".
     if v not = 1 display "last rec incomplete".
     display number-of-blocks, " blocks read".
     if pcj > 0 display pcj," pcj recs ignored".
     display tot-no-of-recs," records written in total".
     display pcsave," is longest".
     display save-fcno," was last control no.".
 
error1.
     display " ".
     display "first 5 chs of block not numeric, no.-of-blks=",number-of-blocks.
     display cg.
error2.
    display " ".
    display "ch.pos. in curr.rec.=",v," seg.len.=",l," no.-of-blks=",Number-of-blocks.
    display cg.
error3.
    display " ".
    display "ch.pos. in curr.blk.=",u," seg.len.=",l," no.-of-blks=",number-of-blocks.
    display cg.
error4.
    display " ".
    display next-scw," first 5 chs of seg not numeric, no.-of-blks=",number-of-blocks,"  ch.pos. in curr.blk=",u  .
error5.
    if not searching
      display " ",
      display "rec starts badly, no.-of-blks=",number-of-blocks,
      display fg.
error6.
      display " ". display "no 001 field ".  display fg.
 
move-char.
      move cbc(u) to f(v).
      if cbc(u) = "\" move b-acc(u) to r-acc(v).
      add 1 to u,v.
 
hex-conv.
     if hex(x) < 10 move hex(x) to char(x)
     else if hex(x) = 10 move "A" to char(x)
     else if hex(x) = 11 move "B" to char(x)
     else if hex(x) = 12 move "C" to char(x)
     else if hex(x) = 13 move "D" to char(x)
     else if hex(x) = 14 move "E" to char(x)
     else if hex(x) = 15 move "F" to char(x)
     else display "illegal hex ",hex(x) move "?" to char(x).
 
move-scw.
      move cbc(u) to next-scw-ch(l).
      add 1 to u.
 
reopen-proj.
     add 1 to pfileno.
     move pfileno to proj-no.
     open output proj.
     move bell-char to p300rec. write p300rec.
     move record-save-area to prec. move space to record-save-area.
     display " ". display proj-id. display "*********".
     move 0 to total-chars, number-of-records.
 
read-eight-bit.
     read eight-bit at end perform reAD-NEXT-FILE GO TO READ-EIGHT-BIT.
 
READ-NEXT-FILE.
     CLOSE EIGHT-BIT.
     DISPLAY " NEXT EIGHT-BIT FILENAME? " WITH NO ADVANCING.
     ACCEPT EIGHT-ID.
       IF EIGHT-ID = " "       perform close-proj perform eof stop run.
       OPEN INPUT EIGHT-BIT.
 
translate-char.
    move e(x) to e-ch.
    if bin < hex-5E perform t-copy-char else
    if bin < hex-61 perform t-backslash-and-save else
    if bin < hex-7B perform t-copy-char else
    if bin < hex-80 perform t-backslash-and-save else
    if bin = hex-E0 move hex-BB to bin
                    perform t-subtract-backslash-and-save else
    if bin = hex-FE move hex-BC to bin
                    perform t-subtract-backslash-and-save
    else            perform t-subtract-backslash-and-save.
t-copy-char.
    add bin to bin.
    move a-ch to cbc(x).
t-backslash-and-save.
    add bin to bin.
    move "\" to cbc(x).
    move a-ch to b-acc(x).
t-subtract-backslash-and-save.
    subtract hex-80 from bin.
    perform t-backslash-and-save.
 
 
**********************
process-record section.
**********************
 
    if searching perform skip-record go to next-record.
 
    if fcno3 = "pcj" add 1 to pcj go to next-record.
    move 0 to a,a240,a245,a505,D,DCed,a082-count,site,analyt.
    move 0 to date1,datecode,site1. move space to a082s,DCed.
*LOOK FOR 002, AND SEE IF RECORD IS ANALYTIC
    IF FTG(2) NOT = "002" MOVE 2 TO Y GO TO GET-LANG-AND-DATE.
    MOVE FDIRS(2) TO FDIR.
    COMPUTE Z = FDIS + FBASE + 1.
    ADD 7 TO Z.
    UNSTRING FREC INTO ADISP POINTER Z.
    COMPUTE ANALYT = (ADISP / 12) - 1.
    MOVE 3 TO Y.
 
GET-LANG-and-date.
    IF FTG(Y) NOT = "008" DISPLAY FCNO," ",FTG(Y)," CAN'T FIND 008" GO TO NEXT-RECORD.
    MOVE FDIRS(Y) TO FDIR.
    COMPUTE Z = FDIS + FBASE + 36.
    UNSTRING FREC INTO LANG POINTER Z.
    compute z = fdis + fbase + 8.
    unstring frec into date1 pointer z.
 
next-tag.
*--------
    add 1 to y.
    move fdirs(y) to fdir.
    if fhash = "#" go to end-of-record.
    move 1 to x,c.
 
check-tag.
*--------
    if ftag < ttag(x) go to NEXT-TAG.
    if ftag > ttag(x) add 1 to x, if x not > tag-occs go to check-tag
                                  else go to NEXT-TAG.
    compute z = fdis + fbase + 1.
    move f(z) to i1. move f(z + 1) to i2.
    if tind(x) = " " go to this-tag.
check-ind-1.
    if tind(x) = i1  go to this-tag.
    add 1 to x.
    if ftag = ttag(x) go to check-ind-1.
    display "unknown ind for this tag ",fcno," ",ftag," ",i1.
    go to NEXT-TAG.
this-tag.
     add 3 to z. move ttentry(x) to current-tag.
 
get-subfield.
*------------
    move 1 to cc.
us1. if f(z) = "$" or "#" move f(z) to wdelim add 1 to z go to us2.
     if f(z) = "\" perform accent
     else move f(z) to st(cc) add 1 to cc.
     add 1 to z.
        IF CC > 500 DISPLAY FTAG, "  >500 chs ",fcno   GO TO US2.
     go to us1.
us2. if cc = 1 display ftag," ",fcno," empty sf" go to next-subfield.
     move "#" to st(cc).
 
    move 1 to b.
 
check-subfield.
*-------------
     if sf not = csf(b) add 1 to b if csf(b) not = " " go to check-subfield
                                   else                go to next-subfield.
     move cpunc(b) to punc.
     if punc1 not numeric perform save-subfield
     else if may-have-article  if i2 = 0 if ftag = "248" move ": ^" to punc perform save-subfield
                                         else            move "#" to punc perform save-subfield
                               else      perform dont-index-article
     else if subfield-is-100h  perform initials thru init-end
                               perform save-subfield
     else if subfield-is-248g  perform dont-index-248g
     else if sf-is-082c        perform look-for-DC-edition
     else if tag-is-505        move 1 to a505 perform save-subfield
     else                      display "unknown punc ",punc," ",fcno go to next-record.
 
next-subfield.
*------------
     if wdelim = "$" go to GET-SUBfield.
     if gfield not = space add 1 to a
                           move ftag to wtag(a)
                           move i1 to wi1(a)
                           move i2 to wi2(a)
                           move "#" to g(c)
                           move gfield   to wfld(a)
                           move space to gfield
                           move 1 to c.
    go to NEXT-TAG.
 
end-of-record.
*************   now assemble fields and output them
    move 1 to pc.
    add 1 to a.
    move "999" to wtag(a).
*FIELD (1) AUTHOR
    perform get-1xx varying a from 1 by 1 until wtag(a) > "111".
    perform finish-field.
 
*FIELD (2) TITLE   (use 240 in preference to 245.0 only)
    perform find-240-245 varying a from 1 by 1 until wtag(a) > "245".
    if a245 = 0 MOVE "no 245" TO MESS PERFORM DIAG write drec
                move space to drec go to next-record .
    if a240 > 0 if wi1(a245) = "0" move " " to wtag(a245)
                else               move " " to wtag(a240).
    perform get-245-or-240 varying a from 1 by 1 until wtag(a) > "245".
    perform finish-field.
 
*FIELD (3)  MAIN ED-PUB
    perform get-250-260 varying a from 1 by 1 until wtag(a) > "260".
    perform finish-field.
 
*FIELD (4) SERIES, PART TITLES, ETC
    perform get-440-490 varying a from 1 by 1 until wtag(a) > "490".
    perform get-248-etc varying a from 1 by 1 until wtag(a) = "999".
    PERFORM GET-LEFTOVERS VARYING A FROM 1 BY 1 UNTIL WTAG(A) = "999".
    perform finish-field.
 
*FIELD (5) ADDED NAMES
    perform get-7XX varying a from 1 by 1 until wtag(a) = "999".
    perform finish-field.
*FIELD (6) DEWEY(S) and SPINE LABEL
    perform get-first-082 varying a from 1 by 1 until wtag(a) > "082" or a082-count = 1.
    perform get-970 varying a from 1 by 1 until wtag(a) > "970".
if a082-count = 0 MOVE "no 082" to mess perform diag go to next-record
   else perform get-other-082s varying a from 1 by 1 until wtag(a) > "082".
    perform finish-field.
 
*FIELD (7) ACCESSION NOS.
    MOVE 0 TO B.
    perform get-96x varying a from 1 by 1 until wtag(a) > "969".
    perform finish-field.
 
*FIELD (8)  SITE code, DATE code, DC edition (from 082$c), LANG CODE , CONTROL NO.
    if f9 = space move "#" to f9.
     compute site = site + site1.
    perform dte.
    string site,datecode,lang,DCed,fcno delimited by "#"  into prec pointer pc.
    if f9 = "#" move " " to f9.
    perform finish-field.
 
*FIELD (9) SUBJECT HEADINGS         (x=no. of subjects, y=pointer, z=subscript for gsubjs)
    move 0 to x.
    perform subj-split varying a from 1 by 1 until wtag(a) > "651".
    perform subj-083 varying a from 1 by 1 until wtag(a) > "083".
    perform subj-get varying a from 1 by 1 until a > x.
    perform finish-field.
*FIELD (10) CONTENTS NOTE (non-analytics)
*     if analyt = 0 perform get-505 varying a from 1 by 1 until wtag(a) = "999".
*     perform finish-field.
*check what`s left
    MOVE 0 TO X.
    perform CHECK-WHAT-IS-LEFT varying a from 1 by 1 until wtag(a) = "999".
    IF X > 0 DISPLAY FCNO," ANALYTIC" WITH NO ADVANCING
           IF A505 > 0 DISPLAY " WITH 505"
           ELSE        DISPLAY " ".
    IF D > 0 WRITE DREC MOVE SPACE TO DREC.
 
*write record (starting new o/p file when necessary)
    add 1 to pc.
    if pc > pcsave move pc to pcsave.
    if pc + total-chars > max-chars perform close-proj, perform reopen-proj.
    if p600 = space write p300rec else
    if p900 = space write p600rec else
    write prec. move " " to prec.
    add 1 to number-of-records.
    add pc to total-chars.
    move fcno to save-fcno.
next-record.
    exit.
 
 
processing-routines section.
****************************
 
check-082.
    if wfld(a) = a082(x) move " " to wtag(a).
 
CHECK-WHAT-IS-LEFT.
    IF WTAG(A) NOT = " "
       IF (WTAG(A) = "100" OR "110" OR "111" OR "240" OR "245") AND (ANALYT > 0)
             MOVE 1 TO X
       ELSE PERFORM DIAG-TAG.
 
dte .
    if date1 < 1750   display "date < 1750 ",fcno else
    if date1 > next-year display " date > ",next-year," ",fcno else
    compute datecode = date1 - 1750.
 
DIAG.
    MOVE FCNO TO DCNO. ADD 1 TO D. MOVE MESS TO DMESS(D).
 
DIAG-TAG.
    MOVE FCNO TO DCNO. ADD 1 TO D. MOVE WTAG(A) TO DTAG(D). MOVE WI1(A) TO DI1(D). MOVE WI2(A) TO DI2(D).
 
dont-index-248g.
      string "_",scgroup delimited by "#" into gfield pointer c.
 
dont-index-article.
*put "don`t index" markers round article at start of
*240, 245, 248 (only used where no $g), 440, 490
      if i2 = 1 string "_",sa1,"^",sf1 delimited by "#" into gfield pointer c else
      if i2 = 2 string "_",sa2,"^",sf2 delimited by "#" into gfield pointer c else
      if i2 = 3 string "_",sa3,"^",sf3 delimited by "#" into gfield pointer c else
      if i2 = 4 string "_",sa4,"^",sf4 delimited by "#" into gfield pointer c else
      if i2 = 5 string "_",sa5,"^",sf5 delimited by "#" into gfield pointer c else
                display fcno," ",FTAG," ",I1,I2,"  indicator 2 > 5" .
      if c > 500 display "too long at dont-index-article ",fcno ," ",FTAG.
 
find-240-245.
     if wtag(a) = "240" move a to a240, else
     if wtag(a) = "245" move a to a245.
 
finish-field.
    if pc = 1 move "#" to p(1) move 2 to pc
    else if p(pc - 1) = " " move "#" to p(pc - 1)
         else move "#" to p(pc) add 1 to pc.
 
get-1xx.
     if wtag(a) = "100" if wi1(a) = "0" move "$" to mark perform string-wfld
                        else            move "|" to mark perform string-wfld
     else if wtag(a) = "110" or "111"   move "$" to mark perform string-wfld.
 
get-245-or-240.
         if wtag(a) = "245" or "240" move "^" to mark perform string-dot-wfld .
 
get-440-490.
    if wtag(a) not = " "
         if (wtag(a) = "440" and wi1(a) = "0")
         or (wtag(a) = "490" and wi2(a) = "0") move "^" to mark perform string-dot-wfld
 
get-250-260.
    if wtag(a) not = " "
         if (wtag(a) = "250" or "260") and wi2(a) = "0" move "@" to mark perform string-dot-wfld .
 
get-248-etc.
         if wtag(a) = "248" move wi1(a) to ind1
                            move "^" to mark perform string-dot-wfld
                            move a to b
                            perform get-248-remainder varying a from b by 1 until wtag(a) = "999"
                            move b to a.
 
get-248-remainder.
    if wtag(a) not = " "
         if (wtag(a) = "440" and wi1(a) = ind1)
         or (wtag(a) = "490" and wi2(a) = ind1) move "^" to mark perform string-dot-wfld
         else
         if (wtag(a) = "250" or "260") and wi2(a) = ind1 move "@" to mark perform string-dot-wfld .
 
get-7xx.
    if wtag(a) not = " "
         if wtag(a) = "700" if wi1(a) = "0" move "$" to mark perform string-wfld
                            else            move "|" to mark perform string-wfld
         else if wtag(a) = "710" or "711"   move "$" to mark perform string-wfld .
 
 
get-first-082.
     if wtag(a) = "082" string wfld(a) delimited by "#" into prec pointer pc
                        move wfld(a) to a082(1) move 1 to a082-count
                        move " " to wtag(a).
 
get-970.
     if wtag(a) = "970"  string " ",wfld(a) delimited by "#" into prec pointer pc
                        move " " to wtag(a).
 
get-505.
     if wtag(a) = "505" move "^" to mark perform string-dot-wfld.
 
get-other-082s.
*                         ignore duplicate deweys (for PCL)
    if wtag(a) = "082" perform check-082 varying x from 1 by 1 until wtag(a) = " " or x > a082-count
                       if wtag(a) = "082" add 1 to a082-count
                                          move wfld(a) to a082(a082-count)
                                          string " ",wfld(a) delimited by "#" into prec pointer pc
                                          move " " to wtag(a).
get-96x.
     if wtagstem(a) = "96" perform get-site-code
          string wfld(a)," " delimited by "#" into prec pointer pc
          move " " to wtag(a) .
 
 
GET-LEFTOVERS.
    IF WTAG(A) NOT = " "
       IF WTAG(A) = "250"           PERFORM DIAG-TAG MOVe "@" TO MARK PERFORM STRING-DOT-WFLD ELSE
       IF WTAG(A) = "440" OR "490"  PERFORM DIAG-TAG MOVe "^" TO MARK PERFORM STRING-DOT-WFLD ELSE
       IF WTAG(A) = "505" and analyt > 0             MOVE "^" TO MARK PERFORM STRING-DOT-WFLD .
 
get-site-code .
    move wtagtail(a) to sitechar.
      if sitechar = "0"  add 1 to site else
      if sitechar = "1"  move 2 to site1 else
      if sitechar = "2" add 4 to site else
      if sitechar = "4" add 8 to site else
      if sitechar = "5"  add 16 to site else
      if sitechar = "6"  move 2 to site1.
 
initials.
*Take initials only from 100,600,700 $h
*keep hyphens (-)
*drop anything beginning with a lower case letter (de,von, al-  etc)
      move 1 to b. move 1 to cc.
init-next.
      unstring scgroup delimited by " " or ". " or ".#" or "." or "-" or "#"
      into initi delimiter in wdelim pointer b.
init-chk.
     if init1 = "\" IF INIT3 = "\" MOVE INIT5 TO INIT1 GO TO INIT-CHK
                    ELSE           move init3 to init1 go to init-chk.
     if init1 > "Z" or init1 = " " if wdelim = "#" or ".#" go to init-end
                                  else go to init-next.
      move init1 to sc(cc). add 1 to cc.
      if wdelim = "-" move "-" to sc(cc) else move " " to sc(cc).
      if wdelim  = "#" or ".#" go to init-end.
      add 1 to cc go to init-next.
init-end.
      move "#" to sc(cc).
      move " #" to punc.
 
look-for-dc-edition.
    move sa2 to DCed.
 
save-subfield.
    if gfield = space and punc2 not = "("  move "#" to punc.
    string punc,scgroup delimited by "#" into gfield pointer c.
    if c > 500 display "too long at save-subfield ",fcno," ",ftag go to next-subfield.
    if punc2 = "(" move ")" to g(c) add 1 to c .
 
skip-record.
  add 1 to skipped.
  display fcno," " with no advancing .
  if skipped > 50 perform close-proj stop run.
  if fcno = previous-fcno move 1 to ffirst display " " ,display proj-id, display "*********".
 
string-wfld.
    string mark, wfld(a) delimited by "#" into prec pointer pc.
    move " " to wtag(a).
 
subj-split.
     if wtag(a) = "650" or "651"    move "^" to mark perform subj
     else if wtag(a) = "600" if wi1(a) = "0" move "$" to mark perform subj
                             else            move "|" to mark perform subj
          else if wtag(a) = "610" or "611"   move "$" to mark perform subj.
 
subj-083.
    if wtag(a) = "083" move "^" to mark perform subj.
 
subj-get.
    move rpunc(a) to punc.
    if punc1 = "-" move " - " to punc
                   move "#" to rmark(a) else
    if punc1 = "," move ", #" to punc else
                   move ". #" to punc.
    string rmark(a), rsubjct(a), punc delimited by "#"
    into prec pointer pc.
 
 
accent section.
*-------------
    move r-acc(z) to a-ch.
    move '\' to st(cc).
    move a-ch to st(cc + 1).
    add 2 to cc.
    move 1 to b.
acc1. if bin = s-c(b) go to acc2.
      add 1 to b.
      if b < 15 go to acc1.
      go to acc3.
acc2. move s-l1(b) to st(cc).
      add 1 to cc.
      if s-l2(b) not = " " move s-l2(b) to st(cc) add 1 to cc.
acc3. exit.
 
 
 
string-dot-wfld section.
*----------------------
        string mark, wfld(a) delimited by "#" into prec pointer pc.
        move " " to wtag(a).
strf1. if p(pc - 1) = " " subtract 1 from pc go to strf1.
     if p(pc - 1)  = "." or "," or "!" or "?" or "]" or ":"go to strf2.
      move "." to p(pc) add 1 to pc.
strf2.     move " " to p(pc). add 1 to pc.
 
subj section.
*-----------
    move 1 to y.
subj1. unstring wfld(a)  delimited by ". " or ".#" or ",#" or  "#"              or "--" or ",. " into ssubjct
    delimiter in wdelim count in cc pointer y.
**      display ssubjct.
    if cc > 119 display "cc>119 (subj)",fcno MOVE 119 TO CC.
    move "#" to ssu(cc + 1).
*now save subject (but for 083 check if unique first)
    if wtag(a) not = "083" go to subj3.
    move 1 to z.
subj2. if z > x go to subj3.
    if ssubjct = rsubjct(z) go to subj4.
    add 1 to z.
    go to subj2.
subj3. add 1 to x.
       move mark to rmark(x).
       move ssubjct to rsubjct(x).
       move wdelim to rpunc(x).
subj4.
    if wdelim = "#" or ".#" or ",#" move " " to wtag(a)
    else  go to subj1.
 

.
