/****************************************************************************

  extract_from_docs.c  SW June 00

  Based on search_and_extract.c and rtgterms_pass1.c

  Query terms on stdin and a file of reldocs

  args: <Rfile> (= file of DOCNOs of relevant or pseudo-relevant docs)

  input: <term>:[<term>:]<source>:<qtf>

  output term:[<term>:]N:n:R:r:cttf:tf:Rttf:rtf:qtf:qlen:src:wt:rscore:rsv

  Many environment variables needed.

  May 01: to be modified so "relevant" docs may be given either by
  DOCNO (contents of field 1) or by IRN (Internal Record Number).

  NB Modification not complete

******************************************************************************/

#undef DEBUG
/* #define DEBUG 1 */
#include "bss.h"
#include "bss_protos.h"
#include "expd.h"

extern struct db Cdb ;

char *info = 
"A query is given on stdin in the form <term>:<source>:<qtf>
                                    or <term1>:<term2>:<source>:<qtf>
and the first command line arg is the name of a file of document numbers
of relevant documents.

Terms are extracted from the relevant docs and these and query terms are
given scores using both the Robertson offer weight (rsv) and an
'absolute' score (rscore).

   Input:  qterm1:[qterm2:]source:qtf

  Output:  term:[term2:]N:n:R:r:cttf:tf:Rttf:rtf:qtf:qlen:source:weight:rscore:rsv

In the input qtf must be nonzero.

Environment: BSS_DB is mandatory; BSS_ATTRIBUTE, BSS_DN_ATTRIBUTE, BSS_LIMIT, 
BSS_BIGRLOAD, BSS_RLOAD, BSS_K3, BSS_MEM, EXPD_MINn, EXPD_MINr, EXPD_MINrtf, 
EXPD_NOI, EXPD_QUIET are optional, overriding defaults (see expd.h).
Note that MINn, MINr, MINrtf and NOI don't apply to query terms.
EXPD_WANTED_FIELDS is a colon-separated list of field numbers for extraction
-- the default is to use the last field.
EXPD_QT_MULT (default 1) multiplies the weight of query terms.
EXPD_PREFERRED_FIELD_MULT (default 1) multiplies the weight of query terms
from a single specified EXPD_PREFERRED_FIELD (default tit) of the query source.
EXPD_MAX_DOCLEN (default 0). If > 0 documents of greater length are not used
as a source of terms.
If EXPD_QUIET is non-zero diagnostic messages are off (default on).
This program uses a script ttstat which must be in the current search path.
See also search_and_extract.

EXPD_SCORECMD may also be set (a BSS command). The command defaults to
rscore, but any other command taking the same args may be used.

NB (May 01): although pairs of query terms are accepted, and there is a second,
empty, term in each output line, so long as the environment variable 
EXPD_PAIRS is nonzero, the second term is ignored. If EXPD_PAIRS is NULL or 
zero only single terms are accepted in input, and output (i.e. compatible 
with previous versions).
" ;


int
main(int argc, char **argv)
{
  char *dbname ;
  int num_wf ;
  int wf[MAXNUMFIELDS + 1] ;
  int R ;			/* Number obtained */
  int Rtf = 0 ;			/* Total # tokens in relset */
  int numaddedterms = 0 ;	/* No of non-queryterms from extract */
  int numqterms = 0 ;
  int numterms = 0 ;
  int qlen = 0 ;
  int N ;
  int cttf ;
  char *Command, *Response ;
  char buf[1024], term[1024], term2[1024] ;
  FILE *relfile ;
  FILE *tmpf;
  char tmpfname[256], tmpf1name[256] ;
  BOOL found ;
  int set ;
  int j, k ;
  struct term *terms ;
  int res ;
  char *p ;
  int num_fields ;
  int chars_read ;
  char *envp ;
  char *attr = DEFAULT_ATTRIBUTE ;
  char *dn_attr = DEFAULT_DN_ATTRIBUTE ;
  double k3 = DEFAULT_K3 ;
  int limit = DEFAULT_LIMIT ;
  int Rload = DEFAULT_BIGRLOAD ;
  int rload = DEFAULT_RLOAD ;
  double qt_mult = DEFAULT_QT_MULT ;
  double preferred_field_mult = DEFAULT_PREFERRED_FIELD_MULT ;
  char *preferred_field = DEFAULT_PREFERRED_FIELD ;
  int doclen ;
  int max_doclen = DEFAULT_MAX_DOCLEN ;
  int noI = DEFAULT_NOI ;
  int minn = DEFAULT_MINn, minr = DEFAULT_MINr, minrtf = DEFAULT_MINrtf ;
  int skippedIs = 0, rtoosmall = 0, rtftoosmall = 0, ntoosmall = 0 ;
  int docstoolong = 0 ;
  char *scorecmd = DEFAULT_SCORECMD ;
  int quiet = DEFAULT_QUIET ;
  int pairs = DEFAULT_PAIRS ;

  terms = (struct term *)calloc(MAXTERMS, sizeof(struct term)) ;
  Command = (char *)malloc(LARGE_BUFSIZE) ;
  Response = (char *)malloc(LARGE_BUFSIZE) ;
  if ( ! terms || ! Command || ! Response ) {
    fprintf(stderr, "%s: out of memory\n", argv[0]) ;
    exit(1) ;
  }
  if ( argc < 2 || strcmp(argv[1], "-h") == 0 ) {
    fprintf(stderr, "Usage: %s <relfile>\n", argv[0]) ;
    fprintf(stderr, info) ;
    exit(1) ;
  }
  limit = 0 ;
  relfile = fopen(argv[1], "rb") ;
  if ( relfile == NULL ) {
    fprintf(stderr, "Can't open relfile %s\n", argv[1]) ;
    exit(1) ;
  }

  if ( (envp = getenv("BSS_DB")) && *envp ) dbname = envp ;
  else {
    fprintf(stderr, "No database (BSS_DB) in environment\n") ;
    exit(1) ;
  }
  if ( (envp = getenv("EXPD_SCORECMD"))  && *envp ) scorecmd = envp ;
  if ( (envp = getenv("EXPD_QUIET"))  && *envp ) quiet = atoi(envp) ;
  if ( (envp = getenv("BSS_LIMIT"))  && *envp ) limit = atoi(envp) ;
  if ( (envp = getenv("BSS_ATTRIBUTE")) && *envp ) attr = envp ;
  if ( (envp = getenv("BSS_DN_ATTRIBUTE")) && *envp ) dn_attr = envp ;
  if ( (envp = getenv("BSS_K3")) && *envp ) k3 = atof(envp) ;
  if ( (envp = getenv("BSS_BIGRLOAD")) && *envp ) Rload = atoi(envp) ;
  if ( (envp = getenv("BSS_RLOAD")) && *envp ) rload = atoi(envp) ;
  if ( (envp = getenv("EXPD_MINn")) && *envp ) minn = atoi(envp) ;
  if ( (envp = getenv("EXPD_MINr")) && *envp ) minr = atoi(envp) ;
  if ( (envp = getenv("EXPD_MINrtf")) && *envp ) minrtf = atoi(envp) ;
  if ( (envp = getenv("EXPD_NOI")) && *envp ) noI = atoi(envp) ;
  if ( (envp = getenv("EXPD_QT_MULT")) && *envp ) qt_mult = atof(envp) ;
  if ( (envp = getenv("EXPD_PREFERRED_FIELD")) && *envp )
    preferred_field = envp ;
  if ( (envp = getenv("EXPD_PREFERRED_FIELD_MULT")) && *envp )
    preferred_field_mult = atof(envp) ;
  if ( (envp = getenv("EXPD_PAIRS")) && *envp ) pairs = atoi(envp) ;
  if ( (envp = getenv("EXPD_MAX_DOCLEN")) && *envp ) max_doclen = atoi(envp) ;

  iinit() ;
  sprintf(Command, "choose %s\n", dbname) ;
  I0 ;
  if ( ! quiet ) fprintf(stderr, "Using database %s\n", dbname) ;
  sprintf(Command, "info db\n") ;
  I0 ;
  num_fields = atoi(strstr(Response, "nfields=") + 8) ;
  num_wf = 1 ;
  if ( (envp = getenv("EXPD_WANTED_FIELDS")) && *envp ) {
    char *p ;
    wf[0] = atoi(strtok(envp, ":")) ;
    for ( j = 1 ; j < num_fields ; j++ ) {
      p = strtok(NULL, ":") ;
      if ( p == NULL ) break ;
      wf[j] = atoi(p) ;
      num_wf++ ;
    }
  }
  else wf[0] = num_fields ;	/* Assume last field */
  sprintf(Command, "set attr=%s\n", attr) ;
  I0 ;
  if ( ! quiet ) {
    fprintf(stderr, "Using index attribute %s\n", attr) ;
    fprintf(stderr, "Using docnum attribute %s\n", dn_attr) ;
  }
  if ( (envp = getenv("BSS_MEM")) && *envp ) {
    sprintf(Command, "set memory=%d\n", atoi(envp)) ;
    I0 ;
  }
  if ( (envp = getenv("BSS_DEBUG")) && *envp ) {
    sprintf(Command, "set debug=%d\n", atoi(envp)) ;
    I0 ;
  }
  if ( limit > 0 ) {
    sprintf(Command, "set limit=%d\n", limit) ;
    I0 ;
    if ( ! quiet ) fprintf(stderr, "Using limit %d\n", limit) ;
  }
  sprintf(Command, "d bign\n") ;
  I0 ;
  N = atoi(Response) ;
  sprintf(Command, "d cttf\n") ;
  I0 ;
  cttf = atoi(Response) ;
  if ( ! quiet ) {
    fprintf(stderr, "Using N %d\n", N) ;
    fprintf(stderr, "Using Rload %d\n", Rload) ;
    fprintf(stderr, "Using rload %d\n", rload) ;
    fprintf(stderr, "Using cttf %d\n", cttf) ;
    fprintf(stderr, "Using k3 %.2f\n", k3) ;
    fprintf(stderr, "Using min n %d\n", minn) ;
    fprintf(stderr, "Using min r %d\n", minr) ;
    fprintf(stderr, "Using min rtf %d\n", minrtf) ;
    fprintf(stderr, "Using no I %d\n", noI) ;
    fprintf(stderr, "Using topic term weight multiplier %.1f\n", qt_mult) ;
    fprintf(stderr, "Using topic preferred field %s\n", preferred_field) ;
    fprintf(stderr, "Using preferred field wt mult %.1f\n",
	    preferred_field_mult) ;
    fprintf(stderr, "Using score command %s\n", scorecmd) ;
    if ( pairs ) fprintf(stderr,
	 "Accepting 4-field input and outputting 2nd (empty) term\n") ;
    if ( max_doclen ) fprintf(stderr, "Max doclength %d\n", max_doclen) ;
  }

  /* Form query termsets */
  if ( Debug ) fprintf(stderr, "Looking up query terms\n") ;
  while ( ! feof(stdin) ) {
    if ( fgets(buf, 1023, stdin) == NULL ) break ;
    if ( numterms >= MAXTERMS ) {
      fprintf(stderr, "Too many terms\n") ;
      goto error ;
    }
    if ( ! pairs )
      sscanf(buf, "%[^:]:%[^:]:%d",
	   term, terms[numterms].source, &terms[numterms].qtf) ;
    else {
      /* %[ doesn't work with empty strings 
      sscanf(buf, "%[^:]:%[^:]:%[^:]:%d",
	     term, term2, terms[numterms].source, &terms[numterms].qtf) ;
      */
      char *p = buf ;
      int l ;
      l = strdcpy(term, p, ':') ;
      p += l + 1 ;
      l = strdcpy(term2, p, ':') ;
      p += l + 1 ;
      l = strdcpy(terms[numterms].source, p, ':') ;
      p += l + 1 ;
      terms[numterms].qtf = atoi(p) ;
    }
    qlen += terms[numterms].qtf ;
    sprintf(Command, "f save=0 t=%s", term) ;
    I0 ;
    terms[numterms].n = atoi(strstr(Response, "np=") + 3) ;
    if ( terms[numterms].n == 0 && rload == 0 ) {
      fprintf(stderr, "Warning: term %s not found\n", term) ;
      /* continue ; */
    }
    terms[numterms].set = -1 ;
    sprintf(Command, "d ttf") ;
    I0 ;
    terms[numterms].ttf = atoi(Response) ;
    
    terms[numterms].term= strdup(term) ;
    if ( pairs ) {
      terms[numterms].term2 = strdup(term2) ;
      if ( strlen(term2) ) continue ; /* Skipping genuine pairs at present */
    }
    numterms++ ;
    numqterms++ ;
  }
  if ( ! quiet ) {
    fprintf(stderr, "Got %d query term", numqterms) ;
    if ( numqterms == 1 ) fprintf(stderr, "\n") ;
    else fprintf(stderr, "s\n") ;
    fprintf(stderr, "Reading relevant docnos\n") ;
  }
  R = 0 ;
  sprintf(tmpf1name, "/tmp/tmpf1.%d", (int)getpid()) ;
  sprintf(tmpfname, "/tmp/tmpf.%d", (int)getpid()) ;
  tmpf = fopen(tmpfname, "w+b") ;
  while ( fgets(buf, 1023, relfile) ) {
    int termsthis, nk ;
    if ( feof(relfile) ) break ;
    /* Doesn't cope with trailing \n which fgets puts in*/
    *(buf + strlen(buf) - 1) = '\0' ; 
    sprintf(Command, "f a=%s t=%s", dn_attr, buf) ;
    I0 ;
    set = atoi(Response + 1) ;
    if ( atoi(strstr(Response, "np=") + 3) == 0 ) {
      fprintf(stderr, "%s not found\n", buf) ;
      sprintf(Command, "delete %d", set) ;
      I0 ;
      continue ;
    }
    sprintf(Command, "doclen r=0") ;
    I0 ;
    if ( max_doclen ) {
      doclen = atoi(Response) ;
      if ( doclen > max_doclen ) {
	docstoolong++ ;
	if ( ! quiet ) fprintf(stderr, "Skipping doc of length %d\n", doclen) ;
	continue ;
      }
    }
    for ( k = 0 ; k < num_wf ; k++ ) {
      sprintf(Command, "extract s=%d r=0 sources=0 field=%d", set, wf[k]) ;
      I0 ;
      nk = atoi(Response) ;
      p = Response ;
      sscanf(p, "%*[^\n]\n%n", &chars_read) ;
      p += chars_read ;
      termsthis = 0 ;
      while ( nk-- ) {
	sscanf(p, "%[^\n]\n%n", buf, &chars_read) ;
	p += chars_read ;
	fprintf(tmpf, "%s:%d\n", buf, R) ; /* First char is GSL code */
	termsthis++ ;
      }
      if ( Debug ) 
	fprintf(stderr, "Extracted %d terms from field %d of doc %d\n",
		termsthis, wf[k], R) ;
      I0 ;
    }
    R++ ;
    sprintf(Command, "delete %d\n", set) ;
  }
  fclose(tmpf) ;
  if ( ! quiet ) {
    fprintf(stderr, "Read and extracted from %d reldocs\n", R) ;
    fprintf(stderr, "Sorting and collating terms\n") ;
  }
  sprintf(Command, "ttstat < %s > %s\n", tmpfname, tmpf1name) ;
  system(Command) ;
  /* Now have tmpf1 containing all the terms uniquely, each followed by
     Rtf, tf, r. */
  
  {
    char buf[256] ;
    char p[256] ;
    char code ;
    int rtf ;
    int r ;
    int n ;
    int termnum ;
    char response[1024], command[1024] ;
    tmpf = fopen(tmpf1name, "rb") ;
    while ( ! feof(tmpf) ) {
      if ( numterms >= MAXTERMS ) {
	fprintf(stderr, "Too many terms\n") ;
	break ;
      }
      fgets(buf, 256, tmpf) ;
      if ( feof(tmpf) ) break ;
      sscanf(buf, "%c%[^:]:%*d:%d:%d", &code, p, &rtf, &r) ;
      for ( k = 0, found = FALSE ; k < numqterms ; k++ ) {
	if ( strcmp(p, terms[k].term) == 0 ) {
	  found = TRUE ;
	  if ( ! quiet ) fprintf(stderr, "Found %s\n", p) ;
	  break ;
	}
      }
	
      if ( found ) termnum = k ;
      else {			/* Add it unless I-term and noI */
	if ( code == 'I' && noI ) {
	  if ( Debug >= 1 ) fprintf(stderr, "Skipping I-term %s\n", p) ;
	  skippedIs++ ;
	  continue ;
	}
	if ( r < minr ) {
	  rtoosmall++ ;
	  continue ;
	}
	if ( rtf < minrtf ) {
	  rtftoosmall++ ;
	  continue ;
	}
	sprintf(command, "f save=0 t=%s", p) ;
	I0_ ;
	n = atoi(response + 3) ;
	if ( n < minn ) {
	  ntoosmall++ ;
	  continue ;
	}
	termnum = numterms ;
	terms[termnum].term = strdup(p) ;
	terms[termnum].term2 = "" ;
	terms[termnum].set = -1 ;
	terms[termnum].n = n ;
	sprintf(command, "d ttf") ;
	I0_ ;
	terms[termnum].ttf = atoi(response) ;
	terms[termnum].qtf = 0 ;
	sprintf(terms[termnum].source, "doc") ;
	numterms++ ;
	numaddedterms++ ;
	if ( Debug >= 1 ) fprintf(stderr, "Added %s, term %d\n", p, numterms) ;
      }
      terms[termnum].r = r ;
      terms[termnum].rtf = rtf ;
      Rtf += rtf ;

    } /* (while !feof(tmpf)) */
  }

  /* Output section */

  for ( j = 0 ; j < numterms; j++ ) {
    if ( terms[j].qtf > 0 ) {
      /* Aug 01: after discussion with SER set n to 1 if really 0
	 Previously incremented n by rload for any n. */
      int real_n = terms[j].n ;
      if ( real_n == 0 ) real_n = 1 ;
      sprintf(Command,
 "weight n=%d wt_fn=2 rload=%d bigrload=%d r=%d bigr=%d k3=%.3f qtf=%d\n",
	      real_n, rload, Rload, terms[j].r, R,
	      k3, terms[j].qtf) ;
      I0 ;
      terms[j].wt = atof(Response) * qt_mult ;
      if ( strcmp(terms[j].source, preferred_field) == 0 )
	terms[j].wt *= preferred_field_mult ;
      if ( R + Rload == 0 ) terms[j].rsv = 0 ;
      else terms[j].rsv = terms[j].wt * (terms[j].r + rload) / (R + Rload);
      sprintf(Command, "%s n=%d r=%d bigr=%d",
	      scorecmd, real_n, terms[j].r + rload, R + Rload) ;
      I0 ;
      terms[j].rscore = atof(Response) ;
    }
    else {
      sprintf(Command, "weight n=%d wt_fn=1 r=%d bigr=%d\n",
	      terms[j].n, terms[j].r, R) ;
      I0 ;
      terms[j].wt = atof(Response) ;
      terms[j].rsv = ( R == 0 ) ? 0 : terms[j].wt * terms[j].r / R;
      sprintf(Command, "%s n=%d r=%d bigr=%d",
	      scorecmd, terms[j].n, terms[j].r, R) ;
      I0 ;
      terms[j].rscore = atof(Response) ;
    }
    /* term:N:n:R:r:cttf:tf:Rttf:rtf:qtf:qln:src:rscore:rsv */

    if ( ! pairs ) 
      printf("%s:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%s:%.3f:%.3f:%.3f\n", 
	     terms[j].term, N, terms[j].n, R, terms[j].r,
	     cttf, terms[j].ttf, Rtf, terms[j].rtf, terms[j].qtf, qlen,
	     terms[j].source, terms[j].wt, terms[j].rscore, terms[j].rsv) ;
    else
      printf("%s:%s:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%s:%.3f:%.3f:%.3f\n", 
	     terms[j].term, terms[j].term2, N, terms[j].n, R, terms[j].r,
	     cttf, terms[j].ttf, Rtf, terms[j].rtf, terms[j].qtf, qlen,
	     terms[j].source, terms[j].wt, terms[j].rscore, terms[j].rsv) ;
  }

  if ( ! quiet ) {
    fprintf(stderr,
	    "%d docs too long, %d qterms %d added %d total %d skippedI %d skippedr %d skippedrtf %d skippedn\n",
	    docstoolong, numqterms, numaddedterms, numterms,
	    skippedIs, rtoosmall, rtftoosmall, ntoosmall) ;
  }

 error:
     
  unlink(tmpfname) ; unlink(tmpf1name) ;
  iexit() ;
  exit(1) ;
}

