/******************************************************************************

  extract_nth  SW May 97

  Extracts every nth b-byte-long record from its input

  Intended for making odd/even etc databases by frigging dir files

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
char *usage="\
Usage: extract_nth -l|-reclen <reclen> [-n|-step <step>] \\\n\
[-s|-skip <recs to skip at start>] [-m|-max <max recs to do>] [-v|-verbose]\n\
       extract_nth -h|-help\n\
step and length must be greater than 0 \n\
-max 0 does all (default)  \n\
Extracts records of length <reclen> from its input subject to given step\n\
(default 1), skip (default 0) and number to do (default all).\n\
" ;

main(argc, argv)
     int argc ;
     char **argv ;
{
  int i, n = 0, l = 0, s = 0 ;
  int count = 0, inrecs = 0 ; 
  int skipped = 0 ;
  int verbose = 0 ;
  int m = 0 ;
  int nread ;
  char *buf ;

  if ( argc < 3 ) goto USAGE ;
  for ( i = 1 ; i <= argc - 1 ; i++ ) {
    if ( strcmp(argv[i], "-n") == 0 || strcmp(argv[i], "-step") == 0 ) {
      if ( i >= argc - 1 ) goto USAGE ;
      n = atoi(argv[++i]) ;
      if ( n <= 0 ) goto USAGE ;
    }
    else if ( strcmp(argv[i], "-l") == 0 || strcmp(argv[i], "-reclen") == 0 ) {
      if ( i >= argc - 1 ) goto USAGE ;
      l = atoi(argv[++i]) ;
      if ( l <= 0 ) {
	fprintf(stderr, "Must specify record length\n") ;
	goto USAGE ;
      }
    }
    else if ( strcmp(argv[i], "-s") == 0 || strcmp(argv[i], "-skip") == 0 ) {
      if ( i >= argc - 1 ) goto USAGE ;
      s = atoi(argv[++i]) ;
      if ( s < 0 ) goto USAGE ;
    }
    else if ( strcmp(argv[i], "-m") == 0 || strcmp(argv[i], "-max") == 0 ) {
      if ( i >= argc - 1 ) goto USAGE ;
      m = atoi(argv[++i]) ;
      if ( m < 0 ) goto USAGE ;
    }
    else if ( strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "-verbose") == 0 ) {
      verbose = 1 ;
    }
    else {
    USAGE:
      fprintf(stderr, "%s", usage) ;
      exit(1) ;
    }
  }
  buf = malloc(l) ;
  if ( buf == NULL ) {
    if ( verbose ) 
      fprintf(stderr, "Out of memory allocating read buffer\n") ;
    exit(5) ;
  }
  /* Skip at start */
  for ( i = 0 ; i < s ; i++ ) {
    nread = fread(buf, l, 1, stdin) ;
    if ( nread < 1 && verbose ) {
      fprintf(stderr, "Only read %d item(s) at start\n", skipped) ;
      goto END ;
    }
    skipped++ ;
  }
  if ( skipped && verbose ) fprintf(stderr, "Skipped %d recs at start\n", skipped) ;
  inrecs += s ;
  while ( 1 ) {
    if ( feof(stdin) ) break ;
    nread = fread(buf, 1, l, stdin) ;
    if ( nread == 0 ) break ;
    if ( nread < l ) 
      fprintf(stderr, "Record %d is only %d bytes\n", count + 1, nread) ;
    inrecs++ ;
    if ( fwrite(buf, nread, 1, stdout) != 1 ) {
      fprintf(stderr, "Unable to write record # %d\n", count + 1) ;
      break ;
    }
    count++ ;
    if ( m > 0 && count >= m ) break ;
    fread(buf, l, n - 1, stdin) ;
    if ( feof(stdin) && verbose ) {
      fprintf(stderr, "Reached end of input\n") ;
      break ;
    }
    inrecs += n - 1 ;
  }
END:
  if ( verbose ) {
    fprintf(stderr, "Input %d record(s)\n", inrecs) ;
    fprintf(stderr, "Output %d record(s)\n", count) ;
  }
  exit(0) ;
}
