/******************************************************************************

  extract_text.c SW March 97

  Text on stdin is parsed and output as

    <term1>:<gsl1>[:<term2>:<gsl2>]

  DB name and, if needed,  attr must be in environment.

******************************************************************************/
#include "defines.h"
#include "I0.h"

#define BUFSIZE 1024 * 1024 * 8

main(argc, argv)
     int argc ;
     char **argv ;
{
  int terms = 0 ;
  char *attr = "default" ;
  int result, res ;
  char *envp ;
  char *buf1, *buf2 ;
  char *p ;
  char *ip ;
  int nk ;
  int size = BUFSIZE ;
  int c ;
  int err = 0 ;
  int spurious = 0 ;
  int verboseness = 0 ;
  char *prog = argv[0] ;

  extern char *optarg ;

  while ((c = getopt(argc, argv, "phv:")) != -1)
    switch (c) {
    case 'p':
      fprintf(stderr, "%s: -p option not supported yet\n", prog) ;
      err = 1 ;
      break ;
    case 'h':
      err = 1 ;
      break ;
    case 'v':
      verboseness = atoi(optarg) ;
      break ;
    }
  if ( err ) {
    fprintf(stderr, "Usage: %s [-p] [-v #} < \n", argv[0]) ;
    fprintf(stderr,
 "Outputs <term1>:<gsl1>[:<term2>:<gsl2>]\n") ;
    fprintf(stderr, "NB: doesn't do pairs at present\n") ;
    fprintf(stderr, "Needs (at least) BSS_DB in env.\n") ;
    exit(1) ;
  }

  if ( (envp = getenv("BSS_ATTRIBUTE")) != NULL ) attr = envp ;
  if ( startup_bss() < 0 ) {
    fprintf(stderr, "%s: can't start BSS\n", prog) ;
    exit(1) ;
  }

  buf1 = malloc(BUFSIZE) ;
  buf2 = malloc(BUFSIZE) ;
  if ( buf1 == NULL || buf2 == NULL ) {
    fprintf(stderr,
 "%s: can't allocate memory for two buffers of size %d\n",
	    prog, BUFSIZE) ;
    iexit() ;
    exit(2) ;
  }
  p = buf1 ;
  while ( 1 ) {
    if ( size <= 0 ) {
      fprintf(stderr, "%s: buffers not big enough\n", prog) ;
      exit(2) ;
    }
    if ( fgets(p, size, stdin) == NULL ) break ;
    size -= strlen(p) ; p += strlen(p) ;
    if ( feof(stdin) ) break ;
  }
  if ( verboseness > 0 ) 
    fprintf(stderr, "%s: read %d chars\n", prog, BUFSIZE - size) ;

  /* Next bit more or less same as extract_rec() in basic_expand */

  nk = bss_superparse(buf1, strlen(buf1), attr, buf2) ;
  if ( nk < 0 ) {
    fprintf(stderr, "Can't parse the text (length %d)\n", strlen(buf1)) ;
    exit(1) ;
  }
  ip = buf2 ;
  /* Need something quite different if pairs */
  /* If not pairs..*/
  while ( nk-- ) {
    if ( *ip == 'H' || *ip == 'F' ||
	 strlen(ip + 1) == 0 || strlen(ip) == 0) {
      ip += strlen(ip) + 1 ;
      ip += strlen(ip) + 1 ;
      continue ;
    }
    else if ( *ip != ' ' && *ip != 'S' && *ip != 'G' && *ip != 'I' ) {
      fprintf(stderr, "%s: spurious term '%s'\n", prog, ip) ;
      spurious++ ;
      ip += strlen(ip) + 1 ;
      ip += strlen(ip) + 1 ;
      continue ;
    }
    if ( *ip == ' ' ) *ip = 'N' ;
    printf("%s:%c::\n", ip + 1, *ip) ;	/* First char is GSL code */
    terms++ ; 
    ip += strlen(ip) + 1 ;
    ip += strlen(ip) + 1 ;
  }
  if ( verboseness > 0 )
    fprintf(stderr, "%s: %d terms\n", prog, terms) ;
  if ( spurious > 0 )
    fprintf(stderr, "%s: %d spurious terms\n", prog, spurious) ;
  iexit() ;
  exit(0) ;
}

