/******************************************************************************

  read_pf   SW Jan 96

  Prog to read through type 4-7 postings file, expanding pos fields if needed

  Mod Nov 97 to produce ascii files of <recnum> <npstgs> or
  <recnum> <ttf> as required

  Mod Dec 99 to produce on stdout output to form doclengths file of number
  of indexed tokens per rec.

******************************************************************************/

#include <stdio.h>
#include "defines.h"
unsigned Toktab ;		/* Dummy */

char *Usage = "Usage: %s -itype <index type> -N <N> [-stats] [-pstgs] [-ttf]\\\n\
[-nodoclens] [-noisy] [-h]\n\
%s takes a postings file as input and by default produces\n\
a document lengths file of 4-byte integers in which record n is the length\n\
in indexed tokens of document n+1.\n\n\
The number of documents in the database (-N arg) and the index type number\n\
(-itype arg) must be specified (index types 8, 9, 10, 11, 12 and 13 supported)\n\
For other args see the source. Note it will not work correctly at present for\n\
indexes with variable length record numbers.\n" ;

main(argc, argv)
     int argc ; char **argv ;
{
  int silent = 1 ;
  int numpos ;
  unsigned pos ;
  unsigned rec = 0 ;
  unsigned offset = 0 ;
  unsigned start_offset ;
  int plength ;
  int j ;
  long long bignumpos = 0, pstgs = 0 ;
  long long ttf = 0 ;
  int nopos = 0, stats = 0, doclens = 1 ;
  int itype = -1 ;
  int large_rn = 0 ;
  int do_ttf = 0 ;		/* If 0 does npstgs */
  int N = -1 ;
  int *recarray ;
  long long ovfrecs = 0 ;

  for ( j = 1 ; j < argc ; j++ ) {
    if ( strcmp(argv[j], "-itype") == 0 ) itype = atoi(argv[++j]) ;
    else if ( strcmp(argv[j], "-ttf") == 0 ) do_ttf = 1 ;
    else if ( strcmp(argv[j], "-pstgs") == 0 ) do_ttf = 0 ;
    else if ( strcmp(argv[j], "-stats") == 0 ) stats = 1 ;
    else if ( strcmp(argv[j], "-N") == 0 ) N = atoi(argv[++j]) ;
    else if ( strcmp(argv[j], "-nodoclens") == 0 ) doclens = 0 ;
    else if ( strcmp(argv[j], "-noisy") == 0 ) silent = 0 ;
    else if ( strcmp(argv[j], "-h") == 0 ) {
      fprintf(stderr, Usage, argv[0], argv[0]) ;
      exit(1) ;
    }
    else {
      fprintf(stderr, "Unknown arg %s\n", argv[j]) ;
      fprintf(stderr, Usage, argv[0], argv[0]) ;
      exit(1) ;
    }
  }
  if ( N < 0 ) {
    fprintf(stderr, "%s: must specify N as -N <N>\n", argv[0]) ;
    exit(1) ;
  }
  if ( itype < 0 ) {
    fprintf(stderr, "%s: must specify index type number as -itype <itype>\n",
	    argv[0]) ;
    exit(1) ;
  }
  if ( itype < 8 || itype > 13 ) {
    fprintf(stderr, "%s: only index types 8 - 13 supported at present\n",
	    argv[0]) ;
    exit(1) ;
  }
  if ( itype == 10 || itype == 11 ) nopos = 1 ;
  if ( itype == 11 || itype == 12 || itype == 13 ) large_rn = 1 ;
  if ( doclens ) {
    stats = 1 ;
    do_ttf = 1 ;
  }
  if ( stats ) {
    recarray = (int *)calloc(N + 1, sizeof(int)) ;
    if ( recarray == NULL ) {
      fprintf(stderr, "Not enough memory to allocate recnum array\n") ;
      exit(1) ;
    }
  }
    
  while ( 1 ) {
    numpos = (int)fget_vshort(stdin) ;
    if ( numpos == EOF ) break ;
    start_offset = offset ;
    ttf += numpos ;
    if ( numpos < 128 ) offset++ ;
    else {
      offset += 2 ; bignumpos++ ;
    }
    if ( large_rn ) {
      rec = getw(stdin) ;
      offset += 4 ;
    }
    else {
#if HILO
      fread(((char *)&rec) + 1, 3, 1, stdin) ;
#else
      fread(((char *)&rec), 3, 1, stdin) ;
#endif
      offset += 3 ;
    }
    if ( stats ) {
      if ( rec <= N )  {
	if ( ! do_ttf ) recarray[rec]++ ;
	else recarray[rec] += numpos ;
      }
      else ovfrecs++ ;
    }
    if ( ! nopos ) {
      for ( j = 0 ; j < numpos ; j++ ) {
	
	pos = (unsigned)getw(stdin) ;
	offset += sizeof(unsigned) ;
      }
      plength = offset - start_offset ;
    }
    pstgs++ ;
    if ( ! silent && pstgs % 1000000 == 0 )
      fprintf(stderr, "Done %lld pstgs\n", pstgs) ;
  }
  if ( ! stats ) {
    printf("%lld pstgs, %lld ttf, %lld more than 127 pos fields\n",
	   pstgs, ttf, bignumpos) ;
  }
  else {
    if ( ! silent ) fprintf(stderr, "Writing output\n") ;
    for ( j = 1 ; j <= N ; j++ ) {
      if ( ! silent ) fprintf(stderr, "%d %d\n", j, recarray[j]) ;
      (void)putw(recarray[j], stdout) ;
    }
    fprintf(stderr, "END\n") ;
    fprintf(stderr, "%lld pstgs, %lld ttf, %lld more than 127 pos fields\n",
	    pstgs, ttf, bignumpos) ;
    if ( ovfrecs ) fprintf(stderr, "%lld postings were for recnums > N (%d)\n",
			   ovfrecs, N) ;
  }
  exit(0) ;
}
