/******************************************************************************

  read_si  SW Feb 95

  Prog to output Okapi secondary index readably to stdout

******************************************************************************/
#include "bss.h"
int silent ;

int
main(int argc, char **argv)
{
  int itype = -1 ;
  int i ;
  unsigned short chunkword ;
  unsigned stnp = 0, ftnp = 0xffffff ;
  char term[512] ;
  char sterm[512] ;
  char fterm[512] ;
  int finished = 0 ;
  int hex = 0 ;

  char buf[512] ;
  Transfer_itr tr ;
  int result ;

  int totterms = 0;		/* Stats added Aug 2000 */
  long long totpstgs = 0, tottf = 0, tottermlen = 0 ; /* Ditto */
  int statsonly = 0 ;

  char *usage = 
"Usage: read_si -itype <index_type> [-silent] [-hex] [-start <start string>]\n\
  [-finish <finish string>] [-stnp <min np>] [-ftnp <max np>] [-statsonly]\n\n\
Output is <lims>:<maxtf>:<ttf>:<tnp>:<plength>:[<pfvol>:]<poffset>:<term>\n\
or if statsonly just a final report of numbers of terms, postings and tokens\n\
" ;

  *sterm = '\0' ;
  *fterm = '\177' ;
  *(fterm + 1) = '\0' ;

  if ( argc < 2 ) {
    fprintf(stderr, usage) ;
    exit(1) ;
  }
  for ( i = 1 ; i < argc ; i++ ) {
    if ( strcmp(argv[i], "-itype") == 0 ) {
      itype = atoi(argv[++i]) ;
      if ( itype > 13 || itype < 8 ) {
	fprintf(stderr, "Index types are 8 - 13\n") ;
	exit(1) ;
      }
    }
    else if ( strcmp(argv[i], "-silent") == 0 ) silent = 1 ;
    else if ( strcmp(argv[i], "-help") == 0 ) {
      fprintf(stderr, usage) ;
      exit(1) ;
    }
    else if ( strcmp(argv[i], "-start") == 0 )
      strcpy(sterm, argv[++i]) ;
    else if ( strcmp(argv[i], "-finish") == 0 )
      strcpy(fterm, argv[++i]) ;
    else if ( strcmp(argv[i], "-stnp") == 0 ) stnp = atoi(argv[++i]) ;
    else if ( strcmp(argv[i], "-ftnp") == 0 ) ftnp = atoi(argv[++i]) ;
    else if ( strcmp(argv[i], "-hex") == 0 ) hex = 1 ;
    else if ( strcmp(argv[i], "-statsonly") == 0 ) statsonly = 1 ;
    else {
      fprintf(stderr, "Unknown arg %s\n", argv[i]) ;
      fprintf(stderr, usage) ;
      exit(1) ;
    }
  }
  if ( itype < 0 ) {
    fprintf(stderr, "Must specify index type as -itype <index type>\n") ;
    fprintf(stderr, usage) ;
    exit(1) ;
  }

  /* Start first chunk */
  fread((void *)&chunkword, sizeof(short), 1, stdin) ;
  while ( 1 ) {
    result = read_itr_seq(buf, itype, stdin) ;
    if ( result == 0 ) break ;	/* EOI */
    parse_itr(buf, itype, &tr) ;
    memcpy(term, tr.term, tr.termlen) ;
    *(term + tr.termlen) = '\0' ;
    if ( strcmp(term, sterm) < 0 ) continue ;
    if ( strcmp(term, fterm) > 0 ) {
      finished = 1 ;
      break ;
    }
    if ( tr.tnp < stnp || tr.tnp > ftnp ) continue ;
    
    if ( ! statsonly ) {
      if ( hex )
	printf("%04x:%5d:%9d:%9d:%9d:%1d:%11llu:%s\n",
	       (unsigned)tr.limits, (int)tr.maxtf,
	       tr.ttf, tr.tnp, tr.plength, tr.pstg_vol, tr.pstg_offset, term) ;
      else 
	printf("%05d:%5d:%9d:%9d:%9d:%1d:%11llu:%s\n",
	       (unsigned)tr.limits, (int)tr.maxtf,
	       tr.ttf, tr.tnp, tr.plength, tr.pstg_vol, tr.pstg_offset, term) ;
	}
    totterms++ ;
    tottermlen += tr.termlen ;
    totpstgs += tr.tnp ;
    tottf += tr.ttf ;
  } /* Main loop */
  if ( statsonly ) 
    printf("%d terms %lld ttlen %lld postings %lld tokens\n",
	   totterms, tottermlen, totpstgs, tottf) ;
  else if ( ! silent ) 
    fprintf(stderr, "%d terms %lld ttlen %lld postings %lld tokens\n",
	    totterms, tottermlen, totpstgs, tottf) ;
  exit(0) ;
}

