/******************************************************************************

  read_si  SW Feb 95

  Prog to output Okapi secondary index readably to stdout

******************************************************************************/
/*#include "defines.h"*/
#include <stdio.h>
#include <stdlib.h>

int
main(int argc, char **argv)
{
  int itype = -1 ;
  int i ;
  unsigned short chunkword ;
  unsigned short itrlen ;
  unsigned char termlen, closing_termlen ;
  unsigned short logtf ;
  unsigned short limits ;
  unsigned short maxtf ;
  int ttf = 0 ;
  int plength ;
  unsigned tnp = 0 ;
  unsigned stnp = 0, ftnp = 0xffffff ;
  int pf_vol = 0 ;
  unsigned long long poffset = 0 ;
  unsigned short closing_itrlen ;
  char term[512] ;
  char sterm[512] ;
  char fterm[512] ;
  int chunknum = 0, termnum = 0 ;
  int finished = 0, silent = 0 ;
  int hex = 0 ;
  int ch ;

  int totterms = 0;		/* Stats added Aug 2000 */
  int nchunks = 0 ;
  long long totpstgs = 0, tottf = 0 ; /* Ditto */
  int statsonly = 0 ;

  char *usage = 
"Usage: read_si -itype <index_type> [-silent] [-hex] [-start <start string>]\n\
  [-finish <finish string>] [-stnp <min np>] [-ftnp <max np>] [-statsonly]\n\n\
Output is <lims>:<maxtf>:<ttf>:<tnp>:<plength>:[<pfvol>:]<poffset>:<term>
or if statsonly just a final report of numbers of terms, postings and tokens
" ;

  *sterm = '\0' ;
  *fterm = '\177' ;
  *(fterm + 1) = '\0' ;

  if ( argc < 2 ) {
    fprintf(stderr, usage) ;
    exit(1) ;
  }
  for ( i = 1 ; i < argc ; i++ ) {
    if ( strcmp(argv[i], "-itype") == 0 ) {
      itype = atoi(argv[++i]) ;
      if ( itype > 13 || itype < 4 ) {
	fprintf(stderr, "Index types are 4 - 13\n") ;
	exit(1) ;
      }
    }
    else if ( strcmp(argv[i], "-silent") == 0 ) silent = 1 ;
    else if ( strcmp(argv[i], "-help") == 0 ) {
      fprintf(stderr, usage) ;
      exit(1) ;
    }
    else if ( strcmp(argv[i], "-start") == 0 )
      strcpy(sterm, argv[++i]) ;
    else if ( strcmp(argv[i], "-finish") == 0 )
      strcpy(fterm, argv[++i]) ;
    else if ( strcmp(argv[i], "-stnp") == 0 ) stnp = atoi(argv[++i]) ;
    else if ( strcmp(argv[i], "-ftnp") == 0 ) ftnp = atoi(argv[++i]) ;
    else if ( strcmp(argv[i], "-hex") == 0 ) hex = 1 ;
    else if ( strcmp(argv[i], "-statsonly") == 0 ) statsonly = 1 ;
    else {
      fprintf(stderr, "Unknown arg %s\n", argv[i]) ;
      fprintf(stderr, usage) ;
      exit(1) ;
    }
  }
  if ( itype < 0 ) {
    fprintf(stderr, "Must specify index type as -itype <index type>\n") ;
    fprintf(stderr, usage) ;
    exit(1) ;
  }

  while ( 1 ) {
    /* Start a chunk */
    fread((char *)&chunkword, sizeof(short), 1, stdin) ;
    chunknum++ ;
    nchunks++ ;
    while ( 1 ) {
      if ( itype >= 4 && itype <= 7 ) {
	if (fread((char *)&itrlen, sizeof(short), 1, stdin) < 0 ) {
	  if ( ! silent ) fprintf(stderr, "Reached end of file\n") ;
	  finished = 1 ;
	  break ;
	}
	if ( itrlen == 0x8000 ) {	/* End of index */
	  if ( ! silent ) fprintf(stderr, "Reached end of index\n") ;
	  finished = 1 ;
	  break ;
	}
	else if ( itrlen == 0 ) { /* End of chunk */
	  /* Read off null bytes */
	  while ( ( ch = getchar() ) == 0 ) ;
	  ungetc(ch, stdin) ;
	  if ( ! silent ) fprintf(stderr, "End of chunk %d\n", chunknum) ;
	  break ;
	}
	fread((char *)&termlen, sizeof(char), 1, stdin) ;
	fread(term, termlen, 1, stdin) ;
	*(term + termlen) = '\0' ;
	fread((char *)&limits, sizeof(short), 1, stdin) ;
	fread((char *)&logtf, sizeof(short), 1, stdin) ;
	fread(((char *)&tnp) + 1, 3, 1, stdin) ;
	if ( itype == 6 || itype == 7 ) pf_vol = getchar() ;
#if HILO
	fread((char *)&poffset + 4, 4, 1, stdin) ;
#else
	fread((char *)&poffset, 4, 1, stdin) ;
#endif
	fread((char *)&closing_itrlen, sizeof(short), 1, stdin) ;
	termnum++ ;
	if ( closing_itrlen != itrlen ) {
	  fprintf(stderr, "Closing itrlen %d not same as itrlen %d\n",
		  closing_itrlen, itrlen) ;
	  exit(1) ;
	}
	if ( strcmp(term, sterm) < 0 ) continue ;
	if ( strcmp(term, fterm) > 0 ) {
	  finished = 1 ;
	  break ;
	}
	if ( tnp < stnp || tnp > ftnp ) continue ;

	if ( ! statsonly ) {
	  if ( hex ) 
	    printf("%04x:%3d:%6d:%1d:%11llu:%s\n",
		   (unsigned)limits, (int)logtf,
		   (int)tnp, pf_vol, poffset, term) ;
	  else 
	    printf("%05d:%3d:%6d:%1d:%11llu:%s\n",
		   (unsigned)limits, (int)logtf,
		 (int)tnp, pf_vol, poffset, term) ;
	}
      }	/* itype <=7 */
      else {
	termlen = (unsigned char)getchar() ;
	if ( termlen == 0xff ) {	/* End of index */
	  if ( ! silent ) fprintf(stderr, "Reached end of index\n") ;
	  finished = 1 ;
	  break ;
	}
	else if ( termlen == 0 ) { /* End of chunk */
	  /* Read off null bytes */
	  while ( ( ch = getchar() ) == 0 ) ;
	  ungetc(ch, stdin) ;
	  if ( ! silent ) fprintf(stderr, "End of chunk %d\n", chunknum) ;
	  break ;
	}
	fread(term, termlen, 1, stdin) ;
	*(term + termlen) = '\0' ;
	fread((char *)&limits, sizeof(short), 1, stdin) ;
	fread((char *)&maxtf, sizeof(short), 1, stdin) ;
	fread((char *)&ttf, sizeof(int), 1, stdin) ;
	fread((char *)&tnp, sizeof(int), 1, stdin) ;
	fread((char *)&plength, sizeof(int), 1, stdin) ;
	pf_vol = getchar() ;
#if HILO
	if ( itype <= 10 ) 
	  fread((char *)&poffset + 4, 4, 1, stdin) ;
	else fread((char *)&poffset + 3, 5, 1, stdin) ;
#else
	if ( itype <= 10 ) 
	  fread((char *)&poffset, 4, 1, stdin) ;
	else fread((char *)&poffset, 5, 1, stdin) ;
#endif
	closing_termlen = (unsigned char)getchar() ;
	termnum++ ;
	if ( closing_termlen != termlen ) {
	  fprintf(stderr,
		  "Closing termlen %d not same as termlen %d, term %s\n",
		  (int)closing_termlen, (int)termlen, term) ;
	  exit(1) ;
	}
	if ( strcmp(term, sterm) < 0 ) continue ;
	if ( strcmp(term, fterm) > 0 ) {
	  finished = 1 ;
	  break ;
	}
	if ( tnp < stnp || tnp > ftnp ) continue ;

	if ( ! statsonly ) {
	  if ( hex )
	    printf("%04x:%5d:%9d:%9d:%9d:%1d:%11llu:%s\n",
		   (unsigned)limits, (int)maxtf,
		   (int)ttf, (int)tnp, (int)plength, pf_vol, poffset, term) ;
	  else 
	    printf("%05d:%5d:%9d:%9d:%9d:%1d:%11llu:%s\n",
		   (unsigned)limits, (int)maxtf,
		   (int)ttf, (int)tnp, (int)plength, pf_vol, poffset, term) ;
	}
      }	/* itype > 7 */
      totterms++ ;
      totpstgs += tnp ;
      tottf += ttf ;
    } /* chunk */
    if ( finished ) break ;
  } /* Main loop */
  if ( statsonly ) 
    printf("%d chunks %d terms %lld postings %lld tokens\n",
	   nchunks, totterms, totpstgs, tottf) ;
  else if ( ! silent ) 
    fprintf(stderr, "%d chunks %d terms %lld postings %lld tokens\n",
	    nchunks, totterms, totpstgs, tottf) ;
  exit(0) ;
}
