/****************************************************************************

  search_and_extract.c  SW Nov 99, June 00

  from expd.c

  input: <term>:<weight>:<src>:<qtf>

  output term:N:n:R:r:cttf:tf:Rttf:rtf:qtf:qlen:src:s1:s2

  args: -R <R> (= number of docs from pilot search, mandatory)

  Many environment variables may be supplied and two are required

******************************************************************************/

#undef DEBUG

#include "expd.h"
#include "bss.h"

extern struct db Cdb ;

char *info = 
"A query is given on stdin in the form <term>:<weight>:<source>:<qtf>\n\
and an initial search done using it. The top argv[1] docs from this search\n\
become the 'relevant set'. Terms are then extracted from the docs of the \n\
relevant set, everything is reweighted and 'scored' using both the Robertson \n\
offer weight (rsv) and an 'absolute' score (rscore).\n\n\
   Input:  qterm:weight:source:qtf\n\
  Output:  term:N:n:R:r:cttf:tf:Rttf:rtf:qtf:qlen:source:rscore:rsv\n\
\n\
In the output, source is given as \"doc\" for non-query terms.\n\
Environment: BSS_DB is mandatory\n\
BSS_LIMIT, BSS_ATTRIBUTE, BSS_OP, BSS_K1, BSS_BM25B, BSS_K3, BSS_MEM\n\
EXPD_MINn, EXPD_MINr, EXPD_MINrtf, EXPD_NOI are optional, overriding defaults\n\
(see expd.h). Note that MINn, MINr, MINrtf and NOI don't apply to query terms.\n\
EXPD_WANTED_FIELDS may be a colon-separated list of field nos. for extraction\n\
-- the default is to use the last field.\n\n\
Note that the program uses a script ttstat which must be in the search path.\n" ;


int
main(int argc, char **argv)
{
  char *dbname ;
  int num_wf ;
  int wf[MAXNUMFIELDS + 1] ;
  int Nrels = 0 ;		/* Number of "rel" docs wanted */
  int R ;			/* Number obtained */
  int Rtf = 0 ;			/* Total # tokens in relset */
  int numaddedterms = 0 ;	/* No of non-queryterms from extract */
  int numqterms = 0 ;
  int numterms = 0 ;
  int qlen = 0 ;
  int N ;
  int cttf ;
  char *Command, *Response ;
  char buf[1024], term[1024] ;
  FILE *tmpf;
  char tmpfname[256], tmpf1name[256] ;
  BOOL found ;
  int initial_set ;
  int relset ;
  int j, k ;
  struct term *terms ;
  int res ;
  int target ;
  char *p ;
  int num_fields ;
  int chars_read ;
  char *envp ;
  char *attr = DEFAULT_ATTRIBUTE ;
  char *op = DEFAULT_OP ;
  double k1 = DEFAULT_K1, b = DEFAULT_B, k3 = DEFAULT_K3 ;
  int limit = DEFAULT_LIMIT ;
  int noI = DEFAULT_NOI ;
  int minn = DEFAULT_MINn, minr = DEFAULT_MINr, minrtf = DEFAULT_MINrtf ;
  int skippedIs = 0, rtoosmall = 0, rtftoosmall = 0, ntoosmall = 0 ;

  terms = (struct term *)calloc(MAXTERMS, sizeof(struct term)) ;
  Command = (char *)malloc(LARGE_BUFSIZE) ;
  Response = (char *)malloc(LARGE_BUFSIZE) ;
  if ( ! terms || ! Command || ! Response ) {
    fprintf(stderr, "%s: out of memory\n", argv[0]) ;
    exit(1) ;
  }
  if ( argc < 2 || strcmp(argv[1], "-h") == 0 ) {
    fprintf(stderr, "Usage: %s <R>\n", argv[0]) ;
    fprintf(stderr, info) ;
    exit(1) ;
  }
  limit = 0 ;
  Nrels = atoi(argv[1]) ;

  if ( (envp = getenv("BSS_DB")) && *envp ) dbname = envp ;
  else {
    fprintf(stderr, "No database (BSS_DB) in environment\n") ;
    exit(1) ;
  }
  if ( (envp = getenv("BSS_LIMIT"))  && *envp ) limit = atoi(envp) ;
  if ( (envp = getenv("BSS_ATTRIBUTE")) && *envp ) attr = envp ;
  if ( (envp = getenv("BSS_OP")) && *envp ) op = envp ;
  if ( (envp = getenv("BSS_K1")) && *envp ) k1 = atof(envp) ;
  if ( (envp = getenv("BSS_BM25B")) && *envp ) b = atof(envp) ;
  if ( (envp = getenv("BSS_K3")) && *envp ) k3 = atof(envp) ;
  if ( (envp = getenv("EXPD_MINn")) && *envp ) minn = atoi(envp) ;
  if ( (envp = getenv("EXPD_MINr")) && *envp ) minr = atoi(envp) ;
  if ( (envp = getenv("EXPD_MINrtf")) && *envp ) minrtf = atoi(envp) ;
  if ( (envp = getenv("EXPD_NOI")) && *envp ) noI = atoi(envp) ;

  iinit() ;
  sprintf(Command, "choose %s\n", dbname) ;
  I0 ;
  fprintf(stderr, "Using database %s\n", dbname) ;
  sprintf(Command, "info db\n") ;
  I0 ;
  num_fields = atoi(strstr(Response, "nfields=") + 8) ;
  num_wf = 1 ;
  if ( (envp = getenv("EXPD_WANTED_FIELDS")) && *envp ) {
    char *p ;
    wf[0] = atoi(strtok(envp, ":")) ;
    for ( j = 1 ; j < num_fields ; j++ ) {
      p = strtok(NULL, ":") ;
      if ( p == NULL ) break ;
      wf[j] = atoi(p) ;
      num_wf++ ;
    }
  }
  else wf[0] = num_fields ;	/* Assume last field */
  sprintf(Command, "set attr=%s\n", attr) ;
  I0 ;
  fprintf(stderr, "Using attribute %s\n", attr) ;
  if ( (envp = getenv("BSS_MEM")) && *envp ) {
    sprintf(Command, "set memory=%d\n", atoi(envp)) ;
    I0 ;
  }
  if ( (envp = getenv("BSS_DEBUG")) && *envp ) {
    sprintf(Command, "set debug=%d\n", atoi(envp)) ;
    I0 ;
  }
  if ( limit > 0 ) {
    sprintf(Command, "set limit=%d\n", limit) ;
    I0 ;
    fprintf(stderr, "Using limit %d\n", limit) ;
  }
  sprintf(Command, "d bign\n") ;
  I0 ;
  N = atoi(Response) ;
  sprintf(Command, "d cttf\n") ;
  I0 ;
  cttf = atoi(Response) ;
  fprintf(stderr, "Using N %d\n", N) ;
  fprintf(stderr, "Using cttf %d\n", cttf) ;
  fprintf(stderr, "Using op %s\n", op) ;
  fprintf(stderr, "Using k1 %.2f\n", k1) ;
  fprintf(stderr, "Using bm25b %.2f\n", b) ;
  fprintf(stderr, "Using k3 %.2f\n", k3) ;
  fprintf(stderr, "Using min n %d\n", minn) ;
  fprintf(stderr, "Using min r %d\n", minr) ;
  fprintf(stderr, "Using min rtf %d\n", minrtf) ;
  fprintf(stderr, "Using no I %d\n", noI) ;

  /* Form query termsets */
  if ( Debug ) fprintf(stderr, "Looking up query terms\n") ;
  while ( ! feof(stdin) ) {
    if ( fgets(buf, 1023, stdin) == NULL ) break ;
    if ( numterms >=MAXTERMS ) {
      fprintf(stderr, "Sorry, too many terms\n") ;
      goto error ;
    }
    sscanf(buf, "%[^:]:%lf:%[^:]:%d",
	   term, &terms[numterms].wt,
	   terms[numterms].source, &terms[numterms].qtf) ;
    qlen += terms[numterms].qtf ;
    sprintf(Command, "f t=%s", term) ;
    I0 ;
    terms[numterms].n = atoi(strstr(Response, "np=") + 3) ;
    if ( terms[numterms].n == 0 ) {
      fprintf(stderr, "Warning: term %s not found\n", term) ;
      continue ;
    }
    terms[numterms].set = atoi(Response + 1) ;
    sprintf(Command, "dump ttf") ;
    I0 ;
    terms[numterms].ttf = atoi(Response + 4) ;
    
    terms[numterms].term= strdup(term) ;

    numterms++ ;
    numqterms++ ;
  }
  fprintf(stderr, "Got %d query terms\n", numqterms) ;
  fprintf(stderr, "Forming initial feedback set\n") ;
  p = Command ;
  sprintf(p, "f") ; p += strlen(p) ;
  target = Nrels ;
  for ( j = 0 ; j < numqterms ; j++ ) {
    sprintf(p, " s=%d w=%.3f", terms[j].set, terms[j].wt) ;
    p += strlen(p) ;
  }
  sprintf(p, " op=%s k1=%.3f bm25b=%.3f target=%d", op, k1, b, target) ;
  I0 ;
  initial_set = atoi(Response + 1) ;
  sprintf(Command, "f s=%d top=%d", initial_set, target) ;
  I0 ;

  R = atoi(strstr(Response, "np=") + 3) ;
  relset = atoi(Response + 1) ;

  {
    int termsthis, j, nk ;
    sprintf(tmpf1name, "/tmp/tmpf1.%d", (int)getpid()) ;
    sprintf(tmpfname, "/tmp/tmpf.%d", (int)getpid()) ;
    tmpf = fopen(tmpfname, "w+") ;
    for ( j = 0 ; j < R ; j++ ) {
      /* For treatment of passages see older version of this, expd */
      for ( k = 0 ; k < num_wf ; k++ ) {
	sprintf(Command,
		"extract s=%d r=%d sources=0 field=%d",
		relset, j, wf[k]) ;
	I0 ;
	nk = atoi(Response) ;
	p = Response ;
	sscanf(p, "%*[^\n]\n%n", &chars_read) ;
	p += chars_read ;
	termsthis = 0 ;
	while ( nk-- ) {
	  sscanf(p, "%[^\n]\n%n", buf, &chars_read) ;
	  p += chars_read ;
	  fprintf(tmpf, "%s:%d\n", buf, j) ; /* First char is GSL code */
	  termsthis++ ;
	}
	if ( Debug ) 
	  fprintf(stderr, "Extracted %d terms from field %d of %d of %d\n",
		  termsthis, wf[k], j, R) ;
      }
    }
    fclose(tmpf) ;
    sprintf(Command,
	    "ttstat < %s > %s\n",
	    tmpfname, tmpf1name) ;
    system(Command) ;
    /* Now have tmpf1 containing all the terms uniquely, each followed by
       Rtf, tf, r. */

    {
      char buf[256] ;
      char p[256] ;
      char code ;
      int rtf ;
      int r ;
      int n ;
      int termnum ;
      char response[1024], command[1024] ;
      tmpf = fopen(tmpf1name, "r") ;
      while ( ! feof(tmpf) ) {
	if ( numterms >= MAXTERMS ) {
	  fprintf(stderr, "Too many terms\n") ;
	  break ;
	}
	fgets(buf, 256, tmpf) ;
	if ( feof(tmpf) ) break ;
	/* Recalc Rtf as may exclude counts for non-topic I-terms */
	sscanf(buf, "%c%[^:]:%*d:%d:%d", &code, p, &rtf, &r) ;
	for ( k = 0, found = FALSE ; k < numqterms ; k++ ) {
	  if ( strcmp(p, terms[k].term) == 0 ) {
	    found = TRUE ;
	    fprintf(stderr, "Found %s\n", p) ;
	    break ;
	  }
	}
	
	if ( found ) termnum = k ;
	else {			/* Add it unless I-term and noI */
	  if ( code == 'I' && noI ) {
	    if ( Debug >= 1 )
	      fprintf(stderr, "Skipping I-term %s\n", p) ;
	    skippedIs++ ;
	    continue ;
	  }
	  if ( r < minr ) {
	    rtoosmall++ ;
	    continue ;
	  }
	  if ( rtf < minrtf ) {
	    rtftoosmall++ ;
	    continue ;
	  }
	  sprintf(command, "f save=0 t=%s", p) ;
	  I0_ ;
	  n = atoi(response + 3) ;
	  if ( n < minn ) {
	    ntoosmall++ ;
	    continue ;
	  }
	  termnum = numterms ;
	  terms[termnum].term = strdup(p) ;
	  terms[termnum].set = -1 ;
	  terms[termnum].n = n ;
	  sprintf(command, "d ttf") ;
	  I0_ ;
	  terms[termnum].ttf = atoi(response) ;
	  terms[termnum].qtf = 0 ;
	  sprintf(terms[termnum].source, "doc") ;
	  numterms++ ;
	  numaddedterms++ ;
	  if ( Debug >= 1 )
	    fprintf(stderr, "Added %s, term %d\n", p, numterms) ;
	}
	/* found or not */
	terms[termnum].r = r ;
	terms[termnum].rtf = rtf ;
	Rtf += rtf ;
	sprintf(command, "rscore n=%d r=%d bigr=%d",
		terms[termnum].n, terms[termnum].r, R) ;
	I0_ ;
	terms[termnum].rscore = atof(response) ;
	if ( terms[termnum].qtf > 1 )
	  sprintf(command, "weight n=%d wt_fn=1 r=%d bigr=%d k3=%.3f qtf=%d\n",
		  terms[termnum].n, terms[termnum].r, R,
		  k3, terms[termnum].qtf) ;
	else
	  sprintf(command, "weight n=%d wt_fn=1 r=%d bigr=%d\n",
		  terms[termnum].n, terms[termnum].r, R) ;
	I0_ ;
	terms[termnum].rsv = atof(response) * terms[termnum].r / R ;
      } /* (while !feof(tmpf)) */
    }
  }

  /* Output section */

  /* term:N:n:R:r:cttf:tf:Rttf:rtf:qtf:qln:src:rscore:rsv */

  for ( j = 0 ; j < numterms; j++ ) {
    printf("%s:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%s:%.3f:%.3f\n", 
	   terms[j].term, N, terms[j].n, R, terms[j].r,
	   cttf, terms[j].ttf, Rtf, terms[j].rtf, terms[j].qtf, qlen,
	   terms[j].source, terms[j].rscore, terms[j].rsv) ;
  }

  fprintf(stderr,
	  "%d qterms %d added %d total %d skippedI %d skippedr %d skippedrtf %d skippedn\n",
	  numqterms, numaddedterms, numterms,
	  skippedIs, rtoosmall, rtftoosmall, ntoosmall) ;

 error:
     
  unlink(tmpfname) ; unlink(tmpf1name) ;
  iexit() ;
  exit(1) ;
}

