static char *stem_sccsid = "@(#) stem.c 1.4 01/11/23\n";

/******************************************************************************

  stem.c SW Nov 92

  Filter prog for Porter stemming

******************************************************************************/

#include <stdio.h>

int wstem_nospell(), sstem_nospell(), nostem_spell() ;
int wstem(), sstem(), nostem(), psstem() ;

int Secs ;			/* Dummy to avoid undefined */

char *usage = "Usage: %s -w[eak]|-s[trong]|-n[one] -sp [-[spel]l] [-p[hrase]]\n" ;
char *helpmes = "Must specify one of -w -s -n -sp (weak, strong, none, strict Porter resp.)\n\
-p[hrase] causes a line to be stemmed word by word rather than as a whole\n\
-spell does a little spelling conflation; it won't work very well \n\
without strong or weak stemming and is ignored if strict Porter\n\n\
Input must be in lower case. \n\
Punctuation is not handled at all except apostrophe s" ;

main(argc, argv)
     int argc ;
     char **argv ;
{
  char strength = '\0' ;
  int spell = 0, phrase = 0 ;
  int i, l, ol ;
  char *s ;
  char *res ;
  int (*func)() ;
  char inbuf[4096], obuf[4096] ;

  if ( argc < 2 ) {
    fprintf(stderr, usage, argv[0]) ;
    exit(1) ;
  }
  
  for ( i = 1 ; i < argc ; i++ ) {
    s = argv[i] ;
    if ( strcmp(s, "-w") == 0 || strcmp(s, "-weak") == 0 ) {
      if ( strength != '\0' ) {
	fprintf(stderr, "%s\n", helpmes) ;
	exit(1) ;
      }
      else strength = 'w' ;
    }
    else if ( strcmp(s, "-s") == 0 || strcmp(s, "-strong") == 0 ) {
      if ( strength != '\0' ) {
	fprintf(stderr, "%s\n", helpmes) ;
	exit(1) ;
      }
      else strength = 's' ;
    }
    else if ( strcmp(s, "-n") == 0 || strcmp(s, "-none") == 0 ) {
      if ( strength != '\0' ) {
	fprintf(stderr, "%s\n", helpmes) ;
	exit(1) ;
      }
      else strength = 'n' ;
    }
    else if ( strcmp(s, "-sp") == 0 ) {
      if ( strength != '\0' ) {
	fprintf(stderr, "%s\n", helpmes) ;
	exit(1) ;
      }
      else strength = 'p' ;
    }
    else if ( strcmp(s, "-l") == 0 || strcmp(s, "-spell") == 0 ) 
      spell = 1 ;
    else if ( strcmp(s, "-p") == 0 || strcmp(s, "-phrase") == 0 ) 
      phrase = 1 ;
    else if ( strcmp(s, "-?") == 0 || strcmp(s, "-help") == 0 ) {
      fprintf(stderr, usage, argv[0]) ;
      fprintf(stderr, "%s\n", helpmes) ;
      exit(0) ;
    }
    else {
      fprintf(stderr, "Unknown flag %s\n", s) ;
      fprintf(stderr, usage, argv[0]) ;
      fprintf(stderr, "%s\n", helpmes) ;
      exit(1) ;
    }
  }
  
  if ( strength == 'w' ) {
    if ( spell ) func = wstem ;
    else func = wstem_nospell ;
  }
  else if ( strength == 's' ) {
    if ( spell ) func = sstem ;
    else func = sstem_nospell ;
  }
  else if ( strength == 'n' ) {
    if ( spell ) func = nostem_spell ;
    else func = nostem ;
  }
  else if ( strength == 'p' ) {
    func = psstem ;
  }
  else {
    fprintf(stderr, helpmes) ;
    exit(1) ;
  }

  while ( (res=gets(inbuf)) != NULL ) {
    l = strlen(inbuf) ;
    if ( phrase == 0 ) ol = func(inbuf, l, obuf) ;
    else ol = stem_phrase_token(inbuf, l, obuf, func, NULL) ;
    while ( ol > 0 && obuf[ol - 1] == ' ' ) ol-- ;
    obuf[ol] = '\0' ;
    (void) puts(obuf) ;
  }
  exit(0) ;
}

      

