/******************************************************************************

  text_filter.c SW Oct 96
  generalises hi_bit_off (which is now a script which execs text_filter +old)

  Exit 0 if no control or high characters, 2 if arg error, otherwise 1

******************************************************************************/
#include <stdio.h>
char *usage =
"Usage: %s [+|-maskhi] [+|-skiplo] [+|-counthi] [+|-countlo] [+|-verbose]\n\
   [+|-old]\n\
Filter which can remove, mask and/or count non-ASCII characters on stdin.\n\
'maskhi' (default off) masks out the hi bit, \n\
'skiphi' (default off) skips chars with hi bit set,\n\
'skiplo' (default off) prevents output of control chars,\n\
'no_out' (default off) sends output to /dev/null\n\
'verbose' (default off) reports after every megabyte of input\n\
'counthi' and 'countlo' (defaults off) count occurrence of each non-ASCII\n\
character (it counts and reports totals anyway).\n\
'old' (default off) sets maskhi and skiplo and changes FF and CR to LF\n\n\
Diagnostic output is to stderr.\n" ;

main(argc, argv)
     int argc ;
     char **argv ;
{
  int ch, hicount = 0, locount = 0, ccount = 0 ;
  int counts[256] ;
  int old = 0 ;
  int counthi = 0, countlo = 0 ;
  int maskhi = 0 ;
  int skiplo = 0, skiphi = 0 ;
  int no_out = 0 ;
  int verbose = 0 ;
  int j ;
  char *p ;

  for ( j = 1 ; j < argc ; j++ ) {
    if ( strcmp((p = argv[j]), "-old") == 0 ) old = 0 ;
    else if ( strcmp(p, "+old") == 0 ) old = 1 ;
    else if ( strcmp(p, "+counthi") == 0 ) counthi = 1 ;
    else if ( strcmp(p, "-counthi") == 0 ) counthi = 0 ;
    else if ( strcmp(p, "+countlo") == 0 ) countlo = 1 ;
    else if ( strcmp(p, "-countlo") == 0 ) countlo = 0 ;
    else if ( strcmp(p, "+maskhi") == 0 ) maskhi = 1 ;
    else if ( strcmp(p, "-maskhi") == 0 ) maskhi = 0 ;
    else if ( strcmp(p, "+skiphi") == 0 ) skiphi = 1 ;
    else if ( strcmp(p, "-skiphi") == 0 ) skiphi = 0 ;
    else if ( strcmp(p, "+skiplo") == 0 ) skiplo = 1 ;
    else if ( strcmp(p, "-skiplo") == 0 ) skiplo = 0 ;
    else if ( strcmp(p, "+no_out") == 0 ) no_out = 1 ;
    else if ( strcmp(p, "-no_out") == 0 ) no_out = 0 ;
    else if ( strcmp(p, "+verbose") == 0 ) verbose = 1 ;
    else if ( strcmp(p, "-verbose") == 0 ) verbose = 0 ;
    else {
      fprintf(stderr, usage, argv[0]) ;
      exit(2) ;
    }
  }    
  if ( counthi || countlo ) memset(counts, '\0', sizeof(counts)) ;
  if ( old ) {
    maskhi = 1 ;
    skiplo = 1 ;
  }
  while ( (ch = getchar()) != EOF ) {
    ccount++ ;
    if ( verbose && (ccount % (1024 * 1024) == 0) )
      fprintf(stderr, "High bit set: %d, control %d, total chars %d\n",
	      hicount, locount, ccount) ;
    if ( ch >= 0x80 ) {
      hicount++ ;
      if ( counthi ) counts[(unsigned)ch]++ ;
      if ( ! skiphi ) {
	if ( maskhi ) ch &= 0x7f ;
	if ( ! no_out ) putchar(ch) ;
      }
    }
    else if ( ch < 0x20 ) {
      switch (ch) {
      case 0x0a:
      case 0x0d:
      case 0x0c:
	if ( old ) ch = 0x0a ;
	if ( ! no_out ) putchar(ch) ;
	break ;
      case 0x09:
	if ( ! no_out ) putchar(ch) ;
	break ;
      default:
	locount++ ;
	if ( ! skiplo && ! no_out ) putchar(ch) ;
	if ( countlo ) counts[ch]++ ;
	break ;
      }
      
    }
    else if ( ! no_out ) putchar(ch) ;
  }
  if ( old )
    fprintf(stderr,
	    "hi_bit_off: %d character(s) altered, %d control chars skipped\n",
	    hicount, locount) ;
  else {
    fprintf(stderr, "High bit set: %d, control %d, total chars %d\n",
	    hicount, locount, ccount) ;
  }
  if ( countlo ) {
    for ( j = 0 ; j < 0x20 ; j++ )
      if ( counts[j] ) fprintf(stderr, "%d (%d) ", j, counts[j]) ;
    if ( locount ) fprintf(stderr, "\n") ;
  }
  if ( counthi ) {
    for ( j = 0x80 ; j < 0x100 ; j++ )
      if ( counts[j] ) fprintf(stderr, "%d (%d) ", j, counts[j]) ;
    if ( hicount ) fprintf(stderr, "\n") ;
  }
  if ( hicount || locount ) exit(1) ;
  else exit(0) ;
}

