#!/usr/bin/perl
#
#
#     topic_negations.pl
#
#
#     Tries to find and remove negated sections in TREC topic narratives.
#
#     Output is topics in standard format, possibly with an additional
#     <neg> tag (not terminated) containing the negated bits.
#
#     Beware:  this messes around with sentence structure --
#     neither <neg> bits nor affected sentences in the narratives
#     will read like proper sentences.
#
#  ser 24/6/99
#

$\ = "\n";

if ($ARGV[0] =~ /h/) {
  print STDERR ;
  print STDERR 'Usage:  raw topics to standard input;
        standard output is modified topics.

        standard error output has topic numbers of changed topics,
           plus the full original and new versions 
           if the first argument contains l (ell).
        ';
#        .log2 file contains details of changed sentences
#           (this file is cumulative -- empty it occasionally!).
  die "\n";
};

if ($ARGV[0] =~ /l/) {$full_log = 1;} else {$full_log = 0;};
undef $/;
@abb1=();@abb2=();$nabb=0;


#  Abbreviations in file in form "abbreviation,expansion\nabb,exp..."
open(ABBREV,"abbreviations") 
  || open(ABBREV,"$ENV{'OKAPI'}/lib/abbreviations") 
  || die "No abbreviations file\n";
$abbrevs = <ABBREV>;
close ABBREV;
foreach $s (split(/\n/,$abbrevs)) {
  if (!($s=~/^#/)) {;
    ($abb1[$nabb],$abb2[$nabb]) = (split(/,/,$s));
    $abb1[$nabb] = quotemeta($abb1[$nabb]);
    $nabb++;
  };
};

#  Subroutine replace_abbrevs
#
#  use the list provided
sub replace_abbrevs {

  local($str) = @_;
  
  for($i=0;$i<$nabb;$i++) {
    if (substr($abb1[$i],-1) ne ".") {
	  $str =~ s/$abb1[$i]\.(,|-| -|(\W*[a-z]))/$abb2[$i]\1/g;    #  abbreviation within sentence
      $str =~ s/$abb1[$i]\./$abb2[$i]./g;             #  all others
	} else {
	  $str =~ s/$abb1[$i]/$abb2[$i]/g;            # more liberal version -- assume all within sentences
	};
  };
  return $str;

};

#open(LOG1,"> topic_negations.log1");
#open(LOG2,">> topic_negations.log2");



#  Basic patterns required
$nr_pat = "(not (be )?(considered )?relevant|irrelevant|non-?relevant|not constitute relevance)";
$doc_pat = "(document|item)s?";
$dneg = "not?|without|unless|rather than|outside of|free from";

#  TNR is a terminal (end of sentence or before a comma) indication of non-relevance
#$TNR = " $nr_pat(,|\\Z)";
#  actually this version finds it anywhere (not already found by INR).
$TNR = " $nr_pat";

#  INR is an initial (beginning of sentence) indication of non-relevance
$INR = "((an? )?($doc_pat )?|it )(is ($nr_pat ))";
$INR .= "|((an? )?($doc_pat )?|it )(((should |will )not )(be about |focus on ))";
$INR .= "|(an? )?$nr_pat $doc_pat ";
$INR .= "|(also )?not relevant,? ";
$INR = "^($INR)";



$data = <STDIN>;         #  Topics file to standard input -- read all at once

#  Extract topic from $data, removing it from $data at the same time
#

while ((($dummy,$data,$top)=&section($data,"top",0))[0]) {
  
  $newtop = &replace_abbrevs($top);
  
  $topno = (&section_ie($newtop,"num",0))[2];
  $topno =~ s/number: *//is;
  $neg = ""; $changed = 0;

  #  Extract narrative from topic, remembering where it was
  #
  ($narr,$pos,$len) = (&section_ie($newtop,"narr",0))[2,4,5];
  $narr =~ s/^narrative:\s+//is;
  $narr =~ s/(\w) *\n *\n/$1. /gs;
  $narr =~ s/\s+/ /gs;
  $newnarr = "";

  #  Extract sentence from narrative
  while ((($sentence,$narr) = split(/[;.]/,$narr,2))[0]) {
    $narr =~ s/^\s+//; $newsentence = "";
	$hold = $sentence;
    
    #  Find INR or TNR in sentence, and identify its scope.
	#  Anything within its scope goes to $neg, anything else to $newsentence.
    if ($sentence =~ s/($INR)/\*INR\* /i) {
      $nbit = $1; $changed = 1;
	  
	  if ($sentence =~ s/INR\*(.*)\W($dneg)\W(.*?,)/\*INR$1,/) { $newsentence = $3; };
	  
	  $neg .= "$sentence.\n";
	  
#	  print LOG2 "$topno: $nbit:\nOriginal: $hold\nNegative: $sentence\nKeep: $newsentence\n";

    } elsif ($sentence =~ s/($TNR)/\ *TNR\*/i) {
      $nbit = $1; $changed = 1;
	  
	  if ($sentence =~ s/(^.*?relevant.*?,)//i) { $newsentence = $1; };
	  if ($sentence =~ s/^($dneg)\W(.*?,)//i) { $newsentence = $2; };
	  if ($sentence =~ s/\W($dneg)\W(.*?)(,| \*TNR)/$3/) { $newsentence = $2; };
	  $sentence =~ s/\*TNR\*(: |,? nor )(.*)/, $2 \*TNR\*/;
	  if ($sentence =~ s/TNR\*(.*\w.*)$/TNR\*/) { $newsentence = $1; };
	  
	  $neg .= "$sentence.\n";
	  
#	  print LOG2 "$topno: $nbit:\nOriginal: $hold\nNegative: $sentence\nKeep: $newsentence\n";

    } else {
	  $newsentence = $sentence;
	};


    #  Reconstruct narrative   (has it got something real left in it?)
    if ($newsentence =~ /\w/) { $newnarr .= "$newsentence.\n"; };
  };


  #  Reconstruct topic
  #
  $newnarr = "<narr> Narrative:\n\n$newnarr";
  if ($neg) { $neg = "<neg> Negative:\n$neg"; };
  substr($newtop,$pos,$len) = "$newnarr\n$neg\n\n";
  $newtop = "<top>\n\n$newtop\n</top>\n\n";
  $top = "<top>\n\n$top\n</top>\n\n";
  
  print STDOUT "$newtop";    #  Reconstructed topics to standard output
# if ($changed) { print STDERR "$topno"; };
  if ($changed * $full_log) { print STDERR "\n$top\n$newtop\n"; };   # for testing

};

#close LOG1;
#print LOG2 "-----------------------\n";
#close LOG2;


#  (this is a general-purpose subroutine, much more elaborate
#  than is required for TREC data).
#
#  Subroutine section:
#
#  extract everything from <xxx> to </xxx>,
#  return whether found or not,
#  the original string with that section removed,
#  the contents of that section (empty if none),
#  any parameters in the opening tag,
#  the position of the extracted section in the original string,
#  and its length in the original string (including tags).
#
sub section {
  
  local($rec,$secname,$spos) = @_;
  local($found) = 0;
  local($contents) = ""; local($pars) = "";
  local($pos) = 0; local($length) = 0;

  if ($rec =~ s:^(.{$spos,}?)(<$secname(>|(\s+(.*?(\".*?\".*?)*?))>)(.*?)</$secname>):$1:is) {
    $contents = $7; $pars = $5;
    $pos = length($1); $length = length($2);
    $found = 1;
  };
  $contents =~ s/^\s*//; $contents =~ s/\s*$//;
  return ($found,$rec,$contents,$pars,$pos,$length);

};


#  Subroutine section_ie:
#
#  like section except that it looks for a section which is
#  implicitly ended by the start of the next section
#  (or the end of the record)
#  -- much simpler (no parameters, no embedded tags),
#  but imitating section in its arguments and returns
#
sub section_ie {

  local($rec,$secname,$spos) = @_;
  local($found) = 0;
  local($contents) = ""; local($pars) = "";
  local($pos) = 0; local($length) = 0;

  if ($rec =~ s:^(.{$spos,}?)(<$secname>(.*?))($|<.*?>):$1:is) {
    $contents = $3;
    $pos = length($1); $length = length($2);
    $found = 1;
  };
  $contents =~ s/^\s*//; $contents =~ s/\s*$//;
  return ($found,$rec,$contents,$pars,$pos,$length);

};
