/******************************************************************************

  expd.c from

  expand.c SW July 95 taken from

  expand.c SW Nov 93

  Cut down and modified version of the trec3 auto expansion prog
  ~okapi/src/trec3/expand

  It takes as input a file of query terms and a number, searches and retrieves
  that number of docs, forms a set from these, extracts terms from them (or,
  if op is bm250, from the best passage of each doc. All terms, including the
  originals are then output as below.

  Input is 

  o a query term file in the format produced by get_qterm;

      <qnum>:<qlength>:<source>:<nterms>:<term>:[<term>:]<qtf>

      (It doesn't cope with <nterms> being other than 1)
      <source> is "doc" and <qtf> 0 if term not from query

  o number of docs to retrieve initially (Nrels);

  o k3 for initial feedback search

  Output is

     <qnum>:<qlen>:<source>:<term>:<qtf>:<n>:<r>:<w0>:<F4>\
        :<init wt>:<fin wt>:<rsv>

******************************************************************************/

/* #define PASS_K1 2.0
#define PASS_K2 0.0
#define PASS_BM25B 0.6
*/
#define PASS_K1 1.2
#define PASS_K2 0.0
#define PASS_BM25B 0.75
#define PASSAGE_TOPNUM 10000

#undef DEBUG
/*#define NOSORT*/
#undef NOSORT
#define OUTTERMS

#define MAXTERMS 50000
#define MAXQTERMS 500
#define RSV_CUTOFF 0
#define I0q if (res = i0(Command, Response) < 0) {\
  sprintf(Command, "perror") ;\
  i0(Command, Response) ;\
  fprintf(stderr, Response) ;\
  goto error ; }

#ifdef DEBUG
#define I0 fprintf(stderr, "%s\n", Command) ;\
if (res = i0(Command, Response) < 0) {\
  sprintf(Command, "perror") ;\
  i0(Command, Response) ;\
  fprintf(stderr, Response) ;\
  goto error ; }\
  else fprintf(stderr, Response) ;
#else
#define I0 I0q
#endif     

#define I0q_ if (res = i0(command, response) < 0) {\
  sprintf(command, "perror") ;\
  i0(command, response) ;\
  fprintf(stderr, response) ;\
  goto error ; }

#ifdef DEBUG
#define I0_ fprintf(stderr, "%s\n", command) ;\
if (res = i0(command, response) < 0) {\
  sprintf(command, "perror") ;\
  i0(command, response) ;\
  fprintf(stderr, response) ;\
  goto error ; }\
  else fprintf(stderr, response) ;
#else
#define I0_ I0q_
#endif     


static int cpterm() ;
#ifdef DEBUG
int Debug = 1 ;
#else
int Debug = 0 ;
#endif

struct term {
  int qnum ;
  int qlen ;
  char source[5] ;
  char *term ;
  int qtf ;
  int r ;
  float w0 ;
  float w1 ;
  float initial_weight ;
  float final_weight ;
  float rsv ;
  int np ;
  int set ;
} ;

#include "bss.h"

extern struct db Cdb ;

char *info = 
"An initial search is done on the output from get_qterms on stdin. \n\
This uses the environment variable BSS_DATABASE. BSS_ATTRIBUTE, BSS_OP,\n\
BSS_LIMIT, BSS_K1 and BSS_BM25B are also used if available, otherwise \n\
default to 'default', bm2500, 0, 2.0 and 0.6 resp. The K3 value must be\n\
given as argv[2]. Env variables BSS_RLOAD and BSS_BIGRLOAD will be used\n\
if available, otherwise they default to 0 respectively. \n\
The top argv[1] docs from this search become the 'intial set'.\n\n\
Terms are then extracted from the docs of the initial set, everything is\n\
reweighted using, in the case of the query terms, argv[3] as K3.\n" ;


char Command[1024 * 1024 * 8] ;
char Response[1024 * 1024 * 8] ;
char Op[10] = "bm2500" ;

main(int argc, char **argv)
{
  int Nrels ;			/* Number of "rel" docs wanted */
  int nrels ;			/* Number obtained */
  int Numexpterms = 0 ;		/* Number of terms requested from extract */
  int numexpterms ;		/* Number got from extract */
  int numaddedterms = 0 ;	/* No of non-queryterms from extract */
  int numqterms = 0 ;
  int numterms = 0 ;
  int num_final_terms ;
  int qnum, qlen ;
  int limit ;
  double Initial_qtf_const = 1000 ;
  double Final_qtf_const = 8 ;
  int N ;
  FILE *termfile ;
  FILE *tmpf, *tmpf1 ;
  char tmpfname[256], tmpf1name[256] ;
  float weight ;
  char buf[1024], term[1024] ;
  BOOL found ;
  int set, initial_set ;
  int relset ;
  int j, k ;
  float x ;
  float k1 = 2.0, k2 = 0.0, b=.6 ;
  int bss_debug = 0 ;
  char *envp ;
  char attr[128] ;
  int rload = 0, bigrload = 0 ;
  struct term terms[MAXTERMS] ;
  int res ;
  int target ;
  int passage_sets[MAXQTERMS] ;
  int p_unit = 1, p_step = 1, p_maxlen = 20 ;
  int passages = 0 ;		/* If true extracts from best passages */
  char *p ;

  if ( argc < 4 ) {
    fprintf(stderr,
	    "Usage: %s <numrels> <initial k3> <final k3>\n",
	    argv[0]) ;
    fprintf(stderr,
   "Must have BSS_DB and optionally BSS_ATTRIBUTE and BSS_LIMIT in envt.\n") ;
    fprintf(stderr,
	    "BSS_OP, BSS_K1 and BSS_BM25B may also be set\n") ;
    fprintf(stderr, info) ;
    exit(1) ;
  }
  limit = 0 ;
  Nrels = atoi(argv[1]) ;
  Initial_qtf_const = atof(argv[2]) ;
  Final_qtf_const = atof(argv[3]) ;

  sprintf(tmpfname, "/tmp/tmpf.%d", getpid()) ;
  tmpf = fopen(tmpfname, "w+") ;
  if ( tmpf == NULL ) {
    fprintf(stderr, "Can't open %s\n", tmpfname) ;
    exit(1) ;
  }

  strcpy(attr, "default") ;
  iinit() ;
  if ( (envp = getenv("BSS_DB")) != NULL && *envp != '\0' ) {
    sprintf(Command, "ch %s\n", envp) ;
    I0 ;
    fprintf(stderr, "Using db %s\n", envp) ;
  }
  else {
    fprintf(stderr, "No database\n") ;
    goto ierror ;
  }
  if ( (envp = getenv("BSS_LIMIT")) != NULL && *envp != '\0' ) {
    sprintf(Command, "set l=%s\n", envp) ;
    I0
  }
  if ( (envp = getenv("BSS_ATTRIBUTE")) != NULL && *envp != '\0' ) {
    sprintf(Command, "set a=%s\n", envp) ;
    I0 ;
    strcpy(attr, envp) ;
  }
  if ( (envp = getenv("BSS_RLOAD")) != NULL && *envp != '\0' ) {
    rload = atoi(envp) ;
  }
  if ( (envp = getenv("BSS_BIGRLOAD")) != NULL && *envp != '\0' ) {
    bigrload = atoi(envp) ;
  }
  if ( (envp = getenv("BSS_OP")) != NULL && *envp != '\0' ) {
    strcpy(Op, envp) ;
  }
  if ( (envp = getenv("BSS_K1")) != NULL && *envp != '\0' ) {
    k1 = atof(envp) ;
  }
  if ( (envp = getenv("BSS_K2")) != NULL && *envp != '\0' ) {
    k2 = atof(envp) ;
  }
  if ( (envp = getenv("BSS_BM25B")) != NULL && *envp != '\0' ) {
    b = atof(envp) ;
  }
  if ( envp = getenv("EXTRACT_PASSAGES") ) passages = 1 ;
  if ( (envp = getenv("BSS_P_UNIT")) != NULL && *envp != '\0' ) {
    p_unit = atoi(envp) ;
  }
  if ( (envp = getenv("BSS_P_STEP")) != NULL && *envp != '\0' ) {
    p_step = atoi(envp) ;
  }
  if ( (envp = getenv("BSS_P_MAXLEN")) != NULL && *envp != '\0' ) {
    p_maxlen = atoi(envp) ;
  }
  if ( (envp = getenv("BSS_DEBUG")) != NULL && *envp != '\0' ) {
    bss_debug = atoi(envp) ;
    sprintf(Command, "set debug=%d\n", bss_debug) ;
    I0 ;
  }

  fprintf(stderr, "Using attribute %s\n", attr) ;
  fprintf(stderr, "Using rload %d\n", rload) ;
  fprintf(stderr, "Using bigrload %d\n", bigrload) ;
  fprintf(stderr, "Using op %s\n", Op) ;
  fprintf(stderr, "Using k1 %.2f\n", k1) ;
  fprintf(stderr, "Using k2 %.2f\n", k2) ;
  fprintf(stderr, "Using bm25b %.2f\n", b) ;
  if ( passages ) {
    fprintf(stderr, "Using passage topnum %d\n", PASSAGE_TOPNUM) ;
    fprintf(stderr, "Using passage k1 %f\n", PASS_K1) ;
    fprintf(stderr, "Using passage k2 %f\n", PASS_K2) ;
    fprintf(stderr, "Using passage bm25b %f\n", PASS_BM25B) ;
    fprintf(stderr, "Using passage unit %d\n", p_unit) ;
    fprintf(stderr, "Using passage step %d\n", p_step) ;
    fprintf(stderr, "Using passage maxlen %d\n", p_maxlen) ;
  }

  /* Form query termsets */
  if ( Debug ) fprintf(stderr, "Looking up query terms\n") ;
  /* Should do something more elaborate if bm250 */
     while ( ! feof(stdin) ) {
    if ( gets(buf) == NULL ) break ;
    if ( numterms >=MAXTERMS ) {
      fprintf(stderr, "Sorry, too many terms\n") ;
      goto error ;
    }
    sscanf(buf, "%d:%d:%[^:]:%*[^:]:%[^:]:%d",
	   &qnum, &qlen, terms[numterms].source, term, &terms[numterms].qtf) ;
    sprintf(Command, "f t=%s", term) ;
    I0 ;
    terms[numterms].np = atoi(strstr(Response, "np=") + 3) ;
    if ( terms[numterms].np == 0 ) {
      fprintf(stderr, "Warning: term %s not found\n", term) ;
      continue ;
    }
    terms[numterms].set = atoi(Response + 1) ;
    
    terms[numterms].term= strdup(term) ;
    sprintf(Command, "w wtfunc=0 n=%d", terms[numterms].np) ;
    I0 ;
    terms[numterms].w0 = atof(Response) ;
    terms[numterms].initial_weight = terms[numterms].w0 ; 
    if ( Initial_qtf_const > 0.001 ) {
      /* If <= 0 no qtf factor */
      x = terms[numterms].initial_weight * terms[numterms].qtf *
	(Initial_qtf_const + 1.0) /
	  (terms[numterms].qtf + Initial_qtf_const) ;
      terms[numterms].initial_weight = x ;    }

    numterms++ ;
    numqterms++ ;
  }
  fprintf(stderr, "Got %d query terms\n", numqterms) ;
  fprintf(stderr, "Forming initial feedback set\n") ;
  p = Command ;
  sprintf(p, "f") ; p += strlen(p) ;
  if ( passages ) target = PASSAGE_TOPNUM ;
  else target = Nrels ;
  for ( j = 0 ; j < numqterms ; j++ ) {
    sprintf(p, " s=%d w=%f", terms[j].set, terms[j].initial_weight) ;
    p += strlen(p) ;
  }
  sprintf(p, " op=%s k1=%f k2=%f bm25b=%f target=%d", Op, k1, k2, b, target) ;
  I0 ;
  initial_set = atoi(Response + 1) ;
  sprintf(Command, "f s=%d top=%d", initial_set, target) ;
  I0 ;
  if ( passages ) {
    fprintf(stderr, "Cutting down the termsets\n") ;
    initial_set = atoi(Response + 1) ;
    for ( j = 0 ; j < numqterms ; j++ ) {
      sprintf(Command, "f s=%d s=%d op=and2", terms[j].set, initial_set) ;
      I0 ;
      passage_sets[j] = atoi(Response + 1) ;
    }
    fprintf(stderr, "Setting up the passage search\n") ;
    p = Command ;
    sprintf(p, "f") ; p += strlen(p) ;
    target = Nrels ;
    for ( j = 0 ; j < numqterms ; j++ ) {
      sprintf(p, " s=%d w=%f", passage_sets[j], terms[j].initial_weight) ;
      p += strlen(p) ;
    }
    sprintf(p, " op=%s k1=%f k2=%f bm25b=%f p_step=%d p_unit=%d p_maxlen=%d target=%d",
	    "bm250", PASS_K1, PASS_K2, PASS_BM25B, p_step, p_unit, p_maxlen,
	    target) ;
    I0 ;
    fprintf(stderr, "%s\n", Response ) ; /* debug */
    fprintf(stderr, "Done the passage search\n") ;
    initial_set = atoi(Response + 1) ;
    sprintf(Command, "f s=%d top=%d", initial_set, target) ;
    I0 ;
  } /* passages */

  nrels = atoi(strstr(Response, "np=") + 3) ;
  relset = atoi(Response + 1) ;
  fprintf(stderr, "Reweighting the %d qterms on these %d recs\n",
	  numqterms, nrels) ;
  for ( j = 0 ; j < numqterms ; j++ ) {
    sprintf(Command, "f s=%d s=%d op=and2 save=0", terms[j].set, relset) ;
    I0 ;
    terms[j].r = atoi(Response + 3) ;
    if ( terms[j].r == 0 ) fprintf(stderr, "Term %s not in relset\n",
				   terms[j].term) ;
    sprintf(Command, "w wtfunc=2 n=%d r=%d bigr=%d rload=%d bigrload=%d",
	    terms[j].np, terms[j].r, nrels, rload, bigrload) ;
    I0 ;
    terms[j].w1 = atof(Response) ;
  }

  {
    int spurious = 0 ;
    int totterms = 0, termsthis ;
    char *ip ;
    int j, nk ;
    sprintf(tmpf1name, "/tmp/tmpf1.%d", getpid()) ;
    sprintf(tmpfname, "/tmp/tmpf.%d", getpid()) ;
    for ( j = 0 ; j < nrels ; j++ ) {
      tmpf = fopen(tmpfname, "w+") ;
      if ( passages ) {
	int messl, sod, p_oset, p_length ;
	int irn ;
	sprintf(Command, "s f=3 s=%d r=%d", relset, j) ;
	I0 ;
	p = Response ;
	sscanf(Response + 56, "%10d", &irn) ;
	fprintf(stderr, "IRN=%d\n", irn) ;
        sscanf(p, "%8d", &messl) ;
	sscanf(p + 86, "%6d", &sod) ;
	p += 92 ;
	p += 3 + 19 * Cdb.nf + 6 + 10 ;
	sscanf(p, "%8d%8d", &p_oset, &p_length) ;
	p = Response + sod + p_oset ;
        if ( Response + p_oset + p_length > Response + messl ) {
          fprintf(stderr,
	  "**Crappy passage record, messl=%d, p_oset=%d, p_length=%d\n",
                  messl, p_oset, p_length) ;
	  fprintf(stderr, "Skipping this record\n") ;
          continue ;
        }
	if ( Debug ) fprintf(stderr, "The good bit is\n**%.*s\n", p_length, p) ;
	else fprintf(stderr, "The good bit is at offset %d, length %d\n",
		     p_oset, p_length) ;
	nk = bss_superparse(p, p_length, attr, Command) ;
      }
      else {
	sprintf(Command, "s f=200 s=%d r=%d", relset, j) ;
	I0 ;
	nk = bss_superparse(Response, strlen(Response), attr, Command) ;
      }
      if ( nk < 0 ) {
	fprintf(stderr, "Can't parse rec %d (length %d)\n",
		j, strlen(Response)) ;
	goto ierror ;
      }
      /* Otherwise write the terms to a tempfile */
      ip = Command ;
      termsthis = spurious = 0 ;
      while ( nk-- ) {
	if ( *ip == 'H' || *ip == 'F' ||
            strlen(ip + 1) == 0 || strlen(ip) == 0) {
          ip += strlen(ip) + 1 ;
          ip += strlen(ip) + 1 ;
          continue ;
        }
        else if ( *ip != ' ' && *ip != 'S' && *ip != 'G' && *ip != 'I' ) {
          spurious++ ;
          ip += strlen(ip) + 1 ;
          ip += strlen(ip) + 1 ;
          continue ;
        }
	if ( *ip == ' ' ) *ip = 'N' ;
	fprintf(tmpf, "%s\n", ip) ; /* First char is GSL code I think */
	termsthis++ ; totterms++ ;
        ip += strlen(ip) + 1 ;
        ip += strlen(ip) + 1 ;
      }
      fclose(tmpf) ;
      sprintf(Command,
 "/usr/bin/sort < %s | /usr/bin/uniq >> %s\n",
	      tmpfname, tmpf1name) ;
      system(Command) ;
      fprintf(stderr, "Extracted %d terms from %d of %d\n",
	      termsthis, j, nrels) ;
      if ( spurious ) fprintf(stderr, "-- %d spurious skipped\n", spurious) ;
    }
    sprintf(Command,
  "/usr/bin/sort < %s | /usr/bin/uniq -c | /usr/bin/sort +0nr -1 > %s\n",
	    tmpf1name, tmpfname) ;
    system(Command) ;
    /* Now have tmpf containing all the terms uniquely, each preceded by
       little r. */

    {
      char buf[256] ;
      char p[256] ;
      char code ;
      int r ;
      tmpf = fopen(tmpfname, "r") ;
      while ( ! feof(tmpf) ) {
	if ( numterms >= MAXTERMS ) {
	  fprintf(stderr, "Too many terms\n") ;
	  break ;
	}
	fgets(buf, 256, tmpf) ;
	if ( feof(tmpf) ) break ;
	sscanf(buf, "%d %c%s", &r, &code, p) ;
	for ( k = 0, found = FALSE ; k < numqterms ; k++ ) {
	  if ( strcmp(p, terms[k].term) == 0 ) {
	    found = TRUE ;
	    fprintf(stderr, "Found %s\n", p) ;
	    break ;
	  }
	}
	
	if ( ! found ) {		/* Add it */
	  char response[1024], command[1024] ;
	  if ( code == 'I' ) {
	    /*fprintf(stderr, "Skipping I-term %s\n", p) ;*/
	    continue ;
	  }
	  terms[numterms].term = strdup(p) ;
	  sprintf(command, "f save=0 t=%s", terms[numterms].term) ;
	  I0_ ;
	  terms[numterms].set = -1 ;
	  terms[numterms].np = atoi(strstr(response, "np=") + 3) ;
	  terms[numterms].r = r ;
	  sprintf(command, "w wtfunc=1 n=%d r=%d bigr=%d",
		  terms[numterms].np, r, nrels) ;
	  I0_ ;
	  terms[numterms].w1 = atof(response) ;
	  sprintf(command, "w wtfunc=0 n=%d", terms[numterms].np) ;
	  I0_ ;
	  terms[numterms].w0 =
	    terms[numterms].initial_weight = atof(response) ; 
	  terms[numterms].qtf = 0 ;
	  sprintf(terms[numterms].source, "doc") ;
	  numterms++ ;
	  numaddedterms++ ;
	  /* fprintf(stderr, "Added %s, term %d\n", p, numterms) ; */
	}
      }
    }
  }

  /* Calc final termweight & rsv */
  for ( j = 0 ; j < numterms ; j++ ) {
    terms[j].final_weight = terms[j].w1 ;
    if ( Final_qtf_const > 0.001 ) {
      /* If <= 0 no qtf factor */
      if ( terms[j].qtf > 0 )
	x = terms[j].final_weight * terms[j].qtf * (Final_qtf_const + 1.0)
	  / (terms[j].qtf + Final_qtf_const) ;
      else
	x = terms[j].final_weight ;
      terms[j].final_weight = x ;
    }
    terms[j].rsv =
      (terms[j].final_weight * terms[j].r / (float)nrels) ;
    if ( terms[j].rsv < 0 ) terms[j].rsv = 0 ;
  }
  

  /* Sort */

#ifndef NOSORT
  gqsort((char *)terms, numterms, sizeof(struct term), cpterm) ;
#endif
  fprintf(stderr, "numterms=%d\n", numterms) ;
  for ( j = 0 ; j < numterms ; j++ ) {
    if ( terms[j].rsv < RSV_CUTOFF ) continue ;
    printf("%d:%d:%s:%s:%d:%d:%d:%f:%f:%f:%f:%f\n", qnum, qlen,
	   terms[j].source, terms[j].term, terms[j].qtf,
	   terms[j].np, terms[j].r, terms[j].w0, terms[j].w1,
	   terms[j].initial_weight, terms[j].final_weight, terms[j].rsv) ;
  }

 ierror:
 error:
     
  unlink(tmpfname) ; unlink(tmpf1name) ;
  iexit() ;
  exit(1) ;
}

static int cpterm(struct term *t1, struct term *t2)
{
  float i ;
  i = t2->rsv - t1->rsv ;
  if ( i == 0 ) i = t2->final_weight - t1->final_weight ;
  return (i>0)? 1 : (i<0)? -1 : 0 ;
}
