/******************************************************************************

  Routing 1996/7

  get_rs.c SW Dec 96

  Accepts a file of relevance records as argv[1] and term and/or pairs
  of terms records on stdin, forms sets of rels and nonrels, looks up the
  terms (in BSS_DB with BSS_ATTRIBUTE and BSS_LIMIT) and outputs

    <term>:<term>:<N>:<n>:<R>:<r>:<S>:<s>

  (For single terms the second <term> field is empty.)

  Input term records are <term>:<term>

  Relevance records are <docno>:0|1 as output by
  getrels <topnum> <disks> all

  vaguely based on rtgterms_pass2.c last rev for TREC-5 May 96 SW
  (orig June 94, rev May 95)

******************************************************************************/

#define DEBUG
/*#undef DEBUG*/

#define I0q if (res = i0(command, response) < 0) {\
  sprintf(command, "perror") ;\
  i0(command, response) ;\
  fprintf(stderr, response) ;\
  goto error ; }

#ifdef DEBUG
#define I0 fprintf(stderr, "%s\n", command) ;\
if (res = i0(command, response) < 0) {\
  sprintf(command, "perror") ;\
  i0(command, response) ;\
  fprintf(stderr, response) ;\
  goto error ; }\
  else fprintf(stderr, response) ;
#else
#define I0 I0q
#endif

#include "bss.h"

char buf[1024] ;
char command[4 * 1024 * 1024] ;
char response[6 * 1024 * 1024] ;

main(argc, argv)
     int argc ;
     char **argv ;
{
  char docno[256] ;
  char term1[256], term2[256] ;
  int status ;
  int limit ;
  int N, n, R = 0, r, S = 0, s ;
  FILE *rels ;
  FILE *f ;
  FILE *tot_tf_file ;
  int rset, nset, set, newset, setnum ;
  int i, j, k ;
  int res ;
  int numterms ;
  char *envp ;
  char *dn_attr ;
  char *dbname ;
  char *p ;
  int zero_np = 0, zero_r = 0, zero_s = 0 ;
  int not_found_docs = 0 ;

  if ( argc < 2 ) {
    fprintf(stderr,
	    "usage: %s <relfile> < term records\n", argv[0]) ;
    exit(1) ;
  }

  if ( (rels = fopen(argv[1], "r")) == NULL ) {
    fprintf(stderr, "%s: can't open relfile %s\n", argv[0], argv[1]) ;
    exit(1) ;
  }

  iinit() ;
  if ( (dbname = getenv("BSS_DB")) == NULL || *dbname == '\0' ) {
    fprintf(stderr, "%s: env var BSS_DB not set\n", argv[0]) ;
    exit(1) ;
  }
  /* Make relset and nonrelset */
  sprintf(command, "ch %s", dbname) ;
  I0 ;
  if ( (envp = getenv("BSS_LIMIT")) && *envp != '\0' ) {
    sprintf(command, "set l=%s", envp) ;
    I0 ;
  }
  if ( (envp = getenv("BSS_ATTRIBUTE")) && *envp != '\0' ) {
    sprintf(command, "set a=%s", envp) ;
    I0 ;
  }
  if ( (envp = getenv("BSS_DN_ATTRIBUTE")) && *envp != '\0' )
    dn_attr = envp ;
  else dn_attr = "dn" ;

  /* Aug 98: MEM and NOPOS now fixed - see the code - this was OK and
     fastest even on 100GB Db on 512M machine */
  sprintf(command, "d bign") ;
  I0q ;
  N = atoi(response) ;
  sprintf(command, "f a=%s t=zzzzzzzzzzzzzz", dn_attr) ;
  I0q ;
  rset = atoi(response + 1) ;
  sprintf(command, "f a=%s t=zzzzzzzzzzzzzz", dn_attr) ;
  I0q ;
  nset = atoi(response + 1) ;
  f = rels ;
  R = S = 0 ;
  while ( 1 ) {
    fscanf(f, "%s", docno) ;
    if ( feof(f) ) break ;
    fscanf(f, "%d", &status) ;
    sprintf(command, "parse a=dn t=%s", docno) ;
    I0q ;
    sscanf(strstr(response, "t=") + 2, "%s", docno) ;
    sprintf(command, "f a=%s t=%s", dn_attr, docno) ;
    I0q ;
    setnum = atoi(response + 1) ;
    if ( atoi(strstr(response, "np=") + 3) == 0 ) {
      not_found_docs++ ;
      /* doc not found, should only be because of limit */
      sprintf(command, "delete %d", setnum ) ;
      I0q ;
      continue ;
    }
    if ( status == 1 ) {
      sprintf(command, "f s=%d s=%d op=or", rset, setnum) ;
      I0q ;
      newset = atoi(response + 1) ;
      sprintf(command, "delete %d", rset) ;
      I0q ;
      rset = newset ;
      R++ ;
    }
    else {
      sprintf(command, "f s=%d s=%d op=or", nset, setnum) ;
      I0q ;
      newset = atoi(response + 1) ;
      sprintf(command, "delete %d", nset) ;
      I0q ;
      nset = newset ;
      S++ ;
    }
    sprintf(command, "delete %d", setnum) ;
    I0q ;
  }
  if ( R > 0 ) {
    sprintf(command, "f s=%d nopos=1 mem=1", rset) ;
    I0q ;
    newset = atoi(response + 1) ;
    sprintf(command, "delete %d", rset) ;
    I0q ;
    rset = newset ;
  }
  if ( S > 0 ) {
    sprintf(command, "f s=%d nopos=1 mem=1", nset) ;
    I0q ;
    newset = atoi(response + 1) ;
    sprintf(command, "delete %d", nset) ;
    I0q ;
    nset = newset ;
  }
  if ( not_found_docs )
    fprintf(stderr, "%d docs not found\n", not_found_docs) ;
  fprintf(stderr, "R=%d S=%d\n", R, S) ;  
  /* Read term records from pass1.
     These are <term>:<term> */

  numterms = 0 ;
  while ( 1 ) {
    *term1 = *term2 = '\0' ;
    scanf(" %[^:]:%[^\n]", term1, term2) ;
    if ( feof(stdin) ) break ;
    if ( *term1 == '\0' ) { 
      fprintf(stderr, "%s: empty term\n", argv[0]) ;
      continue ;
    }
    if ( ! isalnum(*term1) && *term1 != '@' ) {
      fprintf(stderr, "%s: spurious term '%s'\n", argv[0], term1) ;
      continue ;
    }
    /* sprintf(command, "f nopos=1 t=%s", term1) ; */
    sprintf(command, "f nopos=0 mem=1 t=%s", term1) ;
    I0q;
    n = atoi(strstr(response, "np=") + 3) ;
    setnum = atoi(response + 1) ;
    if ( n == 0 ) {
      zero_np++ ;
      fprintf(stderr, "'%s' no postings\n", term1) ;
      sprintf(command, "delete %d", setnum) ;
      I0q ;
      continue ;
    }
    numterms++ ;
    if ( R > 0 ) {
      sprintf(command, "f s=%d s=%d save=0 op=and2 mem=1 nopos=1", setnum, rset) ;
      I0q ;
      r = atoi(response + 3) ;
    }
    else r = 0 ;
    if ( S > 0 ) {
      sprintf(command, "f s=%d s=%d save=0 op=and2 mem=1 nopos=1", setnum, nset) ;
      I0q ;
      s = atoi(response + 3) ;
    }
    else s = 0 ;
    sprintf(command, "delete %d", setnum) ;
    I0q ;
    if ( r == 0 ) zero_r++ ;
    if ( s == 0 ) zero_s++ ; 
    printf("%s:%s:%d:%d:%d:%d:%d:%d\n",
	   term1, term2, N, n, R, r, S, s) ;
  }
  iexit() ;
  fprintf(stderr, "%s: %d terms\n", argv[0], numterms) ;
  if ( zero_np ) fprintf(stderr, "Zero np: %d\n", zero_np) ;
  if ( zero_r ) fprintf(stderr, "Zero r: %d\n", zero_r) ;
  if ( zero_s ) fprintf(stderr, "Zero s: %d\n", zero_s) ;
  exit(0) ;

error:
  iexit() ;
  exit(1) ;

}

