/******************************************************************************

  rtg_select.c SW June 95

  & many mods and hackups since (June 96, June 97)

  Implementation of the new procedure for iteratively selecting routing
  terms from a candidate list, using the Brookes measure (see SER working
  paper of June 95). Many other scoring methods added since.

  Input: argv[1] file of docnos of rels
         argv[2] file of <term>:<weight>

  Output: <term>:<weight>:<increase in score>:<term number>
  June 96 added <origweight> to end of each output record.
  (June 97) The third output field is useless.

  Depends greatly on environment variables. See below.

  Writes a log to stderr and output to stdout.

******************************************************************************/

#include "rtg_select.h"
#include <signal.h>
#define SHORT_MESS

#define REMAKE_THRESH 0
#define VW_THRESH 0
#define VWUP .5
#define VWDOWN .33
#define VWUP_MIN .15
#define MAX_BLANK_PASSES 2
#define SILENT 0
#define MAXITERS 100
#define CAP_MAXITERS 7
#define MAXTIME 3600 * 24
#define GAIN_THRESH 0.0
#define P_STAGE_SIZE 2
#define P_BIGSTEP_P .5
/*#define P_BIGSTEP_CHANGE_P .5*/
#define P_BIGSTEP_CHANGE_P .75
#define P_BIGSTEP_WT_CHANGE .5
/*#define P_SMALLSTEP_CHANGE_P .25*/
#define P_SMALLSTEP_CHANGE_P .4
#define P_SMALLSTEP_WT_CHANGE .33
#define P_RETAIN_REAL_BEST 0
#define P_INITIAL_T 4.0
#define P_T_MULT 0.7
#define P_T_MIN 0.4
#define P_SUCCESSIVE_FAILS_MAX 30
#define P_DROP_FACTOR .025	/* Was .05 up to 18/7/97 */

void interrpt() ;
void dump_output() ;
int interrupt_received ;

char *trec_topic ;
int Numterms ;			/* Total number of terms in selection list */
int startsize ;
int no_add = 0, no_remove = 0 ;
int change_since_remake = 0 ;
int change_since_vw = 0 ;
double gain_thresh ;
double btm_wt ;
int relset ;
int top_num ;
int gran ;
int score_code ;
int silent ;
int num_dropped ;
int N, R ;
int relret ;
int successive_failures_max ;
int max_blank_passes = MAX_BLANK_PASSES ;
int max_skip = DEFAULT_MAX_SKIP ;
char *initial_state ;
int startwith ;
int remake_thresh = REMAKE_THRESH ;
int vw_thresh = VW_THRESH ;
int allvw = 0 ;
int do_prov_wts = 0 ;
int vwend=0 ;
int may_perturb = 0 ;
int P_reg = 0 ;			/* Perturbation regime, unused ? */
double p_up, p_down ;
int P_successive_fails_max = P_SUCCESSIVE_FAILS_MAX ;
int P_retain_real_best = P_RETAIN_REAL_BEST ;

/* For PERTURB */
int passes_before_perturb = 0 ;
int passes_after_perturb = 0 ;
int P_stage_size = P_STAGE_SIZE ;
int P_fixed_iters = 0 ;		/* If nonzero does 50 * P_stage_size iterations per stage */
double P_bigstep_p = P_BIGSTEP_P ;
double P_bigstep_change_p = P_BIGSTEP_CHANGE_P, P_bigstep_wt_change = P_BIGSTEP_WT_CHANGE ;
double P_smallstep_change_p = P_SMALLSTEP_CHANGE_P, P_smallstep_wt_change = P_SMALLSTEP_WT_CHANGE ;
double P_initial_T = P_INITIAL_T, T_mult = P_T_MULT, T_min = P_T_MIN ;

/* Defaults: may be overridden by env */
double vwup = VWUP, vwdown = VWDOWN ;/* Initial values */
double vwup_min = VWUP_MIN ;
double vwdown_min = VWDOWN * VWUP_MIN / VWUP ;
double vw_change_fraction = 0.0 ; /* If nonzero vwup and vwdown are mult by
				     vw_change_fraction after each pass */

char *usage = "<relsfile> <termsfile>\n\
The environment variable BSS_DB must be set.\n\
You may also need one or more of BSS_ATTRIBUTE and BSS_LIMIT.\n\
The following environment variables amy also be set if needed:\n\
  RTSEL_SILENT (nonzero inhibits most of the stderr logging output)\n\
  RTSEL_TREC_TOPIC=<topic number> reports topic number to stderr at end\n\
  RTSEL_BROOKES (scores using Brookes measure)\n\
  RTSEL_LBROOKES (scores using Brookes measure on the top <top_num> (see below) docs.)\n\
  RTSEL_TREC_AVEP (scores using mean precision worked out as TREC eval\n\
  RTSEL_T_AVEP_RCL (scores using wtd ave of TREC avep and recall\n\
  RTSEL_CBM1 (scores using mean of BROOKES/10 and TREC_AVEP)\n\
  RTSEL_AVEP (scores using mean of precisions at setsize/gran, setsize/2*gran\n\
    etc\n\
  RTSEL_GRAN=<num> sets the granularity for the AVEP measure (default 10)\n\
  RTSEL_DIFFMEANS (scores using unnormalised difference of means\n\
  RTSEL_RECALL (uses recall score)\n\
  RTSEL_RPREC (uses precision after <numrels> docs)\n\
  RTSEL_COMBINED_MEASURE or RTSEL_3PT (scores using the mean of precisions at 2R, R and .5R)\n\
  RTSEL_NO_ADD (if nonzero terms are only removed, not added - default false)\n\
  RTSEL_NO_REMOVE (if nonzero terms are only added, not removed - default false)\n\
  RTSEL_STARTSIZE (number of terms to start with - default 0)\n\
  RTSEL_MAXITERS (limits max number of iterations - default 100 for CFP or CB,\n\
    7 for CAP)\n\
  RTSEL_GAIN_THRESH (sets the percentage gain threshold below which\n\
    the procedure will stop - default 0)\n\
  RTSEL_MAXTIME (elapsed time after which procedure stops - default 24 hrs)\n\
  RTSEL_MAX_SUCCESSIVE_FAILURES (max successive times a term fails to increase\n\
    the score before it is dropped - default 100)\n\
  RTSEL_MAX_SKIP (number of times a dropped term is skipped before retrying once\n\
    - default 100)\n\
  RTSEL_USE_SMALL_BIGN causes the Brookes procedure to use set size instead\n\
    of real BIGN (dubious)\n\
  RTSEL_DO_TOP_ONLY=<top_num> causes the scoring procedure to use only the top \n\
    <top_num> docs\n\
    or if RTSEL_DO_TOP_ONLY=R<num> to use <num> * R top docs\n\
  RTSEL_CIRCULAR or RTSEL_CAP causes every term to be tried (unless dropped)\n\
    before starting at the top again; this is done up to MAXITERS times\n\
  RTSEL_INITIAL_STATE=\"<termnum>[ <termnum>]\" makes it start by including the\n\
    given query term numbers\n\
  RTSEL_STARTWITH=<termnum> (if RTSEL_INITIAL_STATE set. starts with this term \n\
    on the first iteration\n\
  RTSEL_CHOOSE_FIRST_POSITIVE or RTSEL_CFP finishes an iteration as soon as \n\
    a score exceeds the gain threshold\n\
  RTSEL_USE_CUTOFF limits the minimum weight in a set to RTSEL_CUTOFF_FACTOR * \n\
    the weight of the <top_num>th document in the current set.\n\
  RTSEL_CUTOFF_FACTOR (see previous item, default 0, must be between 0 and 1).\n\
  RTSEL_REMAKE_THRESH=n remakes the current set whenever n sets have been added or removed\n\
    (0 or absent (default) is infinite)\n\
  RTSEL_VW_THRESH=n jiggles term weights whenever n sets have been added or removed\n\
    and at end (0 or absent (default) is infinite)\n\
  RTSEL_VWUP=<real> is the fraction (default .5) by which a term's current \n\
    weight is upped when jiggled\n\
  RTSEL_VWDOWN=<real> is the analogous fraction for weight reduction \n\
    (default .33)\n\
  RTSEL_VWCHANGE=<real> (>=0) causes vwup and vwdown to be multiplied by \n\
    this value after each pass (default 0) \n\
  RTSEL_ALLVW=<int> > 0 causes terms' weights to be jiggled continuously\n\
  RTSEL_PERTURB=<code> allows random perturbation of weights. The (int) code\n\
    determines the perturbation regime (code nonpositive turns off perturbation).\n\
  RTSEL_P_PASSES_BEFORE=<int> does this many passes before going into PERTURB\n\
  RTSEL_P_STAGE_SIZE=<small int> does max this times 50 perturbations per stage\n\
    (default 2)\n\
  RTSEL_P_FIXED_ITERS If nonzero does 50 X RTSEL_P_STAGE_SIZE (def 2) perturbations\n\
    per stage (default zero) \n\
  RTSEL_INITIAL_T=<real> sets initial 'temperature' (default 4, relative to scores * 1000)\n\
  RTSEL_T_MULT=<real> sets temperature reduction coeff (default 0.7)\n\
  RTSEL_T_MIN=<real> sets minimum temperature (default 0.4)\n\
  RTSEL_P_BIGSTEP_INCR=<real> \n\
  RTSEL_P_SMALLSTEP_INCR=<real> \n\
  RTSEL_P_BIGSTEP_P=<prob> proportion of state changes using big step\n\
    as opposed to small step, default .5\n\
  RTSEL_P_BIGSTEP_CHANGE_P=<prob> probability of a weight being changed given \n\
    big step (default .75)\n\
  RTSEL_P_SMALLSTEP_CHANGE_P=<prob> probability of a weight being changed given \n\
    small step (default .4)\n\
  RTSEL_P_BIGSTEP_WT_CHANGE=<prob> mean proportional weight change (big step) (default .5)\n\
  RTSEL_P_SMALLSTEP_WT_CHANGE=<prob> mean proportional weight change (small step) (default .33)\n\
  RTSEL_P_RETAIN_REAL_BEST=<int> If nonzero reverts to `real' best state\n\
    at end of each stage\n\
  RTSEL_VARY_NONINCL (only if ALLVW) allows the weights of non-included\n\
    terms to be altered provisionally depending on current search result.\n\
  RTSEL_VWEND=<int> > 0 does jiggling that many times at the end\n\
  RTSEL_MARK_EARLY causes relevant docs in the atomic sets to be `marked'; this \n\
    obviates the need for marking each evaluation set. Default on.\n\n\
  NB. Most of the VW stuff only works if CAP\n" ;

extern int Dbg ;
extern FILE *bss_syslog ;

main(argc, argv)
     int argc ;
     char **argv ;
{
  int use_small_bign = 0 ;
  FILE *relsfile ;
  FILE *termsfile ;
  /* Now global int relset ;*/
  /* Now global int R ; */
  /* Now global int relret ; */
  /* Now global int N ; */
  int new_set_np ;
  int last_term_altered = -1 ;	/* Term number of last term added or removed */
  int temp_N, temp_R ;
  int result ;
  int do_assoc = 0 ;		
  int choose_first_positive = 0 ;
  int do_top_R = 0 ;		/* If true does Brookes on top bigr docs */
  double R_factor = 1.0 ;	/* R multiplied by this for Brookes if
				   do_top_R */
  int new_set, best_set ;
  int top_set, temp_set ;
  char command[1024], *response ;
  char *envp ;
  double maxbr = -1e34, max2br = -1e34, max3br = -1e34 ;
  double temp_br ;
  double top_br ;
  double gain = 0.0 ;
  double rprec = 0.0 ;
  int maxiters = 0 ;
  int maxtime = MAXTIME ;
  int start_time = time(NULL) ;
  int iter_time, t ;
  int maxterm ;
  int iters, j, k ;
  int dummy ;
  /* Now global int score_code = BROOKES ;	/* Default */
  /* gran now global		/* Granularity for AVEP */

  int circular = 0 ;
 
  char fname[512] ;

  signal(SIGUSR1, interrpt) ;
  signal(SIGUSR2, dump_output) ;

  srand((long)time(NULL)) ;

  Dbg = (0x80|0x20) ;
  Dbg = 0 ;
  bss_syslog = stderr ;

  C_state = getstate() ;
  Best_state = getstate() ;
  Realbest_state = getstate() ;

  successive_failures_max = MAXITERS ;
  score_code = BROOKES ;
  gain_thresh = GAIN_THRESH ;

  if ( ! silent ) fprintf(stderr, "pid=%d\n", getpid()) ;

  if ( argc < 3 ) {
    fprintf(stderr, "Usage: %s %s\n", argv[0], usage) ;
    exit(1) ;
  }
  if ( startup_bss() < 0 ) {
    fprintf(stderr, "Can't start BSS\n") ;
    exit(1) ;
  }
  /* kill(getpid(), SIGSTOP) ; */


#ifdef DEBUG
  sprintf(fname, "/tmp/rtg_select_errors.%d", getpid()) ;
  Err_file = fopen(fname, "w") ;
#endif
  if ( (envp = getenv("BSS_LIMIT")) && *envp != '\0' ) {
    sprintf(command, "set l=%d", atoi(envp)) ;
    I0 ;
    fprintf(stderr, "Set limit to %d\n", atoi(envp));
  }
  if ( (envp = getenv("BSS_MEM")) ) {
    sprintf(command, "set mem=%d", atoi(envp)) ;
    I0 ;
    fprintf(stderr, "Set memory %d\n", atoi(envp));
  }
  if ( (envp = getenv("BSS_HOLD_MEM")) ) {
    Hold = atoi(envp) ;
    fprintf(stderr, "Set Hold %d\n", atoi(envp)) ;
  }
  if ( (envp = getenv("BSS_OP")) ) {
    Op = envp ;
    fprintf(stderr, "Set op for initial termsets %s\n", Op);
  }
  if ( (envp = getenv("BSS_K1")) ) {
    k1 = atof(envp) ;
    fprintf(stderr, "Set k1 %.2f\n", k1);
  }
  if ( (envp = getenv("BSS_BM25B")) ) {
    b = atof(envp) ;
    fprintf(stderr, "Set bm25_b %.2f\n", b);
  }
  if ( (envp = getenv("BSS_DBG")) ) {
    sprintf(command, "set dbg=%d", atoi(envp)) ;
    I0 ;
    fprintf(stderr, "Set debug %d\n", atoi(envp));
  }
  if ( envp = getenv("RTSEL_TREC_TOPIC") ) {
    trec_topic = envp ;
    if ( *trec_topic ) fprintf(stderr, "TREC topic set to %s\n", trec_topic) ;
  }
  sprintf(command, "d bign") ;
  I0q ;
  N = atoi(response) ;
  fprintf(stderr, "BIGN=%d\n", N) ;
  if ( (envp = getenv("RTSEL_USE_SMALL_BIGN")) ) {
    use_small_bign = 1 ;
    score_code |= SMALLN ;
    if ( ! silent )
      fprintf(stderr, "Will use setsize instead of BIGN\n") ;
  }
  if ( (envp = getenv("RTSEL_DO_TOP_ONLY")) ) {
    if ( *envp == 'R' ) {
      do_top_R = 1 ;
      if ( atof(envp + 1) > 0.0999 ) R_factor = atof(envp + 1) ;
      if ( ! silent )
	fprintf(stderr,
	"Will score on top R * %.1f docs\n",
		R_factor) ;
    }
    else {
      top_num = atoi(envp) ;
      if ( top_num < 5 ) {
	fprintf(stderr, "Can't score on less than 5 docs\n") ;
	exit(1) ;
      }
      if ( ! silent )
      fprintf(stderr,
	      "Will score on top %d\n", top_num) ;
    }
  }
  if ( (envp = getenv("RTSEL_LBROOKES")) ) {
    if ( top_num > 0 ) {
      score_code = LBROOKES ;
      if ( ! silent )
	fprintf(stderr, "Will score using Brookes measure on top %d docs\n", top_num) ;
    }
    else {
      score_code = BROOKES ;
      if ( ! silent )
	fprintf(stderr, "Will score using Brookes measure\n") ;
    }
  }
  else if ( (envp = getenv("RTSEL_BROOKES"))) {
    score_code = BROOKES ;
    if ( ! silent )
      fprintf(stderr, "Will score using Brookes measure\n") ;
  }
  else if ( (envp = getenv("RTSEL_CBM1")) ) {
    score_code = CBM1 ;
    if ( ! silent )
      fprintf(stderr, "Will score using mean of BROOKES/10 and TREC_AVEP\n");
  }
  else if ( (envp = getenv("RTSEL_DIFFMEANS")) ) {
    score_code = DIFFMEANS ;
    if ( ! silent )
      fprintf(stderr, "Will score using difference of means\n") ;
  }
  else if ( (envp = getenv("RTSEL_COMBINED_MEASURE")) ||
      (envp = getenv("RTSEL_3PT")) ) {
    score_code = COMBINED_MEASURE ;
    if ( ! silent )
      fprintf(stderr, "Will score using combined measure\n") ;
  }
  else if ( (envp = getenv("RTSEL_RECALL")) ) {
    score_code = RECALL ;
    if ( ! silent )
      fprintf(stderr, "Will score using recall\n") ;
  }
  else if ( (envp = getenv("RTSEL_RPREC")) ) {
    score_code = RPREC ;
    do_top_R = 1 ;
    if ( ! silent )
      fprintf(stderr, "Will score using R-precision\n") ;
  }
  else if ( (envp = getenv("RTSEL_TREC_AVEP")) ) {
    score_code = TREC_AVEP ;
    if ( ! silent )
      fprintf(stderr, "Will score using TREC average precision\n") ;
  }
  else if ( (envp = getenv("RTSEL_LF1")) ) {
    score_code = TS_LF1 ;
    if ( ! silent )
      fprintf(stderr, "Will score using utility function LF1\n") ;
  }
  else if ( (envp = getenv("RTSEL_LF2")) ) {
    score_code = TS_LF2 ;
    if ( ! silent )
      fprintf(stderr, "Will score using utility function LF2\n") ;
  }
  else if ( (envp = getenv("RTSEL_LF3")) ) {
    score_code = TS_LF3 ;
    if ( ! silent )
      fprintf(stderr, "Will score using utility function LF3\n") ;
  }
  else if ( (envp = getenv("RTSEL_NF1")) ) {
    score_code = TS_NF1 ;
    if ( ! silent )
      fprintf(stderr, "Will score using utility function NF1\n") ;
  }
  else if ( (envp = getenv("RTSEL_NF2")) ) {
    score_code = TS_NF2 ;
    if ( ! silent )
      fprintf(stderr, "Will score using utility function NF2\n") ;
  }
  else if ( (envp = getenv("RTSEL_T_AVEP_RCL")) ) {
    score_code = T_AVEP_RCL ;
    if ( ! silent )
      fprintf(stderr,
	      "Will score using wtd ave of ave prec and recall\n") ;
  }
  else if ( (envp = getenv("RTSEL_AVEP")) ) {
    score_code = AVEP ;
    if ( ! silent )
      fprintf(stderr, "Will score using average precision\n") ;
    if ( envp = getenv("RTSEL_GRAN") ) gran = atoi(envp) ;
    if ( gran <= 0 ) gran = AVEP_GRAN ;
    if ( ! silent )
      fprintf(stderr, "Granularity %d\n", gran) ;
  }
  if ( (envp = getenv("RTSEL_GAIN_THRESH")) ) {
    gain_thresh = atof(envp) ;
    if ( ! silent )
      fprintf(stderr,
      "Will stop after an iteration fails to increase the score by %.2f%%\n",
	      gain_thresh) ;
  }
  if ( (getenv("RTSEL_CHOOSE_FIRST_POSITIVE") || getenv("RTSEL_CFP")) &&
      ! getenv("RTSEL_CIRCULAR") && ! getenv("RTSEL_CAP") ) {
    choose_first_positive = 1 ;
    if ( ! silent )
      fprintf(stderr,
      "Iteration will end as soon as a term increases the score by %.2f%%\n",
	      gain_thresh) ;
  }
  if ( (envp = getenv("RTSEL_MAX_SUCCESSIVE_FAILURES")) ) {
    successive_failures_max = atoi(envp) ;
    if ( ! silent )
      fprintf(stderr,
    "Terms will be dropped after %d successive failures to increase the score\n",
	      successive_failures_max) ;
  }
  if ( (envp = getenv("RTSEL_NO_ADD")) ) {
    no_add = atoi(envp) ;
        if ( ! silent )
      fprintf(stderr, "Terms will only be removed, not added\n") ;
  }
  if ( (envp = getenv("RTSEL_NO_REMOVE")) ) {
    no_remove = atoi(envp) ;
        if ( ! silent )
      fprintf(stderr, "Terms will only be added, not removed\n") ;
  }
  if ( (envp = getenv("RTSEL_MAX_SKIP")) ) {
    max_skip = atoi(envp) ;
    if ( max_skip <= 0 ) max_skip = DEFAULT_MAX_SKIP ;
    if ( ! silent )
      fprintf(stderr,
    "Dropped terms will be retried after being skipped %d times\n",
	      max_skip) ;
  }
  if ( getenv("RTSEL_CIRCULAR") || getenv("RTSEL_CAP") ) {
    circular = 1 ;
    if ( ! silent )
      fprintf(stderr,
	      "Will choose all positive terms in each iteration\n") ;
  }
  if ( (envp = getenv("RTSEL_INITIAL_STATE")) ) {
    initial_state = envp ;
    if ( ! silent ) fprintf(stderr, "Initial state is") ;
    while ( sscanf(initial_state, "%d%n", &j, &k) == 1 ) {
      initial_state += k ;
      C_state->incl[j] = 1 ;
      C_state->inclterms++ ;
      if ( ! silent ) fprintf(stderr, " %d", j) ;
    }
    if ( ! silent ) fprintf(stderr, "\n") ;
    if ( (envp = getenv("RTSEL_STARTWITH")) && atoi(envp) > 0 ) {
      startwith = atoi(envp) ;
      if ( ! silent ) fprintf(stderr, "Will start with term %d\n",
			      startwith) ;
    }
  }
  if ( (envp = getenv("RTSEL_STARTSIZE")) && ! initial_state ) {
    /* Ignored if initial state set */
    int j ;
    startsize = atoi(envp) ;
    C_state->inclterms = startsize ;
    if ( ! silent )
      fprintf(stderr, "Will start with first %d term(s)\n", startsize) ;
  }
  if ( envp = getenv("RTSEL_MARK_EARLY") ) {
    Mark_early = 1 ;
    if ( ! silent ) fprintf(stderr, "Will `mark' atomic sets\n") ;
  }
  if ( envp = getenv("RTSEL_USE_CUTOFF") ) {
    Use_cutoff = 1 ;
  }
  if ( envp = getenv("RTSEL_CUTOFF_FACTOR") ) {
    Cutoff_factor = atof(envp) ;
    if ( Cutoff_factor < 0 || Cutoff_factor > 1 ) {
      fprintf(stderr, "Cutoff_factor must be between 0 and 1") ;
      exit(1) ;
    }
    if ( ! silent && Use_cutoff ) fprintf(stderr, "Set Cutoff_factor %.4f\n", Cutoff_factor) ;
  }
  if ( (envp = getenv("RTSEL_REMAKE_THRESH")) ) {
    remake_thresh = atoi(envp) ;
    if ( ! silent && remake_thresh > 0 )
      fprintf(stderr,
     "Current set will be remade when current set has changed by %d\n", remake_thresh) ;
  }
  if ( (envp = getenv("RTSEL_VW_THRESH")) ) {
    vw_thresh = atoi(envp) ;
    if ( ! silent && vw_thresh > 0 )
      fprintf(stderr,
"Weights of included terms will be jiggled when current set has changed by %d\n",
	      vw_thresh) ;
  }
  if ( (envp = getenv("RTSEL_VWUP")) ) {
    vwup = atof(envp) ;
    if ( ! silent )
      fprintf(stderr,
      "VW 'up' fraction is %.3f\n", vwup);
  }
  if ( (envp = getenv("RTSEL_VWDOWN")) ) {
    vwdown = atof(envp) ;
    if ( ! silent )
      fprintf(stderr,
      "VW 'down' fraction is %.3f\n", vwdown);
  }
  if ( (envp = getenv("RTSEL_ALLVW")) ) {
    allvw = atoi(envp) ;
    if ( ! silent && allvw > 0 )
      fprintf(stderr,
      "Weights of all new terms will be jiggled\n");
    if ( (envp = getenv("RTSEL_VWCHANGE")) ) {
      vw_change_fraction = atof(envp) ;
      if ( vw_change_fraction < 0 ) {
	fprintf(stderr, "RTSEL_VWCHANGE must be >= 0\n") ;
	exit(1) ;
      }
      if ( ! silent ) {
	fprintf(stderr,
		"VW change fraction is %.3f\n", vw_change_fraction);
      }
      fprintf(stderr, "Min VW up fraction is %.3f\n", vwup_min) ;
      fprintf(stderr, "Min VW down fraction is %.3f\n", vwdown_min) ;
      fprintf(stderr, "Max number of passes without gain is %d\n",
	      max_blank_passes) ;
    }
    if ( (envp = getenv("RTSEL_VARY_NONINCL")) ) {
      do_prov_wts = 1 ;
      if ( ! silent )
	fprintf(stderr, "Weights of nonincluded terms may be altered\n") ;
    }
  }
  if ( (envp = getenv("RTSEL_MAXITERS")) ) {
    if ( atoi(envp) > 0 ) {
      maxiters = atoi(envp) ;
      if ( ! silent )
	fprintf(stderr, "Maximum number of iterations %d\n", maxiters) ;
    }
  }
  if ( (envp = getenv("RTSEL_PERTURB")) ) {
    P_reg = atoi(envp) ;
    if ( P_reg > 0 ) {
      may_perturb = 1 ;
      if ( ! silent )
	fprintf(stderr,
		"Weights may be perturbed randomly under perturbation regime %d\n",
		P_reg) ;
    }
    if ( (envp = getenv("RTSEL_P_PASSES_BEFORE")) && atoi(envp) > 0 ) {

      passes_before_perturb = atoi(envp) ;
      if ( ! silent ) fprintf(stderr, "Will do %d passes before perturbation begins\n",
			      passes_before_perturb) ;
    }
    if ( (envp = getenv("RTSEL_P_STAGE_SIZE")) && atoi(envp) > 0 ) {
      P_stage_size = atoi(envp) ;
    }
    if ( envp = getenv("RTSEL_P_FIXED_ITERS"))  {
      P_fixed_iters = atoi(envp) ; /* 50 X P_stage_size */
    }
    if ( (envp = getenv("RTSEL_INITIAL_T")) && atof(envp) > .001 ) {
      P_initial_T = atof(envp) ;
      if ( ! silent ) fprintf(stderr, "Set initial temperature %.3f\n", P_initial_T) ;
    }
    if ( (envp = getenv("RTSEL_T_MULT")) && atof(envp) > .001 ) {
      T_mult = atof(envp) ;
      if ( ! silent ) fprintf(stderr, "Temperature multiplier %.3f\n", T_mult) ;
    }
    if ( (envp = getenv("RTSEL_T_MIN")) && atof(envp) > .001 ) {
      T_min = atof(envp) ;
      if ( ! silent ) fprintf(stderr, "Minimum temperature %.3f\n", T_min) ;
    }
    if ( (envp = getenv("RTSEL_P_BIGSTEP_WT_CHANGE")) && atof(envp) > .001 ) {
      P_bigstep_wt_change = atof(envp) ;
      if ( ! silent ) fprintf(stderr, "Mean prop wt change (big step) %.3f\n",
			      P_bigstep_wt_change) ;
    }
    if ( (envp = getenv("RTSEL_P_SMALLSTEP_WT_CHANGE")) && atof(envp) > .001 )
      {
	P_smallstep_wt_change = atof(envp) ;
	if ( ! silent )
	  fprintf(stderr, "Mean prop wt change (small step) %.3f\n",
		  P_smallstep_wt_change) ;
      }
    if ( (envp = getenv("RTSEL_P_BIGSTEP_CHANGE_P")) && atof(envp) > .001 ) {
      P_bigstep_change_p = atof(envp) ;
      if ( ! silent )
	fprintf(stderr, "Prob of wt change (big step) %.3f\n",
		P_bigstep_change_p) ;
    }
    if ( (envp = getenv("RTSEL_P_SMALLSTEP_CHANGE_P")) && atof(envp) > .001 ) {
      P_smallstep_change_p = atof(envp) ;
      if ( ! silent )
	fprintf(stderr, "Prob of wt change (small step) %.3f\n",
		P_smallstep_change_p) ;
    }
    if ( (envp = getenv("RTSEL_P_BIGSTEP_P")) && atof(envp) > .001 ) {
      P_bigstep_p = atof(envp) ;
      if ( ! silent )
	fprintf(stderr, "Prob of big step %.3f and small step %.3f\n",
		P_bigstep_p, 1 - P_bigstep_p) ;
    }
    if ( (envp = getenv("RTSEL_P_RETAIN_REAL_BEST")) ) {
      P_retain_real_best = atoi(envp) ;
      if ( ! silent && atoi(envp) > 0 )
	fprintf(stderr,
		"Will revert to `real' best state at end of each stage\n") ;
    }
  }
  if ( (envp = getenv("RTSEL_VWEND")) ) {
    vwend = atoi(envp) ;
    if ( ! silent && vwend > 0 )
      fprintf(stderr,
"Weights of included terms will be jiggled %d times at the end\n", vwend) ;
  }

  if ( maxiters == 0 ) {
    if ( circular ) maxiters = CAP_MAXITERS ;
    else maxiters = MAXITERS ;
    if ( ! silent )
      fprintf(stderr, "Maximum number of iterations will be %d\n", maxiters) ;
  }

  if ( (relsfile = fopen(argv[1], "r")) == NULL ) {
    fprintf(stderr, "Can't open relsfile %s\n", argv[1]) ;
    exit(1) ;
  }
  if ( (termsfile = fopen(argv[2], "r")) == NULL ) {
    fprintf(stderr, "Can't open termsfile %s\n", argv[2]) ;
    exit(1) ;
  }
  if ( (relset = make_relset(relsfile, &R)) < 0 ) {
    fprintf(stderr, "make_relset() failed somehow\n") ;
    exit(1) ;
  }
  if ( do_top_R ) top_num = (int)rint((double)R * R_factor) ;
  if ( score_code == COMBINED_MEASURE ) top_num = R * 2 ;
  if ( ! silent ) {
    fprintf(stderr, "Made relset, setnum=%d, R=%d\n", relset, R) ;
    if ( top_num > 0 ) fprintf(stderr, "Will score on top %d docs\n",
			       top_num) ;
  }
  if ( (Numterms = setup_termrecs(termsfile, relset, R)) < 0 ) {
    fprintf(stderr, "setup_termrecs() failed somehow\n") ;
    exit(1) ;
  }

  for ( k = 0 ; k < MAXTERMS ; k++ ) C_state->w[k] = Terms[k].ow ;
  C_state->c_set = best_set = -1 ;
  C_state->score = C_state->new_score = 0.0 ;

  if ( do_assoc ) {
    if ( ! silent ) 
      fprintf(stderr, "Setting up term association table\n") ;
    setup_assocs(Numterms) ;
  }

  if ( startsize > 0 ) {	/* Set up initial set of startsize terms */
    if ( startsize > Numterms ) {
      if ( ! silent ) fprintf(stderr, "Reducing start size to %d\n", Numterms) ;
      startsize = Numterms ;
    }
    if ( ! silent )
      fprintf(stderr, "Forming initial set of the top %d term(s)\n", startsize) ;
    result = make_topn_set(startsize) ;
    if ( result < 0 ) goto error ;
    for ( k = 0 ; k < startsize ; k++ ) C_state->incl[k] = 1 ;
    C_state->inclterms = startsize ;
    change_since_remake += startsize ;
    change_since_vw += startsize ;
    C_state->c_set = new_set = result ;
    C_state->score =
      score(C_state->c_set, relset, N, R, &relret, top_num, score_code, gran) ;
    if ( Use_cutoff ) {
      btm_wt = wt_of_nth(C_state->c_set, top_num) ;
      C_state->cssize = setsize(C_state->c_set) ;
      C_state->cutoff = btm_wt * Cutoff_factor ;
    }
    if ( ! silent ) {
      fprintf(stderr, "Score for initial set is %.3f, rcl=%.1f%%\n",
	      C_state->score, (double)relret/(double)R*100) ;
      if ( Use_cutoff ) 
	fprintf(stderr,
		"Wt of rec %d is %.3f, %d terms, setsize %d\n",
		top_num - 1, btm_wt, C_state->inclterms, C_state->cssize) ;
    }
  }
  else /* startsize = 0 */ if ( initial_state ) {
    /* If initial_state make set from the given terms */
    startsize = 0 ;		/* Recalc it for case initial_state */
    sprintf(command, "f ") ;
    for ( j = 0 ; j < Numterms ; j++ ) {
      if ( C_state->incl[j] ) {
	startsize++ ; 
	sprintf(command + strlen(command), "s=%d ", Terms[j].set) ;
	if ( ! silent )
	  fprintf(stderr,
		  "Putting in term #%d, \"%s\", (np %d, wt %.3f)\n",
		  j, Terms[j].t, Terms[j].np, Terms[j].ow) ;
      }
    }
    I0q ;
    C_state->c_set = new_set = atoi(response + 1) ;
    C_state->inclterms = startsize ;
    C_state->cssize = new_set_np = atoi(strstr(response, "np=") + 3) ;
    C_state->score = C_state->scores[startsize - 1] =
      score(C_state->c_set, relset, N, R, &relret, top_num, score_code, gran) ;
    change_since_remake += startsize ;
    change_since_vw += startsize ;
    if ( Use_cutoff) {
      btm_wt = wt_of_nth(C_state->c_set, top_num) ;
      C_state->cssize = setsize(C_state->c_set) ;
      C_state->cutoff = btm_wt * Cutoff_factor ;
    }
    if ( ! silent ) {
      fprintf(stderr,
	      "Score for initial set is %.3f, rcl=%.1f%%\n",
	      C_state->score, (double)relret/(double)R*100) ;
      if ( Use_cutoff )
	fprintf(stderr, "Wt of rec %d is %.3f, %d terms, setsize %d\n",
		top_num - 1, btm_wt, C_state->inclterms, C_state->cssize) ;
    }
  } /* initial_state > 0 and startsize == 0) */

  if ( circular ) {
    circular_add(maxiters) ;
    do_output(stdout) ;
    iexit() ;
    exit(0) ;
  }
  for ( iters = 0 ; ; iters++ ) {
    /* Don't actually know max number of iters if adding AND removing */
    int found_one = 0 ;
    int num_considered ;
    int start ;
    start = (iters == 0) ? startwith : 0 ;
    iter_time = time(NULL) ;
    if ( ! silent )
      fprintf(stderr, "Starting iteration %d\n", iters + 1) ;
    maxbr = max2br = max3br = -1e34 ;
    memset((char *)C_state->examined, '\0', sizeof(C_state->examined)) ;

    for ( j = start, num_considered = 0 ; j < Numterms ; j++ ) {

      /* Look at each term. Skip it if too many failures, or already included
	 not removing. */
      if ( C_state->dropped[j] ) {
	if ( C_state->skipped[j] >= max_skip &&
	    ( C_state->incl[j] && ! no_remove || ! C_state->incl[j] && ! no_add ) ) {
	  if ( ! silent )
	    fprintf(stderr, "Retrying term #%d after %d failures\n",
		    j, C_state->skipped[j]) ;
	  C_state->skipped[j] = 0 ;
	}
	else {
	  C_state->skipped[j]++ ;
	  if ( ! silent )
	    fprintf(stderr, "Skipping \"%s\"\n", Terms[j].t) ;
	  continue ;
	}
      }

      if ( j == last_term_altered ) continue ;

      /* If term's currently included, try removing it unless no_remove */
      if ( C_state->incl[j] ) {
	if ( no_remove ) {
	  if ( ! silent )
	    fprintf(stderr, "Skipping \"%s\", already included\n",
		    Terms[j].t) ;
	  continue ;
	}
	else new_set = remove_set(C_state->c_set, j, C_state->cutoff) ;
      }
      /* If term's not currently included, try adding it */
      else {
	if ( no_add ) continue ;
	/* Combine C_state->c_set and Terms[j].set */
	else new_set =
    add_set(C_state->c_set, Terms[j].set, top_num, C_state->cutoff, C_state->w[j]/Terms[j].ow) ;
	/* NB this works even if current set is -1 */
      } 
      temp_br =
	score(new_set, relset, N, R, &relret, top_num, score_code, gran) ;
      C_state->scores[j] =  temp_br - C_state->score ;
      gain = C_state->scores[j] * 100 / C_state->score ;
      C_state->examined[j] = 1 ;
      num_considered++ ;
      if ( ! silent ) {
	if ( ! C_state->incl[j] )
	  fprintf(stderr,
 "Score for term #%d, \"%s\" (np %d, wt %.3f), is %.3f (%.2f%%), rcl=%.1f%%\n",
 j, Terms[j].t, Terms[j].np, C_state->w[j],
 C_state->scores[j], gain, (double)relret/(double)R*100) ;
	else 
	  fprintf(stderr,
 "Score after removing term #%d, \"%s\" (np %d, wt %.3f), is %.3f (%.2f%%), rcl=%.1f%%\n",
 j, Terms[j].t, Terms[j].np, C_state->w[j],
 C_state->scores[j], gain, (double)relret/(double)R*100) ;
      }
      if ( C_state->scores[j] < 0.0 ) {
	if ( ! C_state->dropped[j] ) {
	  C_state->failures[j]++ ;
	  if ( C_state->failures[j] >= successive_failures_max ) {
	    C_state->dropped[j] = 1 ;
	    num_dropped++ ;
	  }
	}
      }
      else {
	C_state->failures[j] = 0 ;
	if ( C_state->dropped[j] ) {
	  C_state->dropped[j] = 0 ;
	  C_state->skipped[j] = 0 ;
	  num_dropped-- ;
	}
      }
      /* Is this term good? */
      if ( choose_first_positive && gain > gain_thresh ) {
        if ( ! silent ) 
	  fprintf(stderr,
		  "First good term is \"%s\", term #%d, with rel br=%.3f\n",
		  Terms[j].t, j, C_state->scores[j]) ;
	maxterm = j ;
	maxbr = C_state->scores[j] ;
	best_set = new_set ;
	break ;			/* Out of iteration */
      }
      else if ( C_state->scores[j] > maxbr ) {
	if ( ! silent ) {
	  if ( C_state->scores[j] > 0 ) {
	    fprintf(stderr,
		    "Best term is \"%s\", term #%d, with rel br=%.3f\n",
		    Terms[j].t, j, C_state->scores[j]) ;
	    if ( found_one > 0 )
	      fprintf(stderr, "Previous was \"%s\", #%d, with br=%.3f\n",
		      Terms[maxterm].t, maxterm, C_state->scores[maxterm]) ;
	  }
	}
	if ( found_one > 0 && C_state->c_set >= 0 ) delete_set(best_set) ;
	maxterm = j ;
	maxbr = C_state->scores[j]  ;
	best_set = new_set ;
	found_one++ ;
      } /* (br > maxbr) */
      else if ( C_state->c_set >= 0 ) delete_set(new_set) ;
    } /* (end of iteration) */
    /* Gets here with C_state->c_set (the base for the iter just finished),
       and best_set = set after removing or adding maxterm from/to current_set,
       which is OK to be the new current set
       unless top_num > 0, in which case it is "targeted" and
       unsuitable for further use.
       Not really true now (May 96), using cutoff weight instead of target */
       
    t = time(NULL) ;
    if ( ! silent )
      fprintf(stderr, "Iteration time %d secs, elapsed time %d secs\n",
			    t - iter_time, t - start_time) ;
    if ( maxbr > 0.0 ) {
      C_state->score += maxbr ;	/* maxbr has had C_state->score deducted */
      if ( ! C_state->incl[maxterm] ) {
	if ( ! Use_cutoff ) {
	  /* if top_num > 0 need to make "full" (target=0) version of best
	     set */
	  if ( top_num > 0 ) { 
	    delete_set(best_set) ;
	    best_set =
    add_set(C_state->c_set, Terms[maxterm].set, 0, C_state->cutoff,
	    C_state->w[j]/Terms[j].ow) ;
	  }
	}
      }
      if ( C_state->c_set >= 0 ) delete_set(C_state->c_set) ;
      C_state->c_set = best_set ;
      if ( Use_cutoff ) {
	btm_wt = wt_of_nth(C_state->c_set, top_num) ;
	C_state->cssize = setsize(C_state->c_set) ;
	C_state->cutoff = btm_wt * Cutoff_factor ;
	if ( ! silent ) 
	  fprintf(stderr,
		  "Wt of rec %d is %.3f, %d terms, cutoff %.3f, setsize %d\n",
		  top_num - 1, btm_wt, C_state->inclterms, C_state->cutoff, C_state->cssize) ;
      }
      gain = maxbr * 100 / (C_state->score - maxbr) ;
      last_term_altered = maxterm ;
      if ( C_state->incl[maxterm] ) {
	C_state->incl[maxterm] = 0 ;
	C_state->inclterms-- ;
	change_since_remake ++ ;
	change_since_vw ++ ;
	if ( ! silent ) {
	  int printed = 0, i ;
	  fprintf(stderr,
"Removing \"%s\", term #%d, wt %.3f, increases score from %.3f to %.3f",
		Terms[maxterm].t, maxterm, C_state->w[maxterm],
		C_state->score - maxbr, C_state->score) ;
	  if ( gain > 0.000001 ) fprintf(stderr, " (%.2f%%)\n", gain) ;
	  else fprintf(stderr, "\n") ;
	  if ( ! choose_first_positive ) {
	    fprintf(stderr,
   "Other positive terms of the %d considered this iteration are:\n",
		    num_considered) ;
	    for ( i = Numterms - 1 ; i >= 0 ; i-- ) 
	      if ( C_state->incl[i] && C_state->examined[i] && C_state->scores[i] > 0.00001 ) {
		fprintf(stderr, "  \"%s\" (#%d) with %.3f\n",
			Terms[i].t, i, C_state->scores[i]) ;
		printed++ ;
	      }
	    if ( ! printed ) fprintf(stderr, "(none)\n") ;
	  }
	} /* (not silent) */
      }
      else { /* not C_state->incl[maxterm] */
	C_state->incl[maxterm] = 1 ;
	C_state->inclterms++ ;
	change_since_remake ++ ;
	change_since_vw ++ ;
	if ( ! silent ) {
	  int printed = 0, i ;
	  
	  fprintf(stderr,
"Output \"%s\", term #%d, wt %.3f, increases score from %.3f to %.3f",
		Terms[maxterm].t, maxterm, C_state->w[maxterm],
		C_state->score - maxbr, C_state->score) ;
	  if ( gain > 0.000001 ) fprintf(stderr, " (%.2f%%)\n", gain) ;
	  else fprintf(stderr, "\n") ;
	  if ( ! choose_first_positive ) {
	    fprintf(stderr,
  "Other positive terms of the %d considered this iteration are:\n",
		    num_considered) ;
	    for ( i = 0 ; i < Numterms ; i++ ) 
	      if ( ! C_state->incl[i] && C_state->examined[i] && C_state->scores[i] > 0.00001 ) {
		fprintf(stderr, "  \"%s\" (#%d) with %.3f\n",
			Terms[i].t, i, C_state->scores[i]) ;
		printed++ ;
	      }
	    if ( ! printed ) fprintf(stderr, "(none)\n") ;
	  }
	} /* (not silent) */
      }

      if ( change_since_remake >= remake_thresh && remake_thresh > 0 ) {
	remake() ;
	change_since_remake = 0 ;
      }
      if ( change_since_vw >= vw_thresh && vw_thresh > 0 ) {
	vw1(vwup, vwdown) ;
	change_since_vw = 0 ;
      }
      if ( gain > 0.000001 && gain < gain_thresh ) {
	if ( ! silent )
	  fprintf(stderr,
  "Stopping after %d iterations because %% gain %.2f < threshold %.2f\n",
		  iters + 1, gain, gain_thresh) ;
	break ;
      }
    } /* (maxbr > 0 at the end of an iteration) */
    else {
      if ( ! silent )
	fprintf(stderr,
"Stopping after %d iterations because no term increases the score\n",
		  iters + 1) ;
      break ;
    }
    if ( iters + 1 >= maxiters ) {
      if ( ! silent )
	fprintf(stderr,
		"Stopping after %d iterations because max iterations = %d\n",
		iters + 1, maxiters) ;
      break ;
    }
    if ( time(NULL) - start_time >  maxtime ) {
      if ( ! silent )
	fprintf(stderr,
 "Stopping after %d iterations because clock time (%d) exceeds maxtime (%d)\n",
		time(NULL) - start_time, maxtime) ;
      break ;
    }
    if ( interrupt_received >= 2 ) {
      if ( ! silent )
        fprintf(stderr,
		"Stopping after %d iterations because interrupt received\n") ;
      break ;
    }
    /* Report */
    /* dump_output() ;		/* Safety measure */
    if ( ! silent ) {
      int k ;
      for ( k = 0 ; k < Numterms ; k++ ) {
	if ( C_state->dropped[k] ) 
	  fprintf(stderr, "Term #%d, \"%s\" dropped\n",
		  k, Terms[k].t) ;
	else if ( C_state->failures[k] >= successive_failures_max ) 
	  fprintf(stderr,
		  "Will drop term #%d, \"%s\", %d successive failures\n",
		  k, Terms[k].t, C_state->failures[k]) ;
	else if ( C_state->failures[k] > 1 ) 
	  fprintf(stderr,
		  "Term #%d, \"%s\", %d successive failures\n",
		  k, Terms[k].t, C_state->failures[k]) ;
      }
    } /* (report) */
  } /* Outer loop */

  do_output(stdout) ;
  iexit() ;
  exit(0) ;
 error:
  iexit() ;
  exit(1) ;
}

do_output(stream)
     FILE *stream ;
{
  int output_terms = 0, j ;
  if ( ! silent ) {
    if ( trec_topic && *trec_topic )
      fprintf(stderr, "Topic=%s. ", trec_topic) ;
    fprintf(stderr, "Output terms are:\n") ;
  }
  for ( j = 0 ; j <= Numterms ; j++ ) {
    if ( C_state->incl[j] ) {
      if ( j < startsize - 1 ) {
	fprintf(stream, "%s:%.3f:%s:%d:%.3f\n",
	       Terms[j].t, C_state->w[j], "", j, Terms[j].ow) ;
	if ( ! silent ) 
	  fprintf(stderr, "%s:%.3f:%s:%d:%.3f\n",
		  Terms[j].t, C_state->w[j], "", j, Terms[j].ow) ;
      }
      else {
	fprintf(stream, "%s:%.3f:%.3f:%d:%.3f\n",
	       Terms[j].t, C_state->w[j], C_state->scores[j], j, Terms[j].ow) ;
	if ( ! silent ) 
	  fprintf(stderr, "%s:%.3f:%.3f:%d:%.3f\n",
		  Terms[j].t, C_state->w[j], C_state->scores[j], j, Terms[j].ow) ;
      }
      output_terms++ ;
    }
  }
  if ( ! silent ) 
    fprintf(stderr, "Output total of %d terms, score %.3f\n",
	    output_terms, C_state->score) ;
  return output_terms ;
}

  
/* Needs the arrays + C_state->score, C_state->c_set, top_num, gran, score_code
   silent, Numterms, num_dropped, relset, N, R */

circular_add(num_passes)
     int num_passes ;
{
  char *message ;
  int j = startwith, passes = 0 ;
  int new_set, temp_set ;
  double start_br = C_state->score, temp_br ;
  double gain, cycle_gain ;
  int something_happened = 0 ;
  int last_one_changed, last_one_changed_previous_run = Numterms ;
  int added_this_cycle = 0, removed_this_cycle = 0 ;
  int w_change = 0, blank_passes = 0 ;

  char answer[256] ;

  while ( 1 ) {

    /* One cycle loop */
    if ( passes > 0 ) j = 0 ;
    something_happened = 0 ;
    start_br = C_state->score ;
    last_one_changed_previous_run = last_one_changed ;
    added_this_cycle = removed_this_cycle = 0 ;
    w_change = 0 ;

    if ( may_perturb && passes == passes_before_perturb ) {
      all_perturb() ;
      if ( ! passes_after_perturb ) break ;
    }

    for ( ; j < Numterms ; j++ ) {
      if ( passes > 0 && ! may_perturb &&
	   (! allvw || vw_change_fraction < .0001) &&
	   ! something_happened && j >= last_one_changed_previous_run ) {
	if ( ! silent )
	  fprintf(stderr, "Measure can't improve, quitting\n") ;
	break ;
      }
      if ( C_state->dropped[j] ) continue ;
      if ( ! allvw ) {
	if ( C_state->incl[j] ) {
	  if ( ! no_remove ) {
	    new_set = remove_set(C_state->c_set, j, C_state->cutoff) ;
	    message = "removing" ;
	  }
	  else continue ;
	}
	else { /* Not incl */
	  if ( ! no_add ) {
	    message = "adding" ;
	    new_set =
	      add_set(C_state->c_set, Terms[j].set, 0, C_state->cutoff,
		      C_state->w[j]/Terms[j].ow) ;
	  }
	  else continue ;
	}
	temp_br =
	  score(new_set, relset, N, R, &relret, top_num, score_code, gran) ;
	C_state->scores[j] = temp_br - C_state->score ;
	gain = (temp_br - C_state->score) * 100 / C_state->score ;
	if ( ! silent )
	  fprintf(stderr,
  "Score after %s #%d, \"%s\" (np %d, wt %.3f), is %.3f (%.2f%%), rcl=%.1f%%\n",
		  message, j, Terms[j].t, Terms[j].np, C_state->w[j],
		  temp_br - C_state->score, gain, (double)relret/(double)R*100) ;
	if ( gain > 0 ) {
	  something_happened = 1 ;
	  last_one_changed = j ;
	  change_since_remake++ ;
	  change_since_vw++ ;
	  if ( ! C_state->incl[j] ) {
	    C_state->inclterms++ ;
	    added_this_cycle++ ;
	  }
	  else {
	    C_state->inclterms-- ;
	    removed_this_cycle++ ;
	  }
	  if ( ! silent ) {
	    fprintf(stderr,
		    "%s \"%s\", term #%d, wt %.3f, incs score from %.3f to %.3f",
		    message, Terms[j].t, j, C_state->w[j], C_state->score,
		    temp_br) ;
	    if ( gain > 0.000001 ) fprintf(stderr, " (%.2f%%)\n", gain) ;
	    else fprintf(stderr, "\n") ;
	  }
	  C_state->score = temp_br ;
	  if ( ! Use_cutoff ) {
	    if ( top_num > 0 && ! C_state->incl[j] ) {
	      delete_set(new_set) ;
	      new_set =
		add_set(C_state->c_set, Terms[j].set, 0, C_state->cutoff,
			C_state->w[j]/Terms[j].ow) ;
	    }
	  }
	  else {
	    btm_wt = wt_of_nth(new_set, top_num) ;
	    C_state->cssize = setsize(new_set) ;
	    C_state->cutoff = btm_wt * Cutoff_factor ;
	    if ( ! silent ) 
	      fprintf(stderr,
		      "Wt of rec %d is %.3f, %d terms, cutoff %.3f, setsize %d\n",
		      top_num - 1, btm_wt, C_state->inclterms, C_state->cutoff,
		      C_state->cssize) ;
	  }
	  delete_set(C_state->c_set) ;
	  C_state->c_set = new_set ;
	  C_state->incl[j] = 1 - C_state->incl[j] ;
	  C_state->dropped[j] = 0 ; C_state->failures[j] = 0 ;
	  /* Haven't put this stuff in anywhere else (CB or CFP) yet */
	  if ( change_since_remake >= remake_thresh && remake_thresh > 0 ) {
	    remake() ;
	    change_since_remake = 0 ;
	  }
	  if ( change_since_vw >= vw_thresh && vw_thresh > 0 ) {
	    vw1(vwup, vwdown) ;
	    change_since_vw = 0 ;
	  }
	}
	else { /* (gain <= 0 ) */
	  delete_set(new_set) ;
	  C_state->failures[j]++ ;
	  if ( C_state->failures[j] >= successive_failures_max ) {
	    C_state->dropped[j] = 1 ;
	    num_dropped++ ;
	  }
	}
      }
      else { /* allvw */
	gain = vw(j, vwup, vwdown) ;
	if ( gain > 0 ) {
	  C_state->score += gain ;
	  C_state->dropped[j] = C_state->failures[j] = 0 ;
	  something_happened = 1 ;
	  last_one_changed = j ;
	  change_since_remake++ ;
	  change_since_vw++ ;
	  w_change++ ;
	  if ( ! C_state->incl[j] ) {
	    added_this_cycle++ ;
	    C_state->incl[j] = 1 ;
	  }
	}
	else {
	  C_state->failures[j]++ ;
	  if ( C_state->failures[j] >= successive_failures_max ) {
	    C_state->dropped[j] = 1 ;
	    num_dropped++ ;
	  }
	}
      } /* (allvw) */
    } /* ( cycle ) */

    if ( j < Numterms ) break ;	/* Case where no further gain possible */
    passes++ ;
    cycle_gain = (C_state->score - start_br) * 100 / start_br ;
    if ( ! silent ) {
      if ( passes == 1 ) fprintf(stderr, "1 pass completed\n") ;
      else fprintf(stderr, "%d passes completed\n", passes) ;
      fprintf(stderr, "Increase in score this cycle %.3f (%.2f%%)\n",
	      C_state->score - start_br, cycle_gain) ;
      fprintf(stderr, "%d term(s) added, %d term(s) removed\n",
	      added_this_cycle, removed_this_cycle) ;
      fprintf(stderr, "%d weights changed this cycle\n", w_change) ;
    }
    trec_score_report() ;
    if ( interrupt_received >= 2 ) {
      if ( ! silent )
	fprintf(stderr, "Quitting because two interrupts received\n") ;
      break ;
    }
    if ( ! something_happened ) blank_passes++ ;
    else blank_passes = 0 ;

    if ( ! may_perturb && ( ! allvw || vw_change_fraction < .0001) &&
	 ( ! something_happened || cycle_gain < gain_thresh ) ) {
      if ( ! silent )
	fprintf(stderr,
		"Complete cycle with gain %.2f%%, quitting\n", cycle_gain) ;
      break ;
    }
    else if ( ! may_perturb && allvw && vw_change_fraction >= .0001 ) {
      if ( blank_passes >= max_blank_passes ) {
	fprintf(stderr, "%d cycles with negligible gain, quitting\n",
		blank_passes) ;
	break ;
      }
      else if ( vwup <= vwup_min ) {
	fprintf(stderr, "Weight change fraction %.3f too small to continue\n",
		vwup) ;
	break ;
      }
    }
    if ( passes >= num_passes ) {
      if ( ! silent )
	fprintf(stderr, "Leaving circular_add() because done %d passes\n",
		num_passes) ;
      if ( may_perturb ) continue ;
      break ;
    }
    if ( vw_change_fraction > .001 ) {
      if ( vwup * vw_change_fraction > vwup_min ) {
	fprintf(stderr, "Changing vwup from %.3f to %.3f\n",
		vwup, vwup * vw_change_fraction) ;
	fprintf(stderr, "Changing vwdown from %.3f to %.3f\n",
		vwdown, vwdown * vw_change_fraction) ;
	vwup *= vw_change_fraction ;
	vwdown *= vw_change_fraction ;
      }
      else {
	if ( vwup > vwup_min ) {
	  fprintf(stderr, "Changing vwup from %.3f to %.3f\n", vwup, vwup_min) ;
	  fprintf(stderr, "Changing vwdown from %.3f to %.3f\n",
		  vwdown, vwdown_min) ;
	}
	vwup = vwup_min ;
	vwdown = vwdown_min ;
      }
    }
  } /* (while ( 1 ) */
  if ( vwend > 0 ) {
    double vgain ;
    if ( ! silent ) {
      fprintf(stderr, "Starting %d final weight variation passes\n", vwend) ;
      fprintf(stderr,
	      "Changing vwup from %.3f to %.3f\n", vwup, vwup * 1.2) ;
      fprintf(stderr,
	      "Changing vwdown from %.3f to %.3f\n", vwdown, vwdown * 1.2);
    }
    vwup *= 1.2 ; vwdown *= 1.2 ;
    for ( j = 0 ; j < vwend ; j++ ) {
      vgain = vw1(vwup, vwdown) ;
      if ( ! silent )
	fprintf(stderr,
		"Done weight variation pass %d of %d, gain %.3f\n",
		j + 1, vwend, vgain) ;
      if ( vgain <= 0 && j < vwend - 1 ) {
	if ( ! silent ) {
	  fprintf(stderr,
		  "Changing vwup from %.3f to %.3f\n", vwup, vwup * .75) ;
	  fprintf(stderr,
		  "Changing vwdown from %.3f to %.3f\n", vwdown, vwdown * .75);
	}
	vwup *= .75 ;
	vwdown *= .75 ;
      }
      if ( vgain > 0 && j < vwend - 1 ) trec_score_report() ;
    }
  }
  trec_score_report() ;
}



void
interrpt()
{
  interrupt_received++ ;
  if ( interrupt_received < 3 )
    fprintf(stderr, "SIGUSR1 received, kill again to hasten end\n") ;
  if ( interrupt_received == 2 )
    fprintf(stderr, "Will stop at end of iteration\n") ;
  else if ( interrupt_received > 2 ) {
    fprintf(stderr, "Will hup after outputting anything available\n") ;
    do_output(stdout) ;
    kill(getpid(), 1) ;
    iexit() ;
    exit(1) ;
  }
  signal(SIGUSR1, interrpt) ;
}

void
dump_output()
{
  static gen = 0 ;
  char buff[512] ;
  FILE *f ;
  sprintf(buff, "/tmp/rtg_select.output.%d.%d", gen++, getpid()) ;
  if ( (f = fopen(buff, "w")) == NULL ) 
    fprintf(stderr, "Can't open %s for write\n", buff) ;
  else {
    if ( ! silent )
      fprintf(stderr, "Writing current output state to %s\n", buff) ;
    do_output(f) ;
    fclose(f) ;
  }
  signal(SIGUSR2, dump_output) ;
}

double
vw(termnum, up, down)
     int termnum ;
     double up, down ;
{
  double ff, alpha ;
  double br, gain ;
  int new_s ;
  int j = termnum ;
  double gmax, prov_wt ;

  /* If not already included try it at current weight */
  if ( ! C_state->incl[j] ) {
    alpha = 0 ;
    ff = C_state->w[j] / Terms[j].ow + alpha ;
    if ( ! silent )
      fprintf(stderr, "#%d (\"%s\") at %.3f ..",
	      j, Terms[j].t, C_state->w[j]) ;
    new_s =
      add_set(C_state->c_set, Terms[j].set, 0, C_state->cutoff, ff) ;
    br = score(new_s, relset, N, R, &relret, top_num, score_code, gran) ;
    gain = br - C_state->score ;
    if ( gain > 0 ) {
      delete_set(C_state->c_set) ;
      C_state->c_set = new_s ;
      btm_wt = wt_of_nth(new_s, top_num) ;
      C_state->cssize = setsize(new_s) ;
      C_state->cutoff = btm_wt * Cutoff_factor ;
      C_state->inclterms++ ;
      if ( ! silent )
	fprintf(stderr,
	  " %.3f --> %.3f (rcl=%.1f) [nt=%d ss=%d b/wt=%.3f]\n",
	  C_state->score, br, (double)relret/(double)R * 100.0,
	  C_state->inclterms, C_state->cssize, btm_wt) ;
      return gain ;
    }
    else { /* (gain <= 0 ) (and not incl) */
      delete_set(new_s) ;
      gmax = gain ;
      prov_wt = C_state->w[j] ;
      if ( ! silent )
	fprintf(stderr, " %.3f --> %.3f (rcl=%.1f)\n",
		C_state->score, br, (double)relret/(double)R * 100.0) ;
    }
  } /* (not incl) */

  alpha = -down ;		/* Try reduce first */
  if ( C_state->incl[j] ) {
    fprintf(stderr, "CW #%d (\"%s\") %.3f --> %.3f ..",
	    j, Terms[j].t, C_state->w[j], C_state->w[j] * (1 + alpha)) ;
    ff = alpha * C_state->w[j] / Terms[j].ow ;
  }
  else {
    fprintf(stderr, "#%d (\"%s\") at %.3f ..",
	    j, Terms[j].t, C_state->w[j] * (1 + alpha)) ;
    ff = C_state->w[j] / Terms[j].ow * (1 + alpha) ;
  }
  new_s = add_set(C_state->c_set, Terms[j].set, 0, C_state->cutoff, ff) ;
  br = score(new_s, relset, N, R, &relret, top_num, score_code, gran) ;
  gain = br - C_state->score ;
  if ( gain > 0 ) {
    delete_set(C_state->c_set) ;
    C_state->c_set = new_s ;
    btm_wt = wt_of_nth(new_s, top_num) ;
    C_state->cssize = setsize(new_s) ;
    C_state->w[j] *= (1 + alpha) ;
    C_state->cutoff = btm_wt * Cutoff_factor ;
    if ( ! C_state->incl[j] ) C_state->inclterms++ ;
    if ( ! silent )
      fprintf(stderr, " %.3f --> %.3f (rcl=%.1f) [nt=%d ss=%d b/wt=%.3f]\n",
	      C_state->score, br, (double)relret/(double)R * 100.0,
	      C_state->inclterms, C_state->cssize, btm_wt) ;
    return gain ;
  }
  else {
    delete_set(new_s) ;
    if ( ! silent )
    fprintf(stderr, " %.3f --> %.3f (rcl=%.1f)\n",
	    C_state->score, br, (double)relret/(double)R * 100.0) ;
    if ( ! C_state->incl[j] && gain > gmax ) {
      gmax = gain ;
      prov_wt = C_state->w[j] * (1 + alpha) ; /* Check! */
    }
  }
  alpha = up ;		/* Finally, if still here, try increase */
  if ( C_state->incl[j] ) {
    if ( ! silent ) 
      fprintf(stderr, "CW #%d (\"%s\") %.3f --> %.3f ..",
	      j, Terms[j].t, C_state->w[j], C_state->w[j] * (1 + alpha)) ;
    ff = alpha * C_state->w[j] / Terms[j].ow ;
  }
  else {
    if ( ! silent ) 
      fprintf(stderr, "#%d (\"%s\") at %.3f ..",
	      j, Terms[j].t, C_state->w[j] * (1 + alpha)) ;
    ff = C_state->w[j] / Terms[j].ow * (1 + alpha) ;
  }
  new_s = add_set(C_state->c_set, Terms[j].set, 0, C_state->cutoff, ff) ;
  br = score(new_s, relset, N, R, &relret, top_num, score_code, gran) ;
  gain = br - C_state->score ;
  if ( gain > 0 ) {
    delete_set(C_state->c_set) ;
    C_state->c_set = new_s ;
    btm_wt = wt_of_nth(new_s, top_num) ;
    C_state->cssize = setsize(new_s) ;
    C_state->cutoff = btm_wt * Cutoff_factor ;
    C_state->w[j] *= (1 + alpha) ;
    if ( ! C_state->incl[j] ) C_state->inclterms++ ;
    if ( ! silent )
      fprintf(stderr, " %.3f --> %.3f (rcl=%.1f) [nt=%d ss=%d b/wt=%.3f]\n",
	      C_state->score, br, (double)relret/(double)R * 100.0,
	      C_state->inclterms, C_state->cssize, btm_wt) ;
    return gain ;
  }
  else {
    delete_set(new_s) ;
    if ( ! silent )
    fprintf(stderr, " %.3f --> %.3f (rcl=%.1f)\n",
	    C_state->score, br, (double)relret/(double)R * 100.0) ;
    if ( ! C_state->incl[j] ) {
      if ( gain > gmax ) prov_wt = C_state->w[j] * 1 + alpha ; /* Check! */
      if ( do_prov_wts && fabs(prov_wt - C_state->w[j]) > .0001 ) {
	fprintf(stderr,
		"Provisionally changing weight from %.3f to %.3f\n",
		C_state->w[j], prov_wt) ;
	C_state->w[j] = prov_wt ;
      }
    }
    return gain ;
  }
}

double
vw1(up, down)
     double up, down ;
{
  double ff, alpha ;
  double br ;
  double gain ;
  double cum_gain = 0.0 ;
  int new_s ;
  int j ;
  for ( j = 0 ; j < Numterms ; j++ ) {
    if ( ! C_state->incl[j] ) continue ;
    alpha = -down ;		/* Try reduce by 1/3, say */
    ff = alpha * C_state->w[j] / Terms[j].ow ;
    if ( ! silent )
#ifdef SHORT_MESS
      fprintf(stderr, "CW #%d (\"%s\") %.3f --> %.3f ..",
	      j, Terms[j].t, C_state->w[j], C_state->w[j] * (1 + alpha)) ;
#else
      fprintf(stderr, "Changing wt of #%d (\"%s\") %.3f --> %.3f ..",
	      j, Terms[j].t, C_state->w[j], C_state->w[j] * (1 + alpha)) ;
#endif
    new_s = add_set(C_state->c_set, Terms[j].set, 0, C_state->cutoff, ff) ;
    br = score(new_s, relset, N, R, &relret, top_num, score_code, gran) ;
    if ( ! silent ) fprintf(stderr, " %.3f --> %.3f (rcl=%.1f)\n",
			    C_state->score, br, (double)relret/(double)R * 100.0) ;
    if ( (gain = (br - C_state->score)) > 0 ) {
      delete_set(C_state->c_set) ;
      C_state->c_set = new_s ;
      C_state->score = br ;
      btm_wt = wt_of_nth(new_s, top_num) ;
      C_state->cssize = setsize(new_s) ;
      C_state->w[j] *= (1 + alpha) ;
      C_state->cutoff = btm_wt * Cutoff_factor ;
      if ( Debug ) 
	fprintf(stderr, "Wt of rec %d is %.3f, %d terms, cutoff %.3f\n",
		top_num - 1, btm_wt, C_state->inclterms, C_state->cutoff) ;
      cum_gain += gain ;
    }
    else {
      delete_set(new_s) ;
      alpha = up ;		/* Try increase by 1/2, say */
      ff = alpha * C_state->w[j] / Terms[j].ow ;
      if ( ! silent )
#ifdef SHORT_MESS
	fprintf(stderr, "CW #%d (\"%s\") %.3f --> %.3f ..",
		j, Terms[j].t, C_state->w[j], C_state->w[j] * (1 + alpha)) ;
#else
	fprintf(stderr, "Changing wt of #%d (\"%s\") %.3f --> %.3f ..",
		j, Terms[j].t, C_state->w[j], C_state->w[j] * (1 + alpha)) ;
#endif
      new_s = add_set(C_state->c_set, Terms[j].set, 0, C_state->cutoff, ff) ;
      br = score(new_s, relset, N, R, &relret, top_num, score_code, gran) ;
      if ( ! silent )
	fprintf(stderr, " %.3f --> %.3f (rcl=%.1f)\n",
		C_state->score, br, (double)relret/(double)R * 100.0) ;
      if ( (gain = (br - C_state->score)) > 0 ) {
	delete_set(C_state->c_set) ;
	C_state->c_set = new_s ;
	C_state->score = br ;
	btm_wt = wt_of_nth(new_s, top_num) ;
	C_state->cssize = setsize(new_s) ;
	C_state->cutoff = btm_wt * Cutoff_factor ;
	C_state->w[j] *= (1 + alpha) ;
	if ( Debug ) 
	  fprintf(stderr, "Wt of rec %d is %.3f, %d terms, cutoff %.3f\n",
		  top_num - 1, btm_wt, C_state->inclterms, C_state->cutoff) ;
	cum_gain += gain ;
      }
      else delete_set(new_s) ;
    }
  }

  return cum_gain ;
}

/* Mustn't use if reweighting */

remake()
{
  int temp_set ;
  double br ;
  delete_set(C_state->c_set) ;
  C_state->c_set = remake_current_set(0, (double)0) ;
  btm_wt = wt_of_nth(C_state->c_set, top_num) ;
  C_state->cssize = setsize(C_state->c_set) ;
  C_state->cutoff = btm_wt * Cutoff_factor ;
  if ( ! silent ) {
    fprintf(stderr,
	    "Wt of rec %d is %.3f, %d terms, cutoff %.3f, setsize %d\n",
	    top_num - 1, btm_wt, C_state->inclterms, C_state->cutoff, C_state->cssize) ;
    fprintf(stderr, "Reducing current set to cutoff weight \n") ;
  }
  temp_set  = reduce_set(C_state->c_set, C_state->cutoff) ;
  delete_set(C_state->c_set) ;
  C_state->c_set = temp_set ;
  br = score(C_state->c_set, relset, N, R, &relret, top_num, score_code, gran) ;
  if ( ! silent ) {
    fprintf(stderr,
	    "Score check: %.3f, rcl=%.3f: prev score %.3f\n",
	    br, (double)relret/(double)R, C_state->score) ;
  }
  if ( br != C_state->score ) {
    C_state->score = br ;
    if ( ! silent ) fprintf(stderr, "Correcting score to new value\n") ;
  }
}


trec_score_report()
{
  double avep ;
  avep = score(C_state->c_set, relset, N, R, &relret, 1000, TREC_AVEP, gran) ;
  if ( ! silent )
    fprintf(stderr, "Current TREC rating: avep=%.3f (rcl=%.1f)\n",
	    avep, (double)relret/(double)R*100.0) ;
}

/******************************************************************************

  State functions  SW June 97

******************************************************************************/

init_state(s)
     struct state *s ;
{
  int j ;
  clearstate(s) ;
  s->c_set = s->new_set = -1 ;		/* I.e. no set */
  s->live = 1 ;
}


struct state *
getstate()
{
  int j, found ;
  struct state *s ;
  for ( j = 0, found = 0 ; j < MAXSTATES ; j++ ) {
    if ( ! States[j].live ) {
      found = 1 ;
      break ;
    }
  }
  if ( ! found ) return (struct state *)NULL ;
  else {
    s = &States[j] ;
    init_state(s) ;
    return s ;
  }
}

clearstate(s)
     struct state *s ;
{
  if ( s->live && s->c_set >= 0 ) delete_set(s->c_set) ;
  if ( s->live && s->new_set >= 0 ) delete_set(s->new_set) ;
  memset((char *)s, 0, sizeof(struct state)) ;
  s->c_set = s->new_set = -1 ;
}

copystate(s, t)			/* Copies state t into state s */
				/* Returns >=0 if OK else <0 */
     struct state *s, *t ;
{
  clearstate(s) ;		/* Ensures any set deleted */
  memcpy((char *)s, (char *)t, sizeof(struct state)) ;
  if ( t->c_set >= 0 ) s->c_set = copy_set(t->c_set) ;
  return s->c_set ;
}

plu(p)
     double p ;
{
  if ( p < (double)(rand() & 0x7fff)/32767 ) return 0 ;
  else return 1 ;
}

boltz(new_sc, prev_sc, T, noisy)
     double new_sc, prev_sc ;		/* New score, prev score */
     double T ;
     int noisy ;
{
  double exponent, p ;

  if ( T > .0001 ) {
    exponent = (prev_sc - new_sc)/T ;
    p = exp(-exponent) ;
    if ( noisy ) 
      fprintf(stderr, " [Boltz: T=%.2f, exp(-%.3f)=%.3f] ",
	      T, exponent, p) ;
    return plu(p) ;
  }
  else return 0 ;
}

perturb(s, p, up_p)
     struct state *s ;
     double p ;			/* Probability a term has its weight changed */
     double up_p ;		/* Proportional change for up */
{
  int j ;
  int sets[MAXTERMS] ;
  int num_sets ;
  int relret ;
  double down_p = up_p / (1 + up_p) ;
  double ff[MAXTERMS] ;			/* frig factor */
  double alpha ;

  memcpy((char *)s->nw, (char *)s->w, sizeof(s->w)) ; /* Hack */
  /* Decide what to change and by how much */
  for ( j = 0, num_sets = 0 ; j < MAXTERMS ; j++ ) {
    if ( ! s->incl[j] ) continue ;
    if ( ! plu(p) ) continue ;
    if ( plu(.5) ) alpha = up_p ;
    else alpha = -down_p ;
    ff[num_sets] = alpha * s->w[j]/Terms[j].ow ;
    s->nw[j] = s->w[j] * (1 + alpha) ;
    sets[num_sets++] = Terms[j].set ;
    s->failures[j] = s->dropped[j] = s->skipped[j] = 0 ;
  }
  /* Do it and score */
  s->new_set = add_sets(s->c_set, sets, num_sets, 0, s->cutoff, ff) ;
  s->new_score =
    score(s->new_set, relset, N, R, &relret, top_num, score_code, gran) ;
  s->new_recall = (double)(relret * 100)/(double)R ;
  return num_sets ;
}
  
void
retain_new(s)
     struct state *s ;
{
  if ( s->c_set >= 0 ) delete_set(s->c_set) ;
  s->c_set = s->new_set ; s->new_set = -1 ;
  s->cssize = setsize(s->c_set) ;
  fprintf(stderr, " [Setsize %d] ", s->cssize) ;
  s->cutoff = wt_of_nth(s->c_set, top_num) * Cutoff_factor ;
  s->score = s->new_score ; 
  s->recall = s->new_recall ;
  memcpy((char *)s->w, (char *)s->nw, sizeof(s->nw)) ;
}

static
cmp_dbl(d1, d2)
     double *d1, *d2 ;
{
  return *d1 > *d2 ? 1 : -1 ;
}

check_state(s)
     struct state *s ;
{
  double sort_w[MAXTERMS] ;
  double med ;
  int numterms = 0 ;
  int i ;

  for ( i = 0 ; i < MAXTERMS ; i++ ) {
    if ( ! s->incl[i] ) continue ;
    sort_w[numterms++] = s->w[i] ;
  }
  gqsort((char *)sort_w, numterms, sizeof(double), cmp_dbl) ;
  med = sort_w[numterms/2] ;
  fprintf(stderr, "Checking current state, median wt %.3f\n", med) ;
  for ( i = 0 ; i < MAXTERMS ;i++ ) {
    if ( ! s->incl[i] ) continue ;
    if ( s->w[i] < P_DROP_FACTOR * med ) {
      fprintf(stderr, "Dropping %s, w=%.3f, median weight %.3f\n",
	      Terms[i].t, s->w[i], med) ;
      s->incl[i] = 0 ;
    }
  }
}

/* Start with C_state pointing to real state, Best and Realbest empty */

all_perturb()
{
  int iters, stage ;
  int n_succ, n_real_bests, n_fails, n_keeps ;
  int n_stage_succ, n_stage_real_bests, n_stage_fails, n_stage_keeps ;
  double stage_start_score ;
  int n_perts = 0 ;
  int num_changed ;
  double T = P_initial_T ;	/* Default 2.0 */
  double change_prob, mean_wt_change ;	/* args for perturb() */
  double p_bigstep_wt_change, p_smallstep_wt_change ;
  int iters_since_succ ;
  int cond ;

  copystate(Best_state, C_state) ;
  copystate(Realbest_state, C_state) ;
  p_bigstep_wt_change = P_bigstep_wt_change ;
  p_smallstep_wt_change = P_smallstep_wt_change ;

  for ( T = P_initial_T, stage = 0 ; T > 0 ;
	T *= T_mult, p_smallstep_wt_change *= T_mult,
	  p_bigstep_wt_change *= T_mult, stage++  ) {
    if ( T <= T_min ) T = 0 ;
    if ( ! silent )
      fprintf(stderr, "Starting stage %d, T=%.2f", stage, T) ;
    if ( T < .00001 ) fprintf(stderr, ", final quench") ;
    fprintf(stderr, "\n") ;
    if ( (Stage_final_state[stage] = getstate()) == NULL ) {
      fprintf(stderr, "Run out of State records\n") ;
      break ;
    }
    if ( (T < .00001 || P_retain_real_best) &&
	 Realbest_state->score > C_state->score ) {
      fprintf(stderr,
	      "RB=%.3f (from stage %d), CB=%.3f, reverting to real best\n",
	      Realbest_state->score, Realbest_state->stage, C_state->score) ;
      copystate(C_state, Realbest_state) ;
      copystate(Best_state, C_state) ;
    }
    stage_start_score = C_state->score ;
    n_stage_succ = n_stage_real_bests = n_stage_fails = n_stage_keeps = 0 ;
    for ( iters = n_succ = n_fails = n_keeps = iters_since_succ = 0, cond = 1 ;
	  cond ; iters++ ) {
      if ( P_fixed_iters ) cond = (iters < 50 * P_stage_size) ;
      else cond = (iters < 50 * P_stage_size && n_succ < 5 * P_stage_size) ; 
      if ( plu(P_bigstep_p) ) {
	change_prob = P_bigstep_change_p ;
	mean_wt_change = p_bigstep_wt_change ;
      }
      else {
	change_prob = P_smallstep_change_p ;
	mean_wt_change = p_smallstep_wt_change ;
      }
      if ( ! silent )
	fprintf(stderr, "Pert=%d, P(ch)=%.3f, incr=%.3f .. ",
		n_perts + 1, change_prob, mean_wt_change) ;
      num_changed = perturb(C_state, change_prob, mean_wt_change) ;
      C_state->stage = stage ;
      n_perts++ ;
      if ( ! silent ) fprintf(stderr, "%d wts ch, ", num_changed) ;
      if ( C_state->new_score >= Best_state->score ) {
	n_succ++ ;
	n_stage_succ++ ;
	if ( ! silent )
	  fprintf(stderr,
		  "succ=%d, new %.3f (rcl=%.1f), prev best %.3f, retaining",
		  n_succ, C_state->new_score, C_state->new_recall,
		  Best_state->score) ;
	retain_new(C_state) ;
	copystate(Best_state, C_state) ;
	iters_since_succ = 0 ;
	if ( C_state->score >= Realbest_state->score ) {
	  /* Not sure about the =, changed it after about run43ap */
	  check_state(C_state) ;
	  copystate(Realbest_state, C_state) ;
	  n_real_bests++ ; n_stage_real_bests++ ;
	  if ( ! silent ) fprintf(stderr, " as `real' best\n") ;
	}
	else if ( ! silent ) fprintf(stderr, "\n") ;
      }
      else {			/* Score < best score */
	n_fails++ ; n_stage_fails++ ;
	iters_since_succ++ ;
	if ( ! silent )
	  fprintf(stderr, "fails=%d, new %.3f, best %.3f,",
		  n_fails, C_state->new_score, Best_state->score) ;
	if ( boltz(C_state->new_score * 1000, Best_state->score * 1000, T, 1) ) {
	  retain_new(C_state) ;
	  n_keeps++ ; n_stage_keeps++ ;
	  if ( ! silent ) 
    fprintf(stderr, " keeping %.3f (rcl=%.1f) 'tho current best is %.3f\n",
		    C_state->score, C_state->recall, Best_state->score) ;
	  copystate(Best_state, C_state) ;
	}
	else {			/* forget current state */
	  delete_set(C_state->new_set) ;
	  if ( ! silent ) 
	    fprintf(stderr, " reverting\n") ;
	}
      } /* (score worse) */
      if ( iters_since_succ > P_successive_fails_max ) {
     fprintf(stderr, "Terminating stage %d because %d successive failures\n",
		stage, iters_since_succ) ;
	break ;
      }
    } /* ( for iters .. ) */
    /* End of stage procedures */
    if ( ! silent ) {
      fprintf(stderr,
  "Stage %d stats: gain %.3f succ %d (real bests %d) fails %d (keeps %d)\n",
	      stage, C_state->score - stage_start_score,
	      n_stage_succ, n_stage_real_bests, n_stage_fails, n_stage_keeps) ;
      copystate(Stage_final_state[stage], C_state) ;
    }
  } /* ( stage loop) */
}
