/******************************************************************************

  Routing 1994

  rtgterms_pass2.c

  Modifying for TREC-5 May 96 SW

  Read rels for a topic and build relset.
  Read output from pass1 on stdin and for each term get r and n, and ttf,
  F4 weight and rsv
  
  Input is <term>:<rtf> on stdin (where rtf means tf within relset), and
  relfile, topnum and file containing grand total term frequency within
  reldocs as args.

  Output is
  
    <topnum>:<np>:<R>:<r>:<cttf>:<rgttf>:<rtf>:<F4>:<rsv>:<term>:[<gsl-1>]

  on stdout.

  SW June 94
  May 95: revised for TREC4
  May 99: modified so will output whatever the RSV value

******************************************************************************/

#undef DEBUG

#define I0q if (res = i0(command, response) < 0) {\
  sprintf(command, "perror") ;\
  i0(command, response) ;\
  fprintf(stderr, response) ;\
  goto error ; }

#ifdef DEBUG
#define I0 fprintf(stderr, "%s\n", command) ;\
if (res = i0(command, response) < 0) {\
  sprintf(command, "perror") ;\
  i0(command, response) ;\
  fprintf(stderr, response) ;\
  goto error ; }\
  else fprintf(stderr, response) ;
#else
#define I0 I0q
#endif

#include "bss.h"
#include <string.h>

double weight_to_s_v() ;

char buf[1024] ;
char command[4 * 1024 * 1024] ;
char response[6 * 1024 * 1024] ;

main(argc, argv)
     int argc ;
     char **argv ;
{
  int qnum, qlen ;
  int limit ;
  int N ;
  FILE *rels ;
  FILE *f ;
  FILE *tot_tf_file ;
  int set, newset, res, setnum ;
  int relset ;
  int R, r ;
  int i, j, k ;
  int numterms, termlen ;
  int topnum ;
  int zero_rsv = 0, small_rsv = 0, zero_np = 0, zero_r = 0 ;
  struct term {
    char term[2048] ;
    int np ;
    int r ;
    int ttf ;
    int qtf ;
    float f4weight ;
    float finalweight ;		/* Appears unused (1996) */
    float rsv ;
  } terms ;
  struct term *t ;
  int ttf ;
  int Rtf = 0 ;
  int numpos, doclen, coord, tf, str ;
  char *envp ;
  char *dbname ;
  char *dn_attr ;
  char *p ;

  if ( argc < 3 ) {
    fprintf(stderr,
	    "pass2: usage: pass2 <relfile> <topnum> [<tot_tf_file>]\n") ;
    exit(1) ;
  }

  if ( (rels = fopen(argv[1], "r")) == NULL ) {
    fprintf(stderr, "pass2: can't open relfile %s\n", argv[1]) ;
    exit(1) ;
  }

  topnum = atoi(argv[2]) ;

  if ( argc >= 4 ) tot_tf_file = fopen(argv[3], "r") ;
  if ( tot_tf_file != NULL ) fscanf(tot_tf_file, "%d", &Rtf) ;

  dn_attr = "dn" ;

  iinit() ;
  if ( (envp = getenv("BSS_DB")) != NULL && *envp != '\0' )
    dbname = envp ;
  else {
    fprintf(stderr, "No database in environment (BSS_DB)\n") ;
    goto error ;
  }
  /* Make relset */
  sprintf(command, "ch %s", dbname) ;
  I0 ;
  if ( (envp = getenv("BSS_LIMIT")) != NULL && *envp != '\0' ) {
    sprintf(command, "set l=%s", envp) ;
    I0 ;
  }
  if ( (envp = getenv("BSS_ATTRIBUTE")) != NULL && *envp != '\0' ) {
    sprintf(command, "set a=%s", envp) ;
    I0 ;
  }
  if ( (envp = getenv("BSS_DN_ATTRIBUTE")) != NULL && *envp != '\0' ) dn_attr = envp ;
  sprintf(command, "set mem=1") ;
  I0q ;
  fprintf(stderr, "pass2: %d: forming relset\n", topnum) ;
  sprintf(command, "f a=%s t=zzzzzzzzzzzzzz", dn_attr) ;
  I0q ;
  set = atoi(response + 1) ;
  f = rels ;
  R = 0 ;
  while ( 1 ) {
    fgets(buf, 32, f) ;
    if ( feof(f) ) break ;
    *(buf + strlen(buf) - 1) = '\0' ; /* Doesn't cope with trailing \n */
    sprintf(command, "f a=%s t=%s", dn_attr, buf) ;
    I0q ;
    setnum = atoi(response + 1) ;
    if ( atoi(strstr(response, "np=") + 3) == 0 ) {
      fprintf(stderr, "%s not found\n", buf) ;
      /* doc not found, should only be because of limit */
      sprintf(command, "delete %d", atoi(response + 1) ) ;
      I0 ;
      continue ;
    }
    sprintf(command, "f s=%d s=%d op=or", set, setnum) ;
    I0q ;
    newset = atoi(response + 1) ;
    sprintf(command, "delete %d", setnum) ;
    I0q ;
    sprintf(command, "delete %d", set) ;
    I0q ;
    set = newset ;
    R++ ;
  }

  relset = set ;
  fprintf(stderr, "pass2: %d: now have relset, %d recs\n", topnum, R) ;  

  /* Read the term records from pass1.
     These are <term>:<qtf> */

  numterms = 0 ;
  while ( 1 ) {
    gets(buf) ;
    if ( feof(stdin) ) break ;
    sscanf(buf, "%[^:]:%d", terms.term, &terms.qtf) ;
    if ( *(terms.term) == '\0' ) { 
      fprintf(stderr, "pass2: %d: empty term\n", topnum) ;
      continue ;
    }
    if ( ! isalnum(*(terms.term)) && *(terms.term) != '@' ) {
      fprintf(stderr, "pass2: %d: spurious term '%s'\n", topnum, terms.term) ;
      continue ;
    }
    if ( strchr(terms.term, '\'') ) {
      fprintf(stderr, "pass2: %d: possibly dodgy term '%s' (not skipped)\n", topnum, terms.term) ;
    }
    numterms++ ;

    t = &terms ;
    sprintf(command, "f t=%s", t->term) ;
    I0
    t->np = atoi(strstr(response, "np=") + 3) ;
    setnum = atoi(response + 1) ;
    t->ttf = Setrecs[setnum]->ttf ;
    if ( t->ttf == 0 ) {
      fprintf(stderr, "ttf=0, term=%s\n", t->term) ;
    }
    sprintf(command, "f s=%d s=%d save=0 op=and nopos=1", setnum, relset) ;
    I0
    t->r = atoi(response + 3) ;
    sprintf(command, "delete %d", setnum) ;
    I0
    if ( t->np == 0 ) {
      zero_np++ ;
      fprintf(stderr, "Term %s: no postings, skipping\n", t->term) ;
      continue ;
    }
    if ( t->r == 0 ) {
      zero_r++ ;
      if ( t->qtf > 0 ) {	/* Note: if a topic term but not extracted
				   will have zero here. Should then be
				   output. See rtgterms. */
	fprintf(stderr, "Term %s: extracted but zero r, skipping\n", t->term) ;
	continue ;
      }
    }
    sprintf(command, "w function=1 n=%d r=%d bigr=%d", t->np, t->r, R) ;
    I0
    t->f4weight = atof(response) ;
    t->rsv = weight_to_s_v((double)t->f4weight, R, t->r) ;
    printf("%d:%d:%d:%d:%d:%d:%d:%.3f:%.3f:%s",
	   topnum, t->np, R, t->r, t->ttf, Rtf, t->qtf,
	   t->f4weight, t->rsv, t->term) ;
    if ( *(t->term) == '@' ) {
      sprintf(command, "gsl t=%s", t->term) ;
      I0
        printf(":%s", response) ;
    }
    else printf(":\n") ;
    if ( t->rsv < .1 && t->rsv > 0.0 ) small_rsv++ ;
    else if ( t->rsv <= 0.0 ) zero_rsv++ ;
  }
  fflush(stdout) ;
  if ( zero_np > 0 )
    fprintf(stderr, "pass2: %d: Zero np, skipped: %d\n", topnum, zero_np) ;
  if ( zero_r > 0 )
    fprintf(stderr, "pass2: %d: Zero r: %d\n", topnum, zero_r) ;
  if ( zero_rsv > 0 )
    fprintf(stderr, "pass2: %d: Zero rsv, not skipped: %d\n",
	    topnum, zero_rsv) ;
  if ( small_rsv > 0 )
    fprintf(stderr, "pass2: %d:  0 < Rsv < .1 (not skipped): %d\n",
	    topnum, small_rsv);
  exit(0) ;

 error:
  iexit() ;
  exit(1) ;

}


